/* Generated by Cython 0.28.3 */

#ifndef __PYX_HAVE__cupy__cuda__thrust
#define __PYX_HAVE__cupy__cuda__thrust


#ifndef __PYX_HAVE_API__cupy__cuda__thrust

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C char *cupy_malloc(void *, size_t);
__PYX_EXTERN_C void cupy_free(void *, char *);

#endif /* !__PYX_HAVE_API__cupy__cuda__thrust */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initthrust(void);
#else
PyMODINIT_FUNC PyInit_thrust(void);
#endif

#endif /* !__PYX_HAVE__cupy__cuda__thrust */
