/* Generated by Cython 0.28.3 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "define_macros": [
            [
                "_FORCE_INLINES",
                "1"
            ]
        ],
        "depends": [
            "cupy/cuda/cupy_cuComplex.h",
            "cupy/cuda/cupy_cusolver.h"
        ],
        "extra_compile_args": [
            "-fopenmp"
        ],
        "extra_link_args": [
            "-fopenmp",
            "-Wl,--disable-new-dtags,-rpath,/usr/local/cuda/lib64"
        ],
        "include_dirs": [
            "cupy/cuda",
            "/usr/local/cuda/include"
        ],
        "language": "c++",
        "libraries": [
            "cusolver"
        ],
        "library_dirs": [
            "/usr/local/cuda/lib64"
        ],
        "name": "cupy.cuda.cusolver",
        "sources": [
            "cupy/cuda/cusolver.pyx"
        ]
    },
    "module_name": "cupy.cuda.cusolver"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_3"
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(U other) { return *ptr == other; }
    template<typename U> bool operator !=(U other) { return *ptr != other; }
  private:
    T *ptr;
};

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__cupy__cuda__cusolver
#define __PYX_HAVE_API__cupy__cuda__cusolver
/* Early includes */
#include <stdint.h>
#include "cupy_cuComplex.h"
#include "cupy_cusolver.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "cupy/cuda/cusolver.pyx",
  "cupy/cuda/cusolver.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/*--- Type declarations ---*/

/* "cupy/cuda/driver.pxd":39
 * 
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     CU_JIT_INPUT_CUBIN = 0
 *     CU_JIT_INPUT_PTX = 1
 */
enum  {
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_CUBIN = 0,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_PTX = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_FATBINARY = 2,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_OBJECT = 3,
  __pyx_e_4cupy_4cuda_6driver_CU_JIT_INPUT_LIBRARY = 4,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_MAX_THREADS_PER_BLOCK = 0,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_SHARED_SIZE_BYTES = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_CONST_SIZE_BYTES = 2,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_LOCAL_SIZE_BYTES = 3,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_NUM_REGS = 4,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_PTX_VERSION = 5,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_BINARY_VERSION = 6,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_CACHE_MODE_CA = 7,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_MAX_DYNAMIC_SHARED_SIZE_BYTES = 8,
  __pyx_e_4cupy_4cuda_6driver_CU_FUNC_ATTRIBUTE_PREFERRED_SHARED_MEMORY_CARVEOUT = 9,
  __pyx_e_4cupy_4cuda_6driver_CUDA_ERROR_INVALID_VALUE = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_UNSIGNED_INT8 = 0x01,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_UNSIGNED_INT16 = 0x02,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_UNSIGNED_INT32 = 0x03,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_SIGNED_INT8 = 0x08,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_SIGNED_INT16 = 0x09,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_SIGNED_INT32 = 0x0a,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_HALF = 0x10,
  __pyx_e_4cupy_4cuda_6driver_CU_AD_FORMAT_FLOAT = 0x20,
  __pyx_e_4cupy_4cuda_6driver_CU_TR_ADDRESS_MODE_WRAP = 0,
  __pyx_e_4cupy_4cuda_6driver_CU_TR_ADDRESS_MODE_CLAMP = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_TR_ADDRESS_MODE_MIRROR = 2,
  __pyx_e_4cupy_4cuda_6driver_CU_TR_ADDRESS_MODE_BORDER = 3,
  __pyx_e_4cupy_4cuda_6driver_CU_TR_FILTER_MODE_POINT = 0,
  __pyx_e_4cupy_4cuda_6driver_CU_TR_FILTER_MODE_LINEAR = 1,
  __pyx_e_4cupy_4cuda_6driver_CU_TRSA_OVERRIDE_FORMAT = 0x01,
  __pyx_e_4cupy_4cuda_6driver_CU_TRSF_READ_AS_INTEGER = 0x01,
  __pyx_e_4cupy_4cuda_6driver_CU_TRSF_NORMALIZED_COORDINATES = 0x02,
  __pyx_e_4cupy_4cuda_6driver_CU_TRSF_SRGB = 0x10
};

/* "cupy/cuda/cusolver.pxd":35
 * ###############################################################################
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     CUSOLVER_EIG_TYPE_1 = 1
 *     CUSOLVER_EIG_TYPE_2 = 2
 */
enum  {
  __pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_TYPE_1 = 1,
  __pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_TYPE_2 = 2,
  __pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_TYPE_3 = 3,
  __pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_MODE_NOVECTOR = 0,
  __pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_MODE_VECTOR = 1
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* SetNameInClass.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? _PyDict_SetItem_KnownHash(ns, name, value, ((PyASCIIObject *) name)->hash) : PyObject_SetItem(ns, name, value))
#elif CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_SetNameInClass(ns, name, value)\
    (likely(PyDict_CheckExact(ns)) ? PyDict_SetItem(ns, name, value) : PyObject_SetItem(ns, name, value))
#else
#define __Pyx_SetNameInClass(ns, name, value)  PyObject_SetItem(ns, name, value)
#endif

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionExport.proto */
static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.stdint' */

/* Module declarations from 'cupy.cuda.driver' */

/* Module declarations from 'cython' */

/* Module declarations from 'cupy.cuda' */

/* Module declarations from 'cupy.cuda.stream' */
static intptr_t (*__pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr)(void); /*proto*/

/* Module declarations from 'cupy.cuda.cusolver' */
static PyObject *__pyx_v_4cupy_4cuda_8cusolver_STATUS = 0;
static int __pyx_f_4cupy_4cuda_8cusolver_getProperty(int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver__getVersion(int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_4cupy_4cuda_8cusolver_create(int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_4cupy_4cuda_8cusolver_spCreate(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_destroy(intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spDestroy(intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_setStream(intptr_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusolver_getStream(intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_spotrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dpotrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cpotrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zpotrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spotrf(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dpotrf(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cpotrf(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zpotrf(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spotrs(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dpotrs(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cpotrs(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zpotrs(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgetrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgetrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgetrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgetrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgetrf(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgetrf(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgetrf(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgetrf(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgetrs(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgetrs(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgetrs(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgetrs(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgeqrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgeqrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgeqrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgeqrf_bufferSize(intptr_t, int, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgeqrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgeqrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgeqrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgeqrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sorgqr_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dorgqr_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cungqr_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zungqr_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sorgqr(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dorgqr(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cungqr(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zungqr(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sormqr_bufferSize(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dormqr_bufferSize(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cunmqr_bufferSize(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zunmqr_bufferSize(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sormqr(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dormqr(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cunmqr(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zunmqr(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cormqr(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zormqr(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_ssytrf_bufferSize(intptr_t, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dsytrf_bufferSize(intptr_t, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_csytrf_bufferSize(intptr_t, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zsytrf_bufferSize(intptr_t, int, size_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ssytrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dsytrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_csytrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zsytrf(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgebrd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgebrd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgebrd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgebrd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgebrd(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgebrd(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgebrd(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgebrd(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvd_bufferSize(intptr_t, int, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvd(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvd(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvd(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvd(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static intptr_t __pyx_f_4cupy_4cuda_8cusolver_createGesvdjInfo(int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_destroyGesvdjInfo(intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetTolerance(intptr_t, double, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetMaxSweeps(intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetSortEig(intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static double __pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetResidual(intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetSweeps(intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvdj_bufferSize(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvdj_bufferSize(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvdj_bufferSize(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvdj_bufferSize(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvdj(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvdj(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvdj(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvdj(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched_bufferSize(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched_bufferSize(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched_bufferSize(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched_bufferSize(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_ssyevd_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dsyevd_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cheevd_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zheevd_bufferSize(intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ssyevd(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dsyevd(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cheevd(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zheevd(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_scsrlsvchol(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dcsrlsvchol(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ccsrlsvchol(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zcsrlsvchol(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_scsrlsvqr(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dcsrlsvqr(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ccsrlsvqr(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zcsrlsvqr(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_scsreigvsi(intptr_t, int, int, size_t, size_t, size_t, size_t, float, size_t, int, float, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dcsreigvsi(intptr_t, int, int, size_t, size_t, size_t, size_t, double, size_t, int, double, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ccsreigvsi(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, float, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zcsreigvsi(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, double, size_t, size_t, int __pyx_skip_dispatch); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cupy_4cuda_8cusolver_check_status(int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spSetStream(intptr_t, size_t, int __pyx_skip_dispatch); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusolver_spGetStream(intptr_t, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvdjBatched_bufferSize(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvdjBatched_bufferSize(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvdjBatched_bufferSize(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvdjBatched_bufferSize(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvdjBatched(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvdjBatched(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvdjBatched(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvdjBatched(intptr_t, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch); /*proto*/
#define __Pyx_MODULE_NAME "cupy.cuda.cusolver"
extern int __pyx_module_is_main_cupy__cuda__cusolver;
int __pyx_module_is_main_cupy__cuda__cusolver = 0;

/* Implementation of 'cupy.cuda.cusolver' */
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_super;
static const char __pyx_k_A[] = "A";
static const char __pyx_k_B[] = "B";
static const char __pyx_k_C[] = "C";
static const char __pyx_k_D[] = "D";
static const char __pyx_k_E[] = "E";
static const char __pyx_k_S[] = "S";
static const char __pyx_k_U[] = "U";
static const char __pyx_k_V[] = "V";
static const char __pyx_k_W[] = "W";
static const char __pyx_k_b[] = "b";
static const char __pyx_k_k[] = "k";
static const char __pyx_k_m[] = "m";
static const char __pyx_k_n[] = "n";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_VT[] = "VT";
static const char __pyx_k_mu[] = "mu";
static const char __pyx_k_x0[] = "x0";
static const char __pyx_k_d_A[] = "d_A";
static const char __pyx_k_d_S[] = "d_S";
static const char __pyx_k_d_U[] = "d_U";
static const char __pyx_k_d_V[] = "d_V";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_eps[] = "eps";
static const char __pyx_k_lda[] = "lda";
static const char __pyx_k_ldb[] = "ldb";
static const char __pyx_k_ldc[] = "ldc";
static const char __pyx_k_ldu[] = "ldu";
static const char __pyx_k_ldv[] = "ldv";
static const char __pyx_k_mu0[] = "mu0";
static const char __pyx_k_nnz[] = "nnz";
static const char __pyx_k_tau[] = "tau";
static const char __pyx_k_tol[] = "tol";
static const char __pyx_k_Work[] = "Work";
static const char __pyx_k_econ[] = "econ";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_ipiv[] = "ipiv";
static const char __pyx_k_jobu[] = "jobu";
static const char __pyx_k_jobz[] = "jobz";
static const char __pyx_k_ldvt[] = "ldvt";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_nrhs[] = "nrhs";
static const char __pyx_k_rank[] = "rank";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_side[] = "side";
static const char __pyx_k_tauP[] = "tauP";
static const char __pyx_k_tauQ[] = "tauQ";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_uplo[] = "uplo";
static const char __pyx_k_work[] = "work";
static const char __pyx_k_jobvt[] = "jobvt";
static const char __pyx_k_lwork[] = "lwork";
static const char __pyx_k_rwork[] = "rwork";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_trans[] = "trans";
static const char __pyx_k_csrVal[] = "csrVal";
static const char __pyx_k_d_info[] = "d_info";
static const char __pyx_k_d_work[] = "d_work";
static const char __pyx_k_descrA[] = "descrA";
static const char __pyx_k_handle[] = "handle";
static const char __pyx_k_maxite[] = "maxite";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_params[] = "params";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_status[] = "status";
static const char __pyx_k_stream[] = "stream";
static const char __pyx_k_csrValA[] = "csrValA";
static const char __pyx_k_devInfo[] = "devInfo";
static const char __pyx_k_devIpiv[] = "devIpiv";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_reorder[] = "reorder";
static const char __pyx_k_strideA[] = "strideA";
static const char __pyx_k_strideS[] = "strideS";
static const char __pyx_k_strideU[] = "strideU";
static const char __pyx_k_strideV[] = "strideV";
static const char __pyx_k_h_R_nrmF[] = "h_R_nrmF";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_sort_svd[] = "sort_svd";
static const char __pyx_k_batchSize[] = "batchSize";
static const char __pyx_k_csrColInd[] = "csrColInd";
static const char __pyx_k_csrRowPtr[] = "csrRowPtr";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_tolerance[] = "tolerance";
static const char __pyx_k_csrColIndA[] = "csrColIndA";
static const char __pyx_k_csrRowPtrA[] = "csrRowPtrA";
static const char __pyx_k_max_sweeps[] = "max_sweeps";
static const char __pyx_k_singularity[] = "singularity";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_CUSOLVERError[] = "CUSOLVERError";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_cupy_cuda_cusolver[] = "cupy.cuda.cusolver";
static const char __pyx_k_CUSOLVERError___init[] = "CUSOLVERError.__init__";
static const char __pyx_k_CUSOLVERError___reduce[] = "CUSOLVERError.__reduce__";
static const char __pyx_k_cupy_cuda_cusolver_pyx[] = "cupy/cuda/cusolver.pyx";
static const char __pyx_k_CUSOLVER_STATUS_SUCCESS[] = "CUSOLVER_STATUS_SUCCESS";
static const char __pyx_k_Thin_wrapper_of_CUSOLVER[] = "Thin wrapper of CUSOLVER.";
static const char __pyx_k_CUSOLVER_STATUS_ZERO_PIVOT[] = "CUSOLVER_STATUS_ZERO_PIVOT";
static const char __pyx_k_CUSOLVER_STATUS_ALLOC_FAILED[] = "CUSOLVER_STATUS_ALLOC_FAILED";
static const char __pyx_k_CUSOLVER_STATUS_ARCH_MISMATCH[] = "CUSOLVER_STATUS_ARCH_MISMATCH";
static const char __pyx_k_CUSOLVER_STATUS_INVALID_VALUE[] = "CUSOLVER_STATUS_INVALID_VALUE";
static const char __pyx_k_CUSOLVER_STATUS_MAPPING_ERROR[] = "CUSOLVER_STATUS_MAPPING_ERROR";
static const char __pyx_k_CUSOLVER_STATUS_NOT_SUPPORTED[] = "CUSOLVER_STATUS_NOT_SUPPORTED";
static const char __pyx_k_CUSOLVER_STATUS_INTERNAL_ERROR[] = "CUSOLVER_STATUS_INTERNAL_ERROR";
static const char __pyx_k_CUSOLVER_STATUS_INVALID_LICENSE[] = "CUSOLVER_STATUS_INVALID_LICENSE";
static const char __pyx_k_CUSOLVER_STATUS_MATRIX_TYPE_NOT[] = "CUSOLVER_STATUS_MATRIX_TYPE_NOT_SUPPORTED";
static const char __pyx_k_CUSOLVER_STATUS_NOT_INITIALIZED[] = "CUSOLVER_STATUS_NOT_INITIALIZED";
static const char __pyx_k_CUSOLVER_STATUS_EXECUTION_FAILED[] = "CUSOLVER_STATUS_EXECUTION_FAILED";
static PyObject *__pyx_n_s_A;
static PyObject *__pyx_n_s_B;
static PyObject *__pyx_n_s_C;
static PyObject *__pyx_n_s_CUSOLVERError;
static PyObject *__pyx_n_s_CUSOLVERError___init;
static PyObject *__pyx_n_s_CUSOLVERError___reduce;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_ALLOC_FAILED;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_ARCH_MISMATCH;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_EXECUTION_FAILED;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_INTERNAL_ERROR;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_INVALID_LICENSE;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_INVALID_VALUE;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_MAPPING_ERROR;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_MATRIX_TYPE_NOT;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_NOT_INITIALIZED;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_NOT_SUPPORTED;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_SUCCESS;
static PyObject *__pyx_n_u_CUSOLVER_STATUS_ZERO_PIVOT;
static PyObject *__pyx_n_s_D;
static PyObject *__pyx_n_s_E;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_S;
static PyObject *__pyx_n_s_U;
static PyObject *__pyx_n_s_V;
static PyObject *__pyx_n_s_VT;
static PyObject *__pyx_n_s_W;
static PyObject *__pyx_n_s_Work;
static PyObject *__pyx_n_s_b;
static PyObject *__pyx_n_s_batchSize;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_csrColInd;
static PyObject *__pyx_n_s_csrColIndA;
static PyObject *__pyx_n_s_csrRowPtr;
static PyObject *__pyx_n_s_csrRowPtrA;
static PyObject *__pyx_n_s_csrVal;
static PyObject *__pyx_n_s_csrValA;
static PyObject *__pyx_n_s_cupy_cuda_cusolver;
static PyObject *__pyx_kp_s_cupy_cuda_cusolver_pyx;
static PyObject *__pyx_n_s_d_A;
static PyObject *__pyx_n_s_d_S;
static PyObject *__pyx_n_s_d_U;
static PyObject *__pyx_n_s_d_V;
static PyObject *__pyx_n_s_d_info;
static PyObject *__pyx_n_s_d_work;
static PyObject *__pyx_n_s_descrA;
static PyObject *__pyx_n_s_devInfo;
static PyObject *__pyx_n_s_devIpiv;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_econ;
static PyObject *__pyx_n_s_eps;
static PyObject *__pyx_n_s_h_R_nrmF;
static PyObject *__pyx_n_s_handle;
static PyObject *__pyx_n_s_info;
static PyObject *__pyx_n_s_init;
static PyObject *__pyx_n_s_ipiv;
static PyObject *__pyx_n_s_jobu;
static PyObject *__pyx_n_s_jobvt;
static PyObject *__pyx_n_s_jobz;
static PyObject *__pyx_n_s_k;
static PyObject *__pyx_n_s_lda;
static PyObject *__pyx_n_s_ldb;
static PyObject *__pyx_n_s_ldc;
static PyObject *__pyx_n_s_ldu;
static PyObject *__pyx_n_s_ldv;
static PyObject *__pyx_n_s_ldvt;
static PyObject *__pyx_n_s_lwork;
static PyObject *__pyx_n_s_m;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_max_sweeps;
static PyObject *__pyx_n_s_maxite;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_mu;
static PyObject *__pyx_n_s_mu0;
static PyObject *__pyx_n_s_n;
static PyObject *__pyx_n_s_nnz;
static PyObject *__pyx_n_s_nrhs;
static PyObject *__pyx_n_s_params;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_rank;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reorder;
static PyObject *__pyx_n_s_rwork;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_side;
static PyObject *__pyx_n_s_singularity;
static PyObject *__pyx_n_s_sort_svd;
static PyObject *__pyx_n_s_status;
static PyObject *__pyx_n_s_stream;
static PyObject *__pyx_n_s_strideA;
static PyObject *__pyx_n_s_strideS;
static PyObject *__pyx_n_s_strideU;
static PyObject *__pyx_n_s_strideV;
static PyObject *__pyx_n_s_super;
static PyObject *__pyx_n_s_tau;
static PyObject *__pyx_n_s_tauP;
static PyObject *__pyx_n_s_tauQ;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_tol;
static PyObject *__pyx_n_s_tolerance;
static PyObject *__pyx_n_s_trans;
static PyObject *__pyx_n_s_uplo;
static PyObject *__pyx_n_s_work;
static PyObject *__pyx_n_s_x;
static PyObject *__pyx_n_s_x0;
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_13CUSOLVERError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_status); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_13CUSOLVERError_2__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_2getProperty(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_type); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_4_getVersion(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_6create(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_8spCreate(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_10destroy(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_12spDestroy(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_14setStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, size_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_16getStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_18spSetStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, size_t __pyx_v_stream); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_20spGetStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_22spotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_24dpotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_26cpotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_28zpotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_30spotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_32dpotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_34cpotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_36zpotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_38spotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_40dpotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_42cpotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_44zpotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_46sgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_48dgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_50cgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_52zgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_54sgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_56dgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_58cgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_60zgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_62sgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_64dgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_66cgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_68zgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_70sgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_72dgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_74cgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_76zgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_78sgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_80dgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_82cgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_84zgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_86sorgqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_88dorgqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_90cungqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_92zungqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_94sorgqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_96dorgqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_98cungqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_100zungqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_102sormqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_104dormqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_106cunmqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_108zunmqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_110sormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_112dormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_114cunmqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_116zunmqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_118cormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_120zormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_122ssytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_124dsytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_126csytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_128zsytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_130ssytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_132dsytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_134csytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_136zsytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_138sgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_140dgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_142cgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_144zgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_146sgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_148dgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_150cgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_152zgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_154sgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_156dgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_158cgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_160zgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_162sgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_164dgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_166cgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_168zgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_170createGesvdjInfo(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_172destroyGesvdjInfo(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_174xgesvdjSetTolerance(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info, double __pyx_v_tolerance); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_176xgesvdjSetMaxSweeps(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info, int __pyx_v_max_sweeps); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_178xgesvdjSetSortEig(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info, int __pyx_v_sort_svd); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_180xgesvdjGetResidual(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, intptr_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_182xgesvdjGetSweeps(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, intptr_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_184sgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_186dgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_188cgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_190zgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_192sgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_194dgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_196cgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_198zgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_200sgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_202dgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_204cgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_206zgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_208sgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_210dgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_212cgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_214zgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_216sgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_218dgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_220cgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_222zgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_224sgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_226dgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_228cgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_230zgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_232ssyevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_234dsyevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_236cheevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_238zheevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_240ssyevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_242dsyevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_244cheevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_246zheevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_248scsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_250dcsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_252ccsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_254zcsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_256scsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_258dcsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_260ccsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_262zcsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_264scsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, float __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, float __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_266dcsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, double __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, double __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_268ccsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, float __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x); /* proto */
static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_270zcsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, double __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_11;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_codeobj__3;
static PyObject *__pyx_codeobj__5;
/* Late includes */

/* "cupy/cuda/cusolver.pyx":575
 * class CUSOLVERError(RuntimeError):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 */

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13CUSOLVERError_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_13CUSOLVERError___init__[] = "CUSOLVERError.__init__(self, status)";
static PyMethodDef __pyx_mdef_4cupy_4cuda_8cusolver_13CUSOLVERError_1__init__ = {"__init__", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_13CUSOLVERError_1__init__, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_13CUSOLVERError___init__};
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13CUSOLVERError_1__init__(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_self = 0;
  PyObject *__pyx_v_status = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_status,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_status)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, 1); __PYX_ERR(0, 575, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 575, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_status = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 575, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.CUSOLVERError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_13CUSOLVERError___init__(__pyx_self, __pyx_v_self, __pyx_v_status);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_13CUSOLVERError___init__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self, PyObject *__pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "cupy/cuda/cusolver.pyx":576
 * 
 *     def __init__(self, status):
 *         self.status = status             # <<<<<<<<<<<<<<
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self, __pyx_n_s_status, __pyx_v_status) < 0) __PYX_ERR(0, 576, __pyx_L1_error)

  /* "cupy/cuda/cusolver.pyx":577
 *     def __init__(self, status):
 *         self.status = status
 *         super(CUSOLVERError, self).__init__(STATUS[status])             # <<<<<<<<<<<<<<
 * 
 *     def __reduce__(self):
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_CUSOLVERError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2);
  __Pyx_INCREF(__pyx_v_self);
  __Pyx_GIVEREF(__pyx_v_self);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(__pyx_v_4cupy_4cuda_8cusolver_STATUS == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 577, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_4cupy_4cuda_8cusolver_STATUS, __pyx_v_status); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  if (!__pyx_t_4) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[2] = {__pyx_t_4, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 577, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":575
 * class CUSOLVERError(RuntimeError):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cupy.cuda.cusolver.CUSOLVERError.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":579
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13CUSOLVERError_3__reduce__(PyObject *__pyx_self, PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_13CUSOLVERError_2__reduce__[] = "CUSOLVERError.__reduce__(self)";
static PyMethodDef __pyx_mdef_4cupy_4cuda_8cusolver_13CUSOLVERError_3__reduce__ = {"__reduce__", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_13CUSOLVERError_3__reduce__, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_13CUSOLVERError_2__reduce__};
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13CUSOLVERError_3__reduce__(PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce__ (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_13CUSOLVERError_2__reduce__(__pyx_self, ((PyObject *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_13CUSOLVERError_2__reduce__(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("__reduce__", 0);

  /* "cupy/cuda/cusolver.pyx":580
 * 
 *     def __reduce__(self):
 *         return (type(self), (self.status,))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self, __pyx_n_s_status); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(((PyObject *)Py_TYPE(__pyx_v_self)));
  PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(__pyx_v_self)));
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":579
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.CUSOLVERError.__reduce__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":584
 * 
 * @cython.profile(False)
 * cpdef inline check_status(int status):             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         raise CUSOLVERError(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status); /*proto*/
static CYTHON_INLINE PyObject *__pyx_f_4cupy_4cuda_8cusolver_check_status(int __pyx_v_status, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("check_status", 0);

  /* "cupy/cuda/cusolver.pyx":585
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise CUSOLVERError(status)
 * 
 */
  __pyx_t_1 = ((__pyx_v_status != 0) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "cupy/cuda/cusolver.pyx":586
 * cpdef inline check_status(int status):
 *     if status != 0:
 *         raise CUSOLVERError(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_CUSOLVERError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_status); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_t_4};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 586, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 586, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 586, __pyx_L1_error)

    /* "cupy/cuda/cusolver.pyx":585
 * @cython.profile(False)
 * cpdef inline check_status(int status):
 *     if status != 0:             # <<<<<<<<<<<<<<
 *         raise CUSOLVERError(status)
 * 
 */
  }

  /* "cupy/cuda/cusolver.pyx":584
 * 
 * @cython.profile(False)
 * cpdef inline check_status(int status):             # <<<<<<<<<<<<<<
 *     if status != 0:
 *         raise CUSOLVERError(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("cupy.cuda.cusolver.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_check_status[] = "check_status(int status)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_1check_status(PyObject *__pyx_self, PyObject *__pyx_arg_status) {
  int __pyx_v_status;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_status (wrapper)", 0);
  assert(__pyx_arg_status); {
    __pyx_v_status = __Pyx_PyInt_As_int(__pyx_arg_status); if (unlikely((__pyx_v_status == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 584, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_check_status(__pyx_self, ((int)__pyx_v_status));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_check_status(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_status) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("check_status", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.check_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":593
 * ###############################################################################
 * 
 * cpdef int getProperty(int type):             # <<<<<<<<<<<<<<
 *     cdef int value
 *     with nogil:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_3getProperty(PyObject *__pyx_self, PyObject *__pyx_arg_type); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_getProperty(int __pyx_v_type, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_value;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getProperty", 0);

  /* "cupy/cuda/cusolver.pyx":595
 * cpdef int getProperty(int type):
 *     cdef int value
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverGetProperty(<LibraryPropertyType>type, &value)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":596
 *     cdef int value
 *     with nogil:
 *         status = cusolverGetProperty(<LibraryPropertyType>type, &value)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return value
 */
        __pyx_v_status = cusolverGetProperty(((libraryPropertyType_t)__pyx_v_type), (&__pyx_v_value));
      }

      /* "cupy/cuda/cusolver.pyx":595
 * cpdef int getProperty(int type):
 *     cdef int value
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverGetProperty(<LibraryPropertyType>type, &value)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":597
 *     with nogil:
 *         status = cusolverGetProperty(<LibraryPropertyType>type, &value)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return value
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":598
 *         status = cusolverGetProperty(<LibraryPropertyType>type, &value)
 *     check_status(status)
 *     return value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_value;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":593
 * ###############################################################################
 * 
 * cpdef int getProperty(int type):             # <<<<<<<<<<<<<<
 *     cdef int value
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.getProperty", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_3getProperty(PyObject *__pyx_self, PyObject *__pyx_arg_type); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_2getProperty[] = "getProperty(int type) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_3getProperty(PyObject *__pyx_self, PyObject *__pyx_arg_type) {
  int __pyx_v_type;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getProperty (wrapper)", 0);
  assert(__pyx_arg_type); {
    __pyx_v_type = __Pyx_PyInt_As_int(__pyx_arg_type); if (unlikely((__pyx_v_type == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 593, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.getProperty", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_2getProperty(__pyx_self, ((int)__pyx_v_type));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_2getProperty(CYTHON_UNUSED PyObject *__pyx_self, int __pyx_v_type) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getProperty", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_getProperty(__pyx_v_type, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.getProperty", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":601
 * 
 * 
 * cpdef tuple _getVersion():             # <<<<<<<<<<<<<<
 *     return (getProperty(MAJOR_VERSION),
 *             getProperty(MINOR_VERSION),
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_5_getVersion(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver__getVersion(CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_getVersion", 0);

  /* "cupy/cuda/cusolver.pyx":602
 * 
 * cpdef tuple _getVersion():
 *     return (getProperty(MAJOR_VERSION),             # <<<<<<<<<<<<<<
 *             getProperty(MINOR_VERSION),
 *             getProperty(PATCH_LEVEL))
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_getProperty(MAJOR_VERSION, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "cupy/cuda/cusolver.pyx":603
 * cpdef tuple _getVersion():
 *     return (getProperty(MAJOR_VERSION),
 *             getProperty(MINOR_VERSION),             # <<<<<<<<<<<<<<
 *             getProperty(PATCH_LEVEL))
 * 
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_getProperty(MINOR_VERSION, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 603, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "cupy/cuda/cusolver.pyx":604
 *     return (getProperty(MAJOR_VERSION),
 *             getProperty(MINOR_VERSION),
 *             getProperty(PATCH_LEVEL))             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_getProperty(PATCH_LEVEL, 0)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "cupy/cuda/cusolver.pyx":602
 * 
 * cpdef tuple _getVersion():
 *     return (getProperty(MAJOR_VERSION),             # <<<<<<<<<<<<<<
 *             getProperty(MINOR_VERSION),
 *             getProperty(PATCH_LEVEL))
 */
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 602, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":601
 * 
 * 
 * cpdef tuple _getVersion():             # <<<<<<<<<<<<<<
 *     return (getProperty(MAJOR_VERSION),
 *             getProperty(MINOR_VERSION),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("cupy.cuda.cusolver._getVersion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_5_getVersion(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_4_getVersion[] = "_getVersion() -> tuple";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_5_getVersion(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_getVersion (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_4_getVersion(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_4_getVersion(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_getVersion", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver__getVersion(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver._getVersion", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":611
 * ###############################################################################
 * 
 * cpdef intptr_t create() except? 0:             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     with nogil:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_7create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static intptr_t __pyx_f_4cupy_4cuda_8cusolver_create(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusolverDnHandle_t __pyx_v_handle;
  int __pyx_v_status;
  intptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("create", 0);

  /* "cupy/cuda/cusolver.pyx":613
 * cpdef intptr_t create() except? 0:
 *     cdef Handle handle
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCreate(&handle)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":614
 *     cdef Handle handle
 *     with nogil:
 *         status = cusolverDnCreate(&handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <intptr_t>handle
 */
        __pyx_v_status = cusolverDnCreate((&__pyx_v_handle));
      }

      /* "cupy/cuda/cusolver.pyx":613
 * cpdef intptr_t create() except? 0:
 *     cdef Handle handle
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCreate(&handle)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":615
 *     with nogil:
 *         status = cusolverDnCreate(&handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <intptr_t>handle
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":616
 *         status = cusolverDnCreate(&handle)
 *     check_status(status)
 *     return <intptr_t>handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((intptr_t)__pyx_v_handle);
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":611
 * ###############################################################################
 * 
 * cpdef intptr_t create() except? 0:             # <<<<<<<<<<<<<<
 *     cdef Handle handle
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_7create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_6create[] = "create() -> intptr_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_7create(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_6create(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_6create(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("create", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_create(0); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 611, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.create", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":619
 * 
 * 
 * cpdef intptr_t spCreate() except? 0:             # <<<<<<<<<<<<<<
 *     cdef SpHandle handle
 *     with nogil:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_9spCreate(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static intptr_t __pyx_f_4cupy_4cuda_8cusolver_spCreate(CYTHON_UNUSED int __pyx_skip_dispatch) {
  cusolverSpHandle_t __pyx_v_handle;
  int __pyx_v_status;
  intptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spCreate", 0);

  /* "cupy/cuda/cusolver.pyx":621
 * cpdef intptr_t spCreate() except? 0:
 *     cdef SpHandle handle
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCreate(&handle)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":622
 *     cdef SpHandle handle
 *     with nogil:
 *         status = cusolverSpCreate(&handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <intptr_t>handle
 */
        __pyx_v_status = cusolverSpCreate((&__pyx_v_handle));
      }

      /* "cupy/cuda/cusolver.pyx":621
 * cpdef intptr_t spCreate() except? 0:
 *     cdef SpHandle handle
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCreate(&handle)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":623
 *     with nogil:
 *         status = cusolverSpCreate(&handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <intptr_t>handle
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":624
 *         status = cusolverSpCreate(&handle)
 *     check_status(status)
 *     return <intptr_t>handle             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((intptr_t)__pyx_v_handle);
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":619
 * 
 * 
 * cpdef intptr_t spCreate() except? 0:             # <<<<<<<<<<<<<<
 *     cdef SpHandle handle
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spCreate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_9spCreate(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_8spCreate[] = "spCreate() -> intptr_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_9spCreate(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spCreate (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_8spCreate(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_8spCreate(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("spCreate", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spCreate(0); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 619, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spCreate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":627
 * 
 * 
 * cpdef destroy(intptr_t handle):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDestroy(<Handle>handle)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_11destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_destroy(intptr_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroy", 0);

  /* "cupy/cuda/cusolver.pyx":628
 * 
 * cpdef destroy(intptr_t handle):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDestroy(<Handle>handle)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":629
 * cpdef destroy(intptr_t handle):
 *     with nogil:
 *         status = cusolverDnDestroy(<Handle>handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
        __pyx_v_status = cusolverDnDestroy(((cusolverDnHandle_t)__pyx_v_handle));
      }

      /* "cupy/cuda/cusolver.pyx":628
 * 
 * cpdef destroy(intptr_t handle):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDestroy(<Handle>handle)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":630
 *     with nogil:
 *         status = cusolverDnDestroy(<Handle>handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 630, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":627
 * 
 * 
 * cpdef destroy(intptr_t handle):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDestroy(<Handle>handle)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_11destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_10destroy[] = "destroy(intptr_t handle)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_11destroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  intptr_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroy (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = PyInt_AsSsize_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 627, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_10destroy(__pyx_self, ((intptr_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_10destroy(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_destroy(__pyx_v_handle, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.destroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":633
 * 
 * 
 * cpdef spDestroy(intptr_t handle):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpDestroy(<SpHandle>handle)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13spDestroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spDestroy(intptr_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spDestroy", 0);

  /* "cupy/cuda/cusolver.pyx":634
 * 
 * cpdef spDestroy(intptr_t handle):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDestroy(<SpHandle>handle)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":635
 * cpdef spDestroy(intptr_t handle):
 *     with nogil:
 *         status = cusolverSpDestroy(<SpHandle>handle)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
        __pyx_v_status = cusolverSpDestroy(((cusolverSpHandle_t)__pyx_v_handle));
      }

      /* "cupy/cuda/cusolver.pyx":634
 * 
 * cpdef spDestroy(intptr_t handle):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDestroy(<SpHandle>handle)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":636
 *     with nogil:
 *         status = cusolverSpDestroy(<SpHandle>handle)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":633
 * 
 * 
 * cpdef spDestroy(intptr_t handle):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpDestroy(<SpHandle>handle)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spDestroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13spDestroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_12spDestroy[] = "spDestroy(intptr_t handle)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_13spDestroy(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  intptr_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spDestroy (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = PyInt_AsSsize_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 633, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.spDestroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_12spDestroy(__pyx_self, ((intptr_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_12spDestroy(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spDestroy", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spDestroy(__pyx_v_handle, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spDestroy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":643
 * ###############################################################################
 * 
 * cpdef setStream(intptr_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSetStream(<Handle>handle, <driver.Stream>stream)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_15setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_setStream(intptr_t __pyx_v_handle, size_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setStream", 0);

  /* "cupy/cuda/cusolver.pyx":644
 * 
 * cpdef setStream(intptr_t handle, size_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":645
 * cpdef setStream(intptr_t handle, size_t stream):
 *     with nogil:
 *         status = cusolverDnSetStream(<Handle>handle, <driver.Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
        __pyx_v_status = cusolverDnSetStream(((cusolverDnHandle_t)__pyx_v_handle), ((cudaStream_t)__pyx_v_stream));
      }

      /* "cupy/cuda/cusolver.pyx":644
 * 
 * cpdef setStream(intptr_t handle, size_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":646
 *     with nogil:
 *         status = cusolverDnSetStream(<Handle>handle, <driver.Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":643
 * ###############################################################################
 * 
 * cpdef setStream(intptr_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSetStream(<Handle>handle, <driver.Stream>stream)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_15setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_14setStream[] = "setStream(intptr_t handle, size_t stream)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_15setStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  size_t __pyx_v_stream;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("setStream (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("setStream", 1, 2, 2, 1); __PYX_ERR(0, 643, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "setStream") < 0)) __PYX_ERR(0, 643, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 643, __pyx_L3_error)
    __pyx_v_stream = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stream == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 643, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("setStream", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 643, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_14setStream(__pyx_self, __pyx_v_handle, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_14setStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, size_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("setStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_v_stream, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.setStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":649
 * 
 * 
 * cpdef size_t getStream(intptr_t handle) except? 0:             # <<<<<<<<<<<<<<
 *     cdef driver.Stream stream
 *     with nogil:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_17getStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusolver_getStream(intptr_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudaStream_t __pyx_v_stream;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("getStream", 0);

  /* "cupy/cuda/cusolver.pyx":651
 * cpdef size_t getStream(intptr_t handle) except? 0:
 *     cdef driver.Stream stream
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnGetStream(<Handle>handle, &stream)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":652
 *     cdef driver.Stream stream
 *     with nogil:
 *         status = cusolverDnGetStream(<Handle>handle, &stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>stream
 */
        __pyx_v_status = cusolverDnGetStream(((cusolverDnHandle_t)__pyx_v_handle), (&__pyx_v_stream));
      }

      /* "cupy/cuda/cusolver.pyx":651
 * cpdef size_t getStream(intptr_t handle) except? 0:
 *     cdef driver.Stream stream
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnGetStream(<Handle>handle, &stream)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":653
 *     with nogil:
 *         status = cusolverDnGetStream(<Handle>handle, &stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>stream
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":654
 *         status = cusolverDnGetStream(<Handle>handle, &stream)
 *     check_status(status)
 *     return <size_t>stream             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_stream);
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":649
 * 
 * 
 * cpdef size_t getStream(intptr_t handle) except? 0:             # <<<<<<<<<<<<<<
 *     cdef driver.Stream stream
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.getStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_17getStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_16getStream[] = "getStream(intptr_t handle) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_17getStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  intptr_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("getStream (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = PyInt_AsSsize_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.getStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_16getStream(__pyx_self, ((intptr_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_16getStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("getStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_getStream(__pyx_v_handle, 0); if (unlikely(__pyx_t_1 == ((size_t)0) && PyErr_Occurred())) __PYX_ERR(0, 649, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.getStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":657
 * 
 * 
 * cpdef spSetStream(intptr_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpSetStream(<SpHandle>handle, <driver.Stream>stream)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_19spSetStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spSetStream(intptr_t __pyx_v_handle, size_t __pyx_v_stream, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spSetStream", 0);

  /* "cupy/cuda/cusolver.pyx":658
 * 
 * cpdef spSetStream(intptr_t handle, size_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpSetStream(<SpHandle>handle, <driver.Stream>stream)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":659
 * cpdef spSetStream(intptr_t handle, size_t stream):
 *     with nogil:
 *         status = cusolverSpSetStream(<SpHandle>handle, <driver.Stream>stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
        __pyx_v_status = cusolverSpSetStream(((cusolverSpHandle_t)__pyx_v_handle), ((cudaStream_t)__pyx_v_stream));
      }

      /* "cupy/cuda/cusolver.pyx":658
 * 
 * cpdef spSetStream(intptr_t handle, size_t stream):
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpSetStream(<SpHandle>handle, <driver.Stream>stream)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":660
 *     with nogil:
 *         status = cusolverSpSetStream(<SpHandle>handle, <driver.Stream>stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":657
 * 
 * 
 * cpdef spSetStream(intptr_t handle, size_t stream):             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpSetStream(<SpHandle>handle, <driver.Stream>stream)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spSetStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_19spSetStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_18spSetStream[] = "spSetStream(intptr_t handle, size_t stream)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_19spSetStream(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  size_t __pyx_v_stream;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spSetStream (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_stream,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stream)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spSetStream", 1, 2, 2, 1); __PYX_ERR(0, 657, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spSetStream") < 0)) __PYX_ERR(0, 657, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L3_error)
    __pyx_v_stream = __Pyx_PyInt_As_size_t(values[1]); if (unlikely((__pyx_v_stream == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 657, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spSetStream", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 657, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.spSetStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_18spSetStream(__pyx_self, __pyx_v_handle, __pyx_v_stream);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_18spSetStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, size_t __pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spSetStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_v_stream, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 657, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spSetStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":663
 * 
 * 
 * cpdef size_t spGetStream(intptr_t handle) except *:             # <<<<<<<<<<<<<<
 *     cdef driver.Stream stream
 *     with nogil:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_21spGetStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static size_t __pyx_f_4cupy_4cuda_8cusolver_spGetStream(intptr_t __pyx_v_handle, CYTHON_UNUSED int __pyx_skip_dispatch) {
  cudaStream_t __pyx_v_stream;
  int __pyx_v_status;
  size_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spGetStream", 0);

  /* "cupy/cuda/cusolver.pyx":665
 * cpdef size_t spGetStream(intptr_t handle) except *:
 *     cdef driver.Stream stream
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpGetStream(<SpHandle>handle, &stream)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":666
 *     cdef driver.Stream stream
 *     with nogil:
 *         status = cusolverSpGetStream(<SpHandle>handle, &stream)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <size_t>stream
 */
        __pyx_v_status = cusolverSpGetStream(((cusolverSpHandle_t)__pyx_v_handle), (&__pyx_v_stream));
      }

      /* "cupy/cuda/cusolver.pyx":665
 * cpdef size_t spGetStream(intptr_t handle) except *:
 *     cdef driver.Stream stream
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpGetStream(<SpHandle>handle, &stream)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":667
 *     with nogil:
 *         status = cusolverSpGetStream(<SpHandle>handle, &stream)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <size_t>stream
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 667, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":668
 *         status = cusolverSpGetStream(<SpHandle>handle, &stream)
 *     check_status(status)
 *     return <size_t>stream             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((size_t)__pyx_v_stream);
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":663
 * 
 * 
 * cpdef size_t spGetStream(intptr_t handle) except *:             # <<<<<<<<<<<<<<
 *     cdef driver.Stream stream
 *     with nogil:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spGetStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_21spGetStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_20spGetStream[] = "spGetStream(intptr_t handle) -> size_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_21spGetStream(PyObject *__pyx_self, PyObject *__pyx_arg_handle) {
  intptr_t __pyx_v_handle;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spGetStream (wrapper)", 0);
  assert(__pyx_arg_handle); {
    __pyx_v_handle = PyInt_AsSsize_t(__pyx_arg_handle); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.spGetStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_20spGetStream(__pyx_self, ((intptr_t)__pyx_v_handle));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_20spGetStream(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("spGetStream", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spGetStream(__pyx_v_handle, 0); if (unlikely(__pyx_t_1 == ((size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 663, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_FromSize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 663, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spGetStream", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":676
 * 
 * # Cholesky factorization
 * cpdef int spotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_23spotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_spotrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spotrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":679
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSpotrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":680
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":681
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSpotrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n,
 *             <float*>A, <int>lda, &lwork)
 */
        __pyx_v_status = cusolverDnSpotrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((float *)__pyx_v_A), ((int)__pyx_v_lda), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":680
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":684
 *             <Handle>handle, <FillMode>uplo, n,
 *             <float*>A, <int>lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":685
 *             <float*>A, <int>lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dpotrf_bufferSize(intptr_t handle, int uplo,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":676
 * 
 * # Cholesky factorization
 * cpdef int spotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_23spotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_22spotrf_bufferSize[] = "spotrf_bufferSize(intptr_t handle, int uplo, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_23spotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spotrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 676, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 676, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 676, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 676, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spotrf_bufferSize") < 0)) __PYX_ERR(0, 676, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 677, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spotrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 676, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_22spotrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_22spotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("spotrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spotrf_bufferSize(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 676, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":687
 *     return lwork
 * 
 * cpdef int dpotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_25dpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dpotrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dpotrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":690
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDpotrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":691
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":692
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDpotrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n,
 *             <double*>A, <int>lda, &lwork)
 */
        __pyx_v_status = cusolverDnDpotrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((double *)__pyx_v_A), ((int)__pyx_v_lda), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":691
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":695
 *             <Handle>handle, <FillMode>uplo, n,
 *             <double*>A, <int>lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":696
 *             <double*>A, <int>lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cpotrf_bufferSize(intptr_t handle, int uplo,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":687
 *     return lwork
 * 
 * cpdef int dpotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_25dpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_24dpotrf_bufferSize[] = "dpotrf_bufferSize(intptr_t handle, int uplo, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_25dpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dpotrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 687, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 687, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 687, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 687, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dpotrf_bufferSize") < 0)) __PYX_ERR(0, 687, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 688, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 688, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 688, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dpotrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 687, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_24dpotrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_24dpotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dpotrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dpotrf_bufferSize(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 687, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":698
 *     return lwork
 * 
 * cpdef int cpotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_27cpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cpotrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cpotrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":701
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCpotrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":702
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":703
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCpotrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n,
 *             <cuComplex*>A, <int>lda, &lwork)
 */
        __pyx_v_status = cusolverDnCpotrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuComplex *)__pyx_v_A), ((int)__pyx_v_lda), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":702
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":706
 *             <Handle>handle, <FillMode>uplo, n,
 *             <cuComplex*>A, <int>lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":707
 *             <cuComplex*>A, <int>lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zpotrf_bufferSize(intptr_t handle, int uplo,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":698
 *     return lwork
 * 
 * cpdef int cpotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_27cpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_26cpotrf_bufferSize[] = "cpotrf_bufferSize(intptr_t handle, int uplo, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_27cpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cpotrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 698, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 698, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 698, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 698, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cpotrf_bufferSize") < 0)) __PYX_ERR(0, 698, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 698, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 698, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 699, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 699, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 699, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cpotrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 698, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_26cpotrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_26cpotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cpotrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cpotrf_bufferSize(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 698, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":709
 *     return lwork
 * 
 * cpdef int zpotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_29zpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zpotrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zpotrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":712
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZpotrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":713
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":714
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZpotrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n,
 *             <cuDoubleComplex*>A, <int>lda, &lwork)
 */
        __pyx_v_status = cusolverDnZpotrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), ((int)__pyx_v_lda), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":713
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZpotrf_bufferSize(
 *             <Handle>handle, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":717
 *             <Handle>handle, <FillMode>uplo, n,
 *             <cuDoubleComplex*>A, <int>lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 717, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":718
 *             <cuDoubleComplex*>A, <int>lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef spotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":709
 *     return lwork
 * 
 * cpdef int zpotrf_bufferSize(intptr_t handle, int uplo,             # <<<<<<<<<<<<<<
 *                             int n, size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_29zpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_28zpotrf_bufferSize[] = "zpotrf_bufferSize(intptr_t handle, int uplo, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_29zpotrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zpotrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 709, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 709, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 709, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 709, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zpotrf_bufferSize") < 0)) __PYX_ERR(0, 709, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 709, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 709, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 710, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zpotrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 709, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_28zpotrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_28zpotrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zpotrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zpotrf_bufferSize(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 709, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":720
 *     return lwork
 * 
 * cpdef spotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_31spotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spotrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spotrf", 0);

  /* "cupy/cuda/cusolver.pyx":722
 * cpdef spotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSpotrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":723
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <float*>A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":724
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSpotrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <float*>A,
 *             lda, <float*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnSpotrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":723
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <float*>A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":727
 *             <Handle>handle, <FillMode>uplo, n, <float*>A,
 *             lda, <float*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":720
 *     return lwork
 * 
 * cpdef spotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_31spotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_30spotrf[] = "spotrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_31spotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spotrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 1); __PYX_ERR(0, 720, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 2); __PYX_ERR(0, 720, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 3); __PYX_ERR(0, 720, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 4); __PYX_ERR(0, 720, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 5); __PYX_ERR(0, 720, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 6); __PYX_ERR(0, 720, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, 7); __PYX_ERR(0, 720, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spotrf") < 0)) __PYX_ERR(0, 720, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 720, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 721, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spotrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 720, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_30spotrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_30spotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spotrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spotrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":729
 *     check_status(status)
 * 
 * cpdef dpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_33dpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dpotrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dpotrf", 0);

  /* "cupy/cuda/cusolver.pyx":731
 * cpdef dpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDpotrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":732
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <double*>A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":733
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDpotrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <double*>A,
 *             lda, <double*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnDpotrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":732
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <double*>A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":736
 *             <Handle>handle, <FillMode>uplo, n, <double*>A,
 *             lda, <double*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":729
 *     check_status(status)
 * 
 * cpdef dpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_33dpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_32dpotrf[] = "dpotrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_33dpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dpotrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 1); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 2); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 3); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 4); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 5); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 6); __PYX_ERR(0, 729, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, 7); __PYX_ERR(0, 729, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dpotrf") < 0)) __PYX_ERR(0, 729, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 729, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 730, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dpotrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 729, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_32dpotrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_32dpotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dpotrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dpotrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":738
 *     check_status(status)
 * 
 * cpdef cpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_35cpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cpotrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cpotrf", 0);

  /* "cupy/cuda/cusolver.pyx":740
 * cpdef cpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCpotrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":741
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":742
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCpotrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A,
 *             lda, <cuComplex*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnCpotrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":741
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":745
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A,
 *             lda, <cuComplex*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":738
 *     check_status(status)
 * 
 * cpdef cpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_35cpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_34cpotrf[] = "cpotrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_35cpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cpotrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 1); __PYX_ERR(0, 738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 2); __PYX_ERR(0, 738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 3); __PYX_ERR(0, 738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 4); __PYX_ERR(0, 738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 5); __PYX_ERR(0, 738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 6); __PYX_ERR(0, 738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, 7); __PYX_ERR(0, 738, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cpotrf") < 0)) __PYX_ERR(0, 738, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 738, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 739, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cpotrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 738, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_34cpotrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_34cpotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cpotrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cpotrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":747
 *     check_status(status)
 * 
 * cpdef zpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_37zpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zpotrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zpotrf", 0);

  /* "cupy/cuda/cusolver.pyx":749
 * cpdef zpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZpotrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":750
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":751
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZpotrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A,
 *             lda, <cuDoubleComplex*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnZpotrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":750
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZpotrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":754
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A,
 *             lda, <cuDoubleComplex*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef spotrs(intptr_t handle, int uplo, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":747
 *     check_status(status)
 * 
 * cpdef zpotrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_37zpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_36zpotrf[] = "zpotrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_37zpotrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zpotrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 1); __PYX_ERR(0, 747, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 2); __PYX_ERR(0, 747, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 3); __PYX_ERR(0, 747, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 4); __PYX_ERR(0, 747, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 5); __PYX_ERR(0, 747, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 6); __PYX_ERR(0, 747, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, 7); __PYX_ERR(0, 747, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zpotrf") < 0)) __PYX_ERR(0, 747, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 747, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 748, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 748, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 748, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zpotrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 747, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_36zpotrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_36zpotrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zpotrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zpotrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":756
 *     check_status(status)
 * 
 * cpdef spotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_39spotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_spotrs(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spotrs", 0);

  /* "cupy/cuda/cusolver.pyx":758
 * cpdef spotrs(intptr_t handle, int uplo, int n, int nrhs,
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSpotrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":759
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":760
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSpotrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 *             <const float*>A, lda, <float*>B, ldb,
 */
        __pyx_v_status = cusolverDnSpotrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, __pyx_v_nrhs, ((float const *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":759
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":764
 *             <const float*>A, lda, <float*>B, ldb,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dpotrs(intptr_t handle, int uplo, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 764, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":756
 *     check_status(status)
 * 
 * cpdef spotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_39spotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_38spotrs[] = "spotrs(intptr_t handle, int uplo, int n, int nrhs, size_t A, int lda, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_39spotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("spotrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 1); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 2); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 3); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 4); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 5); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 6); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 7); __PYX_ERR(0, 756, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, 8); __PYX_ERR(0, 756, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "spotrs") < 0)) __PYX_ERR(0, 756, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 756, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 757, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("spotrs", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 756, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_38spotrs(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_38spotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("spotrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spotrs(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 756, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.spotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":766
 *     check_status(status)
 * 
 * cpdef dpotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_41dpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dpotrs(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dpotrs", 0);

  /* "cupy/cuda/cusolver.pyx":768
 * cpdef dpotrs(intptr_t handle, int uplo, int n, int nrhs,
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDpotrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":769
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":770
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDpotrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 *             <const double*>A, lda, <double*>B, ldb,
 */
        __pyx_v_status = cusolverDnDpotrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, __pyx_v_nrhs, ((double const *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":769
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":774
 *             <const double*>A, lda, <double*>B, ldb,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cpotrs(intptr_t handle, int uplo, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":766
 *     check_status(status)
 * 
 * cpdef dpotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_41dpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_40dpotrs[] = "dpotrs(intptr_t handle, int uplo, int n, int nrhs, size_t A, int lda, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_41dpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dpotrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 1); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 2); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 3); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 4); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 5); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 6); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 7); __PYX_ERR(0, 766, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, 8); __PYX_ERR(0, 766, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dpotrs") < 0)) __PYX_ERR(0, 766, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 766, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 767, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dpotrs", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 766, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_40dpotrs(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_40dpotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dpotrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dpotrs(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":776
 *     check_status(status)
 * 
 * cpdef cpotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_43cpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cpotrs(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cpotrs", 0);

  /* "cupy/cuda/cusolver.pyx":778
 * cpdef cpotrs(intptr_t handle, int uplo, int n, int nrhs,
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCpotrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":779
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":780
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCpotrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 *             <const cuComplex*>A, lda, <cuComplex*>B, ldb,
 */
        __pyx_v_status = cusolverDnCpotrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, __pyx_v_nrhs, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((cuComplex *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":779
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":784
 *             <const cuComplex*>A, lda, <cuComplex*>B, ldb,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zpotrs(intptr_t handle, int uplo, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":776
 *     check_status(status)
 * 
 * cpdef cpotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_43cpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_42cpotrs[] = "cpotrs(intptr_t handle, int uplo, int n, int nrhs, size_t A, int lda, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_43cpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cpotrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 1); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 2); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 3); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 4); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 5); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 6); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 7); __PYX_ERR(0, 776, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, 8); __PYX_ERR(0, 776, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cpotrs") < 0)) __PYX_ERR(0, 776, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 776, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 777, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cpotrs", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 776, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_42cpotrs(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_42cpotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cpotrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cpotrs(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":786
 *     check_status(status)
 * 
 * cpdef zpotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_45zpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zpotrs(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zpotrs", 0);

  /* "cupy/cuda/cusolver.pyx":788
 * cpdef zpotrs(intptr_t handle, int uplo, int n, int nrhs,
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZpotrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":789
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":790
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZpotrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 *             <const cuDoubleComplex*>A, lda, <cuDoubleComplex*>B, ldb,
 */
        __pyx_v_status = cusolverDnZpotrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, __pyx_v_nrhs, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":789
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZpotrs(
 *             <Handle>handle, <FillMode>uplo, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":794
 *             <const cuDoubleComplex*>A, lda, <cuDoubleComplex*>B, ldb,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":786
 *     check_status(status)
 * 
 * cpdef zpotrs(intptr_t handle, int uplo, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_45zpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_44zpotrs[] = "zpotrs(intptr_t handle, int uplo, int n, int nrhs, size_t A, int lda, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_45zpotrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zpotrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 1); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 2); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 3); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 4); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 5); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 6); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 7); __PYX_ERR(0, 786, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, 8); __PYX_ERR(0, 786, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zpotrs") < 0)) __PYX_ERR(0, 786, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 786, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 787, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zpotrs", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 786, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_44zpotrs(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_44zpotrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zpotrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zpotrs(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zpotrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":798
 * 
 * # LU factorization
 * cpdef int sgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_47sgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgetrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgetrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":801
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgetrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":802
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgetrf_bufferSize(
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":803
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgetrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnSgetrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":802
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgetrf_bufferSize(
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":805
 *         status = cusolverDnSgetrf_bufferSize(
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":806
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgetrf_bufferSize(intptr_t handle, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":798
 * 
 * # LU factorization
 * cpdef int sgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_47sgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_46sgetrf_bufferSize[] = "sgetrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_47sgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgetrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 798, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 798, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgetrf_bufferSize") < 0)) __PYX_ERR(0, 798, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 799, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgetrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 798, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_46sgetrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_46sgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sgetrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgetrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 798, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":808
 *     return lwork
 * 
 * cpdef int dgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_49dgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgetrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgetrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":811
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgetrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":812
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgetrf_bufferSize(
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":813
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgetrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnDgetrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":812
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgetrf_bufferSize(
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":815
 *         status = cusolverDnDgetrf_bufferSize(
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 815, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":816
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgetrf_bufferSize(intptr_t handle, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":808
 *     return lwork
 * 
 * cpdef int dgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_49dgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_48dgetrf_bufferSize[] = "dgetrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_49dgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgetrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 808, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 808, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 808, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 808, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgetrf_bufferSize") < 0)) __PYX_ERR(0, 808, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 808, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 808, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 808, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 809, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 809, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgetrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 808, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_48dgetrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_48dgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dgetrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgetrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 808, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":818
 *     return lwork
 * 
 * cpdef int cgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_51cgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgetrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgetrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":821
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgetrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":822
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgetrf_bufferSize(
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":823
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgetrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnCgetrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":822
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgetrf_bufferSize(
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":825
 *         status = cusolverDnCgetrf_bufferSize(
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 825, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":826
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgetrf_bufferSize(intptr_t handle, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":818
 *     return lwork
 * 
 * cpdef int cgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_51cgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_50cgetrf_bufferSize[] = "cgetrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_51cgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgetrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 818, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 818, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 818, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 818, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgetrf_bufferSize") < 0)) __PYX_ERR(0, 818, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 818, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 818, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 818, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 819, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 819, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgetrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 818, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_50cgetrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_50cgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cgetrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgetrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 818, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":828
 *     return lwork
 * 
 * cpdef int zgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_53zgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgetrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgetrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":831
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgetrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":832
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgetrf_bufferSize(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":833
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgetrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnZgetrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":832
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgetrf_bufferSize(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":835
 *         status = cusolverDnZgetrf_bufferSize(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":836
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":828
 *     return lwork
 * 
 * cpdef int zgetrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_53zgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_52zgetrf_bufferSize[] = "zgetrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_53zgetrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgetrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 828, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 828, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 828, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 828, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgetrf_bufferSize") < 0)) __PYX_ERR(0, 828, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 829, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 829, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgetrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 828, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_52zgetrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_52zgetrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zgetrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgetrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 828, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":838
 *     return lwork
 * 
 * cpdef sgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_55sgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgetrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgetrf", 0);

  /* "cupy/cuda/cusolver.pyx":840
 * cpdef sgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgetrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":841
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgetrf(
 *             <Handle>handle, m, n, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":842
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgetrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <float*>A, lda,
 *             <float*>work, <int*>devIpiv, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnSgetrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_work), ((int *)__pyx_v_devIpiv), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":841
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgetrf(
 *             <Handle>handle, m, n, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":845
 *             <Handle>handle, m, n, <float*>A, lda,
 *             <float*>work, <int*>devIpiv, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 845, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":838
 *     return lwork
 * 
 * cpdef sgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_55sgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_54sgetrf[] = "sgetrf(intptr_t handle, int m, int n, size_t A, int lda, size_t work, size_t devIpiv, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_55sgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgetrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_devIpiv,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 1); __PYX_ERR(0, 838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 2); __PYX_ERR(0, 838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 3); __PYX_ERR(0, 838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 4); __PYX_ERR(0, 838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 5); __PYX_ERR(0, 838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 6); __PYX_ERR(0, 838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, 7); __PYX_ERR(0, 838, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgetrf") < 0)) __PYX_ERR(0, 838, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 838, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 839, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgetrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 838, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_54sgetrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_54sgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgetrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgetrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":847
 *     check_status(status)
 * 
 * cpdef dgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_57dgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgetrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgetrf", 0);

  /* "cupy/cuda/cusolver.pyx":849
 * cpdef dgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgetrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":850
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgetrf(
 *             <Handle>handle, m, n, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":851
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgetrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <double*>A, lda,
 *             <double*>work, <int*>devIpiv, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnDgetrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_work), ((int *)__pyx_v_devIpiv), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":850
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgetrf(
 *             <Handle>handle, m, n, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":854
 *             <Handle>handle, m, n, <double*>A, lda,
 *             <double*>work, <int*>devIpiv, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 854, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":847
 *     check_status(status)
 * 
 * cpdef dgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_57dgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_56dgetrf[] = "dgetrf(intptr_t handle, int m, int n, size_t A, int lda, size_t work, size_t devIpiv, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_57dgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgetrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_devIpiv,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 1); __PYX_ERR(0, 847, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 2); __PYX_ERR(0, 847, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 3); __PYX_ERR(0, 847, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 4); __PYX_ERR(0, 847, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 5); __PYX_ERR(0, 847, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 6); __PYX_ERR(0, 847, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, 7); __PYX_ERR(0, 847, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgetrf") < 0)) __PYX_ERR(0, 847, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 847, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 848, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 848, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 848, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgetrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 847, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_56dgetrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_56dgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgetrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgetrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":856
 *     check_status(status)
 * 
 * cpdef cgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_59cgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgetrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgetrf", 0);

  /* "cupy/cuda/cusolver.pyx":858
 * cpdef cgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgetrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":859
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgetrf(
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":860
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgetrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 *             <cuComplex*>work, <int*>devIpiv, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnCgetrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((cuComplex *)__pyx_v_work), ((int *)__pyx_v_devIpiv), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":859
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgetrf(
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":863
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 *             <cuComplex*>work, <int*>devIpiv, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 863, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":856
 *     check_status(status)
 * 
 * cpdef cgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_59cgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_58cgetrf[] = "cgetrf(intptr_t handle, int m, int n, size_t A, int lda, size_t work, size_t devIpiv, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_59cgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgetrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_devIpiv,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 1); __PYX_ERR(0, 856, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 2); __PYX_ERR(0, 856, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 3); __PYX_ERR(0, 856, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 4); __PYX_ERR(0, 856, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 5); __PYX_ERR(0, 856, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 6); __PYX_ERR(0, 856, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, 7); __PYX_ERR(0, 856, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgetrf") < 0)) __PYX_ERR(0, 856, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 856, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 856, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 856, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 856, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 856, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 857, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 857, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 857, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgetrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 856, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_58cgetrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_58cgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgetrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgetrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":865
 *     check_status(status)
 * 
 * cpdef zgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_61zgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgetrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgetrf", 0);

  /* "cupy/cuda/cusolver.pyx":867
 * cpdef zgetrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgetrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":868
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgetrf(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":869
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgetrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 *             <cuDoubleComplex*>work, <int*>devIpiv, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnZgetrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex *)__pyx_v_work), ((int *)__pyx_v_devIpiv), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":868
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgetrf(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":872
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 *             <cuDoubleComplex*>work, <int*>devIpiv, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 872, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":865
 *     check_status(status)
 * 
 * cpdef zgetrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t work, size_t devIpiv, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_61zgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_60zgetrf[] = "zgetrf(intptr_t handle, int m, int n, size_t A, int lda, size_t work, size_t devIpiv, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_61zgetrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_work;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgetrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_work,&__pyx_n_s_devIpiv,&__pyx_n_s_devInfo,0};
    PyObject* values[8] = {0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 1); __PYX_ERR(0, 865, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 2); __PYX_ERR(0, 865, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 3); __PYX_ERR(0, 865, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 4); __PYX_ERR(0, 865, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 5); __PYX_ERR(0, 865, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 6); __PYX_ERR(0, 865, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, 7); __PYX_ERR(0, 865, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgetrf") < 0)) __PYX_ERR(0, 865, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 8) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 865, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 866, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 866, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 866, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgetrf", 1, 8, 8, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 865, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_60zgetrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_60zgetrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_work, size_t __pyx_v_devIpiv, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgetrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgetrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_work, __pyx_v_devIpiv, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":876
 * 
 * # LU solve
 * cpdef sgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_63sgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgetrs(intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgetrs", 0);

  /* "cupy/cuda/cusolver.pyx":879
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgetrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 879, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":880
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":881
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgetrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <Operation>trans, n, nrhs,
 *             <const float*> A, lda, <const int*>devIpiv,
 */
        __pyx_v_status = cusolverDnSgetrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasOperation_t)__pyx_v_trans), __pyx_v_n, __pyx_v_nrhs, ((float const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_devIpiv), ((float *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":880
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":885
 *             <const float*> A, lda, <const int*>devIpiv,
 *             <float*>B, ldb, <int*> devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgetrs(intptr_t handle, int trans, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":876
 * 
 * # LU solve
 * cpdef sgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_63sgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_62sgetrs[] = "sgetrs(intptr_t handle, int trans, int n, int nrhs, size_t A, int lda, size_t devIpiv, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_63sgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_trans;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgetrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_trans,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_devIpiv,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 1); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 2); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 3); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 4); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 5); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 6); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 7); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 8); __PYX_ERR(0, 876, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, 9); __PYX_ERR(0, 876, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgetrs") < 0)) __PYX_ERR(0, 876, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 876, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 877, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 877, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 877, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 878, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 878, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 878, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgetrs", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 876, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_62sgetrs(__pyx_self, __pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_62sgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgetrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgetrs(__pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":887
 *     check_status(status)
 * 
 * cpdef dgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_65dgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgetrs(intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgetrs", 0);

  /* "cupy/cuda/cusolver.pyx":890
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgetrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 890, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":891
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":892
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgetrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <Operation>trans, n, nrhs,
 *             <const double*> A, lda, <const int*>devIpiv,
 */
        __pyx_v_status = cusolverDnDgetrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasOperation_t)__pyx_v_trans), __pyx_v_n, __pyx_v_nrhs, ((double const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_devIpiv), ((double *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":891
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":896
 *             <const double*> A, lda, <const int*>devIpiv,
 *             <double*>B, ldb, <int*> devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgetrs(intptr_t handle, int trans, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":887
 *     check_status(status)
 * 
 * cpdef dgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_65dgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_64dgetrs[] = "dgetrs(intptr_t handle, int trans, int n, int nrhs, size_t A, int lda, size_t devIpiv, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_65dgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_trans;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgetrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_trans,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_devIpiv,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 1); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 2); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 3); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 4); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 5); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 6); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 7); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 8); __PYX_ERR(0, 887, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, 9); __PYX_ERR(0, 887, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgetrs") < 0)) __PYX_ERR(0, 887, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 887, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 888, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 889, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 889, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 889, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgetrs", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 887, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_64dgetrs(__pyx_self, __pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_64dgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgetrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgetrs(__pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":898
 *     check_status(status)
 * 
 * cpdef cgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_67cgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgetrs(intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgetrs", 0);

  /* "cupy/cuda/cusolver.pyx":901
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgetrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":902
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":903
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgetrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <Operation>trans, n, nrhs,
 *             <const cuComplex*> A, lda, <const int*>devIpiv,
 */
        __pyx_v_status = cusolverDnCgetrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasOperation_t)__pyx_v_trans), __pyx_v_n, __pyx_v_nrhs, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_devIpiv), ((cuComplex *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":902
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":907
 *             <const cuComplex*> A, lda, <const int*>devIpiv,
 *             <cuComplex*>B, ldb, <int*> devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgetrs(intptr_t handle, int trans, int n, int nrhs,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":898
 *     check_status(status)
 * 
 * cpdef cgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_67cgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_66cgetrs[] = "cgetrs(intptr_t handle, int trans, int n, int nrhs, size_t A, int lda, size_t devIpiv, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_67cgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_trans;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgetrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_trans,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_devIpiv,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 1); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 2); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 3); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 4); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 5); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 6); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 7); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 8); __PYX_ERR(0, 898, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, 9); __PYX_ERR(0, 898, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgetrs") < 0)) __PYX_ERR(0, 898, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 898, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 899, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 900, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgetrs", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 898, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_66cgetrs(__pyx_self, __pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_66cgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgetrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgetrs(__pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":909
 *     check_status(status)
 * 
 * cpdef zgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_69zgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgetrs(intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgetrs", 0);

  /* "cupy/cuda/cusolver.pyx":912
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgetrs(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":913
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":914
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgetrs(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <Operation>trans, n, nrhs,
 *             <const cuDoubleComplex*> A, lda, <const int*>devIpiv,
 */
        __pyx_v_status = cusolverDnZgetrs(((cusolverDnHandle_t)__pyx_v_handle), ((cublasOperation_t)__pyx_v_trans), __pyx_v_n, __pyx_v_nrhs, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((int const *)__pyx_v_devIpiv), ((cuDoubleComplex *)__pyx_v_B), __pyx_v_ldb, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":913
 *              size_t B, int ldb, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgetrs(
 *             <Handle>handle, <Operation>trans, n, nrhs,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":918
 *             <const cuDoubleComplex*> A, lda, <const int*>devIpiv,
 *             <cuDoubleComplex*>B, ldb, <int*> devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 918, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":909
 *     check_status(status)
 * 
 * cpdef zgetrs(intptr_t handle, int trans, int n, int nrhs,             # <<<<<<<<<<<<<<
 *              size_t A, int lda, size_t devIpiv,
 *              size_t B, int ldb, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_69zgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_68zgetrs[] = "zgetrs(intptr_t handle, int trans, int n, int nrhs, size_t A, int lda, size_t devIpiv, size_t B, int ldb, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_69zgetrs(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_trans;
  int __pyx_v_n;
  int __pyx_v_nrhs;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_devIpiv;
  size_t __pyx_v_B;
  int __pyx_v_ldb;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgetrs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_trans,&__pyx_n_s_n,&__pyx_n_s_nrhs,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_devIpiv,&__pyx_n_s_B,&__pyx_n_s_ldb,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 1); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 2); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nrhs)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 3); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 4); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 5); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devIpiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 6); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_B)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 7); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldb)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 8); __PYX_ERR(0, 909, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, 9); __PYX_ERR(0, 909, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgetrs") < 0)) __PYX_ERR(0, 909, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L3_error)
    __pyx_v_nrhs = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_nrhs == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 909, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L3_error)
    __pyx_v_devIpiv = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_devIpiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L3_error)
    __pyx_v_B = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_B == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 911, __pyx_L3_error)
    __pyx_v_ldb = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldb == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 911, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 911, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgetrs", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 909, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_68zgetrs(__pyx_self, __pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_68zgetrs(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_trans, int __pyx_v_n, int __pyx_v_nrhs, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_devIpiv, size_t __pyx_v_B, int __pyx_v_ldb, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgetrs", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgetrs(__pyx_v_handle, __pyx_v_trans, __pyx_v_n, __pyx_v_nrhs, __pyx_v_A, __pyx_v_lda, __pyx_v_devIpiv, __pyx_v_B, __pyx_v_ldb, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgetrs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":922
 * 
 * # QR factorization
 * cpdef int sgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_71sgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgeqrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgeqrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":925
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgeqrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":926
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgeqrf_bufferSize(
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":927
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgeqrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnSgeqrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":926
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgeqrf_bufferSize(
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":929
 *         status = cusolverDnSgeqrf_bufferSize(
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":930
 *             <Handle>handle, m, n, <float*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgeqrf_bufferSize(intptr_t handle, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":922
 * 
 * # QR factorization
 * cpdef int sgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_71sgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_70sgeqrf_bufferSize[] = "sgeqrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_71sgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgeqrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 922, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 922, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 922, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 922, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgeqrf_bufferSize") < 0)) __PYX_ERR(0, 922, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 922, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 922, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 922, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 923, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 923, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgeqrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 922, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_70sgeqrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_70sgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sgeqrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgeqrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 922, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":932
 *     return lwork
 * 
 * cpdef int dgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_73dgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgeqrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgeqrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":935
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgeqrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 935, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":936
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgeqrf_bufferSize(
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":937
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgeqrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnDgeqrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":936
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgeqrf_bufferSize(
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":939
 *         status = cusolverDnDgeqrf_bufferSize(
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":940
 *             <Handle>handle, m, n, <double*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgeqrf_bufferSize(intptr_t handle, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":932
 *     return lwork
 * 
 * cpdef int dgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_73dgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_72dgeqrf_bufferSize[] = "dgeqrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_73dgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgeqrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 932, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 932, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 932, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 932, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgeqrf_bufferSize") < 0)) __PYX_ERR(0, 932, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 933, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgeqrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 932, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_72dgeqrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_72dgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dgeqrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgeqrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 932, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":942
 *     return lwork
 * 
 * cpdef int cgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_75cgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgeqrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgeqrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":945
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgeqrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":946
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgeqrf_bufferSize(
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":947
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgeqrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnCgeqrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":946
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgeqrf_bufferSize(
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":949
 *         status = cusolverDnCgeqrf_bufferSize(
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":950
 *             <Handle>handle, m, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgeqrf_bufferSize(intptr_t handle, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":942
 *     return lwork
 * 
 * cpdef int cgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_75cgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_74cgeqrf_bufferSize[] = "cgeqrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_75cgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgeqrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 942, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 942, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 942, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 942, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgeqrf_bufferSize") < 0)) __PYX_ERR(0, 942, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 943, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 943, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgeqrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 942, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_74cgeqrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_74cgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cgeqrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgeqrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 942, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":952
 *     return lwork
 * 
 * cpdef int zgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_77zgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgeqrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgeqrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":955
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgeqrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":956
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgeqrf_bufferSize(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":957
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgeqrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnZgeqrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":956
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgeqrf_bufferSize(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":959
 *         status = cusolverDnZgeqrf_bufferSize(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 959, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":960
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":952
 *     return lwork
 * 
 * cpdef int zgeqrf_bufferSize(intptr_t handle, int m, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_77zgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_76zgeqrf_bufferSize[] = "zgeqrf_bufferSize(intptr_t handle, int m, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_77zgeqrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgeqrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf_bufferSize", 1, 5, 5, 1); __PYX_ERR(0, 952, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf_bufferSize", 1, 5, 5, 2); __PYX_ERR(0, 952, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf_bufferSize", 1, 5, 5, 3); __PYX_ERR(0, 952, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf_bufferSize", 1, 5, 5, 4); __PYX_ERR(0, 952, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgeqrf_bufferSize") < 0)) __PYX_ERR(0, 952, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 953, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 953, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgeqrf_bufferSize", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 952, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_76zgeqrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_76zgeqrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zgeqrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgeqrf_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 952, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgeqrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":962
 *     return lwork
 * 
 * cpdef sgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_79sgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgeqrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgeqrf", 0);

  /* "cupy/cuda/cusolver.pyx":964
 * cpdef sgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgeqrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":965
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgeqrf(
 *             <Handle>handle, m, n, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":966
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgeqrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <float*>A, lda,
 *             <float*>tau, <float*>work, lwork,
 */
        __pyx_v_status = cusolverDnSgeqrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_tau), ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":965
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgeqrf(
 *             <Handle>handle, m, n, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":970
 *             <float*>tau, <float*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":962
 *     return lwork
 * 
 * cpdef sgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_79sgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_78sgeqrf[] = "sgeqrf(intptr_t handle, int m, int n, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_79sgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgeqrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 1); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 2); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 3); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 4); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 5); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 6); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 7); __PYX_ERR(0, 962, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, 8); __PYX_ERR(0, 962, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgeqrf") < 0)) __PYX_ERR(0, 962, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 962, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 963, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgeqrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 962, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_78sgeqrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_78sgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgeqrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgeqrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":972
 *     check_status(status)
 * 
 * cpdef dgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_81dgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgeqrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgeqrf", 0);

  /* "cupy/cuda/cusolver.pyx":974
 * cpdef dgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgeqrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":975
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgeqrf(
 *             <Handle>handle, m, n, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":976
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgeqrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <double*>A, lda,
 *             <double*>tau, <double*>work, lwork,
 */
        __pyx_v_status = cusolverDnDgeqrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_tau), ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":975
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgeqrf(
 *             <Handle>handle, m, n, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":980
 *             <double*>tau, <double*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":972
 *     check_status(status)
 * 
 * cpdef dgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_81dgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_80dgeqrf[] = "dgeqrf(intptr_t handle, int m, int n, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_81dgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgeqrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 1); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 2); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 3); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 4); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 5); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 6); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 7); __PYX_ERR(0, 972, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, 8); __PYX_ERR(0, 972, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgeqrf") < 0)) __PYX_ERR(0, 972, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 972, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 973, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgeqrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 972, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_80dgeqrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_80dgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgeqrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgeqrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 972, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":982
 *     check_status(status)
 * 
 * cpdef cgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_83cgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgeqrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgeqrf", 0);

  /* "cupy/cuda/cusolver.pyx":984
 * cpdef cgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgeqrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":985
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgeqrf(
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":986
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgeqrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 *             <cuComplex*>tau, <cuComplex*>work, lwork,
 */
        __pyx_v_status = cusolverDnCgeqrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((cuComplex *)__pyx_v_tau), ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":985
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgeqrf(
 *             <Handle>handle, m, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":990
 *             <cuComplex*>tau, <cuComplex*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":982
 *     check_status(status)
 * 
 * cpdef cgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_83cgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_82cgeqrf[] = "cgeqrf(intptr_t handle, int m, int n, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_83cgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgeqrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 1); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 2); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 3); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 4); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 5); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 6); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 7); __PYX_ERR(0, 982, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, 8); __PYX_ERR(0, 982, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgeqrf") < 0)) __PYX_ERR(0, 982, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 982, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 983, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 983, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 983, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 983, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgeqrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 982, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_82cgeqrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_82cgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgeqrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgeqrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":992
 *     check_status(status)
 * 
 * cpdef zgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_85zgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgeqrf(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgeqrf", 0);

  /* "cupy/cuda/cusolver.pyx":994
 * cpdef zgeqrf(intptr_t handle, int m, int n, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgeqrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":995
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgeqrf(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":996
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgeqrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 *             <cuDoubleComplex*>tau, <cuDoubleComplex*>work, lwork,
 */
        __pyx_v_status = cusolverDnZgeqrf(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex *)__pyx_v_tau), ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":995
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgeqrf(
 *             <Handle>handle, m, n, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1000
 *             <cuDoubleComplex*>tau, <cuDoubleComplex*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1000, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":992
 *     check_status(status)
 * 
 * cpdef zgeqrf(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_85zgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_84zgeqrf[] = "zgeqrf(intptr_t handle, int m, int n, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_85zgeqrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgeqrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 1); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 2); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 3); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 4); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 5); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 6); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 7); __PYX_ERR(0, 992, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, 8); __PYX_ERR(0, 992, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgeqrf") < 0)) __PYX_ERR(0, 992, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 992, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 992, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 992, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 992, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 992, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgeqrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 992, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_84zgeqrf(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_84zgeqrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgeqrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgeqrf(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgeqrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1004
 * 
 * # Generate unitary matrix Q from QR factorization
 * cpdef int sorgqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_87sorgqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sorgqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sorgqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1007
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSorgqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1008
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSorgqr_bufferSize(
 *             <Handle>handle, m, n, k, <const float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1009
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSorgqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <const float*>A, lda,
 *             <const float*>tau, &lwork)
 */
        __pyx_v_status = cusolverDnSorgqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((float const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_tau), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1008
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSorgqr_bufferSize(
 *             <Handle>handle, m, n, k, <const float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1012
 *             <Handle>handle, m, n, k, <const float*>A, lda,
 *             <const float*>tau, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1012, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1013
 *             <const float*>tau, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dorgqr_bufferSize(intptr_t handle, int m, int n, int k,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1004
 * 
 * # Generate unitary matrix Q from QR factorization
 * cpdef int sorgqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sorgqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_87sorgqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_86sorgqr_bufferSize[] = "sorgqr_bufferSize(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_87sorgqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sorgqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1004, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1004, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1004, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1004, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1004, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1004, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sorgqr_bufferSize") < 0)) __PYX_ERR(0, 1004, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1005, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1005, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1005, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sorgqr_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1004, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sorgqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_86sorgqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_86sorgqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sorgqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sorgqr_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sorgqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1015
 *     return lwork
 * 
 * cpdef int dorgqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_89dorgqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dorgqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dorgqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1018
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDorgqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1019
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDorgqr_bufferSize(
 *             <Handle>handle, m, n, k, <const double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1020
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDorgqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <const double*>A, lda,
 *             <const double*>tau, &lwork)
 */
        __pyx_v_status = cusolverDnDorgqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((double const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_tau), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1019
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDorgqr_bufferSize(
 *             <Handle>handle, m, n, k, <const double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1023
 *             <Handle>handle, m, n, k, <const double*>A, lda,
 *             <const double*>tau, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1024
 *             <const double*>tau, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cungqr_bufferSize(intptr_t handle, int m, int n, int k,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1015
 *     return lwork
 * 
 * cpdef int dorgqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dorgqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_89dorgqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_88dorgqr_bufferSize[] = "dorgqr_bufferSize(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_89dorgqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dorgqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dorgqr_bufferSize") < 0)) __PYX_ERR(0, 1015, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1015, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1015, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1015, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1015, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1016, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dorgqr_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1015, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dorgqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_88dorgqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_88dorgqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dorgqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dorgqr_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1015, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dorgqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1026
 *     return lwork
 * 
 * cpdef int cungqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_91cungqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cungqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cungqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1029
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCungqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1029, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1030
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCungqr_bufferSize(
 *             <Handle>handle, m, n, k, <const cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1031
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCungqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <const cuComplex*>A, lda,
 *             <const cuComplex*>tau, &lwork)
 */
        __pyx_v_status = cusolverDnCungqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((cuComplex const *)__pyx_v_tau), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1030
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCungqr_bufferSize(
 *             <Handle>handle, m, n, k, <const cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1034
 *             <Handle>handle, m, n, k, <const cuComplex*>A, lda,
 *             <const cuComplex*>tau, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1035
 *             <const cuComplex*>tau, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zungqr_bufferSize(intptr_t handle, int m, int n, int k,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1026
 *     return lwork
 * 
 * cpdef int cungqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cungqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_91cungqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_90cungqr_bufferSize[] = "cungqr_bufferSize(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_91cungqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cungqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1026, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1026, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1026, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1026, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1026, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1026, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cungqr_bufferSize") < 0)) __PYX_ERR(0, 1026, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1027, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1027, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1027, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cungqr_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1026, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cungqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_90cungqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_90cungqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cungqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cungqr_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1026, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cungqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1037
 *     return lwork
 * 
 * cpdef int zungqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_93zungqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zungqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zungqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1040
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZungqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1041
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZungqr_bufferSize(
 *             <Handle>handle, m, n, k, <const cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1042
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZungqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <const cuDoubleComplex*>A, lda,
 *             <const cuDoubleComplex*>tau, &lwork)
 */
        __pyx_v_status = cusolverDnZungqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex const *)__pyx_v_tau), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1041
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZungqr_bufferSize(
 *             <Handle>handle, m, n, k, <const cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1045
 *             <Handle>handle, m, n, k, <const cuDoubleComplex*>A, lda,
 *             <const cuDoubleComplex*>tau, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1046
 *             <const cuDoubleComplex*>tau, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1037
 *     return lwork
 * 
 * cpdef int zungqr_bufferSize(intptr_t handle, int m, int n, int k,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t tau) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zungqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_93zungqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_92zungqr_bufferSize[] = "zungqr_bufferSize(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_93zungqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zungqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1037, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1037, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1037, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1037, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1037, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1037, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zungqr_bufferSize") < 0)) __PYX_ERR(0, 1037, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1038, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zungqr_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1037, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zungqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_92zungqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_92zungqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zungqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zungqr_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1037, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1037, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zungqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1048
 *     return lwork
 * 
 * cpdef sorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_95sorgqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sorgqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sorgqr", 0);

  /* "cupy/cuda/cusolver.pyx":1050
 * cpdef sorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSorgqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1050, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1051
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSorgqr(
 *             <Handle>handle, m, n, k, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1052
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSorgqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <float*>A, lda,
 *             <const float*>tau, <float*>work, lwork,
 */
        __pyx_v_status = cusolverDnSorgqr(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((float *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_tau), ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1051
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSorgqr(
 *             <Handle>handle, m, n, k, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1056
 *             <const float*>tau, <float*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1048
 *     return lwork
 * 
 * cpdef sorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sorgqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_95sorgqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_94sorgqr[] = "sorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_95sorgqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sorgqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 1); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 2); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 3); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 4); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 5); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 6); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 7); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 8); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, 9); __PYX_ERR(0, 1048, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sorgqr") < 0)) __PYX_ERR(0, 1048, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1048, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1049, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1049, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1049, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1049, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sorgqr", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1048, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sorgqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_94sorgqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_94sorgqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sorgqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sorgqr(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sorgqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1058
 *     check_status(status)
 * 
 * cpdef dorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_97dorgqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dorgqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dorgqr", 0);

  /* "cupy/cuda/cusolver.pyx":1060
 * cpdef dorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDorgqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1061
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDorgqr(
 *             <Handle>handle, m, n, k, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1062
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDorgqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <double*>A, lda,
 *             <const double*>tau, <double*>work, lwork,
 */
        __pyx_v_status = cusolverDnDorgqr(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((double *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_tau), ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1061
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDorgqr(
 *             <Handle>handle, m, n, k, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1066
 *             <const double*>tau, <double*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1066, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1058
 *     check_status(status)
 * 
 * cpdef dorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dorgqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_97dorgqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_96dorgqr[] = "dorgqr(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_97dorgqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dorgqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 1); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 2); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 3); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 4); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 5); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 6); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 7); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 8); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, 9); __PYX_ERR(0, 1058, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dorgqr") < 0)) __PYX_ERR(0, 1058, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dorgqr", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1058, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dorgqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_96dorgqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_96dorgqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dorgqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dorgqr(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dorgqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1068
 *     check_status(status)
 * 
 * cpdef cungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_99cungqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cungqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cungqr", 0);

  /* "cupy/cuda/cusolver.pyx":1070
 * cpdef cungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCungqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1070, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1071
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCungqr(
 *             <Handle>handle, m, n, k, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1072
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCungqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <cuComplex*>A, lda,
 *             <const cuComplex*>tau, <cuComplex*>work, lwork,
 */
        __pyx_v_status = cusolverDnCungqr(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((cuComplex const *)__pyx_v_tau), ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1071
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCungqr(
 *             <Handle>handle, m, n, k, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1076
 *             <const cuComplex*>tau, <cuComplex*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1076, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1068
 *     check_status(status)
 * 
 * cpdef cungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cungqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_99cungqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_98cungqr[] = "cungqr(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_99cungqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cungqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 1); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 2); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 3); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 4); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 5); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 6); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 7); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 8); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, 9); __PYX_ERR(0, 1068, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cungqr") < 0)) __PYX_ERR(0, 1068, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1068, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1069, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cungqr", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1068, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cungqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_98cungqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_98cungqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cungqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cungqr(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1068, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cungqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1078
 *     check_status(status)
 * 
 * cpdef zungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_101zungqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zungqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zungqr", 0);

  /* "cupy/cuda/cusolver.pyx":1080
 * cpdef zungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZungqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1081
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZungqr(
 *             <Handle>handle, m, n, k, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1082
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZungqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n, k, <cuDoubleComplex*>A, lda,
 *             <const cuDoubleComplex*>tau, <cuDoubleComplex*>work, lwork,
 */
        __pyx_v_status = cusolverDnZungqr(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex const *)__pyx_v_tau), ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1081
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZungqr(
 *             <Handle>handle, m, n, k, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1086
 *             <const cuDoubleComplex*>tau, <cuDoubleComplex*>work, lwork,
 *             <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1086, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1078
 *     check_status(status)
 * 
 * cpdef zungqr(intptr_t handle, int m, int n, int k, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t tau, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zungqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_101zungqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_100zungqr[] = "zungqr(intptr_t handle, int m, int n, int k, size_t A, int lda, size_t tau, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_101zungqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zungqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 1); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 2); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 3); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 4); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 5); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 6); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 7); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 8); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, 9); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zungqr") < 0)) __PYX_ERR(0, 1078, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1079, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zungqr", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1078, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zungqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_100zungqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_100zungqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zungqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zungqr(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zungqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1090
 * 
 * # Compute Q**T*b in solve min||A*x = b||
 * cpdef int sormqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_103sormqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sormqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sormqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1094
 *                             size_t C, int ldc) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSormqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1095
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSormqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1096
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSormqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const float*>A, lda, <const float*>tau,
 */
        __pyx_v_status = cusolverDnSormqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((float const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_tau), ((float *)__pyx_v_C), __pyx_v_ldc, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1095
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSormqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1100
 *             <const float*>A, lda, <const float*>tau,
 *             <float*>C, ldc, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1101
 *             <float*>C, ldc, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dormqr_bufferSize(intptr_t handle, int side, int trans,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1090
 * 
 * # Compute Q**T*b in solve min||A*x = b||
 * cpdef int sormqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sormqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_103sormqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_102sormqr_bufferSize[] = "sormqr_bufferSize(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_103sormqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sormqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 1); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 2); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 3); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 4); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 5); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 6); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 7); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 8); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 9); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, 10); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sormqr_bufferSize") < 0)) __PYX_ERR(0, 1090, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1090, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1090, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1090, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1091, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1092, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sormqr_bufferSize", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1090, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sormqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_102sormqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_102sormqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sormqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sormqr_bufferSize(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1090, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sormqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1103
 *     return lwork
 * 
 * cpdef int dormqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_105dormqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dormqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dormqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1107
 *                             size_t C, int ldc) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDormqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1108
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDormqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1109
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDormqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const double*>A, lda, <const double*>tau,
 */
        __pyx_v_status = cusolverDnDormqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((double const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_tau), ((double *)__pyx_v_C), __pyx_v_ldc, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1108
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDormqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1113
 *             <const double*>A, lda, <const double*>tau,
 *             <double*>C, ldc, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1114
 *             <double*>C, ldc, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cunmqr_bufferSize(intptr_t handle, int side, int trans,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1103
 *     return lwork
 * 
 * cpdef int dormqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dormqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_105dormqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_104dormqr_bufferSize[] = "dormqr_bufferSize(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_105dormqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dormqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 1); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 2); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 3); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 4); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 5); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 6); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 7); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 8); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 9); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, 10); __PYX_ERR(0, 1103, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dormqr_bufferSize") < 0)) __PYX_ERR(0, 1103, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1105, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dormqr_bufferSize", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1103, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dormqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_104dormqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_104dormqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dormqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dormqr_bufferSize(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1103, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dormqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1116
 *     return lwork
 * 
 * cpdef int cunmqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_107cunmqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cunmqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cunmqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1120
 *                             size_t C, int ldc) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCunmqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1121
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCunmqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1122
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCunmqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const cuComplex*>A, lda, <const cuComplex*>tau,
 */
        __pyx_v_status = cusolverDnCunmqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((cuComplex const *)__pyx_v_tau), ((cuComplex *)__pyx_v_C), __pyx_v_ldc, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1121
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCunmqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1126
 *             <const cuComplex*>A, lda, <const cuComplex*>tau,
 *             <cuComplex*>C, ldc, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1127
 *             <cuComplex*>C, ldc, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zunmqr_bufferSize(intptr_t handle, int side, int trans,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1116
 *     return lwork
 * 
 * cpdef int cunmqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cunmqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_107cunmqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_106cunmqr_bufferSize[] = "cunmqr_bufferSize(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_107cunmqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cunmqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 1); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 2); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 3); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 4); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 5); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 6); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 7); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 8); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 9); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, 10); __PYX_ERR(0, 1116, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cunmqr_bufferSize") < 0)) __PYX_ERR(0, 1116, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1117, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1118, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1118, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cunmqr_bufferSize", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1116, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cunmqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_106cunmqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_106cunmqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cunmqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cunmqr_bufferSize(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1116, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cunmqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1129
 *     return lwork
 * 
 * cpdef int zunmqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_109zunmqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zunmqr_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zunmqr_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1133
 *                             size_t C, int ldc) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZunmqr_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1134
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZunmqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1135
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZunmqr_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const cuDoubleComplex*>A, lda, <const cuDoubleComplex*>tau,
 */
        __pyx_v_status = cusolverDnZunmqr_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex const *)__pyx_v_tau), ((cuDoubleComplex *)__pyx_v_C), __pyx_v_ldc, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1134
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZunmqr_bufferSize(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1139
 *             <const cuDoubleComplex*>A, lda, <const cuDoubleComplex*>tau,
 *             <cuDoubleComplex*>C, ldc, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1140
 *             <cuDoubleComplex*>C, ldc, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1129
 *     return lwork
 * 
 * cpdef int zunmqr_bufferSize(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *                             int m, int n, int k, size_t A, int lda, size_t tau,
 *                             size_t C, int ldc) except? -1:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zunmqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_109zunmqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_108zunmqr_bufferSize[] = "zunmqr_bufferSize(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_109zunmqr_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zunmqr_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,0};
    PyObject* values[11] = {0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 1); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 2); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 3); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 4); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 5); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 6); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 7); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 8); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 9); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, 10); __PYX_ERR(0, 1129, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zunmqr_bufferSize") < 0)) __PYX_ERR(0, 1129, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 11) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1129, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1129, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1129, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1131, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1131, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zunmqr_bufferSize", 1, 11, 11, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1129, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zunmqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_108zunmqr_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_108zunmqr_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zunmqr_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zunmqr_bufferSize(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1129, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zunmqr_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1143
 * 
 * 
 * cpdef sormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_111sormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sormqr(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sormqr", 0);

  /* "cupy/cuda/cusolver.pyx":1146
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSormqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1147
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSormqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1148
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSormqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const float*>A, lda, <const float*>tau,
 */
        __pyx_v_status = cusolverDnSormqr(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((float const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_tau), ((float *)__pyx_v_C), __pyx_v_ldc, ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1147
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSormqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1153
 *             <float*>C, ldc,
 *             <float*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dormqr(intptr_t handle, int side, int trans,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1143
 * 
 * 
 * cpdef sormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_111sormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_110sormqr[] = "sormqr(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_111sormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sormqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 1); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 2); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 3); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 4); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 5); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 6); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 7); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 8); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 9); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 10); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 11); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 12); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, 13); __PYX_ERR(0, 1143, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sormqr") < 0)) __PYX_ERR(0, 1143, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1143, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1143, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1143, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1144, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1145, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1145, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1145, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1145, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1145, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sormqr", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1143, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_110sormqr(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_110sormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sormqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sormqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1155
 *     check_status(status)
 * 
 * cpdef dormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_113dormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dormqr(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dormqr", 0);

  /* "cupy/cuda/cusolver.pyx":1158
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDormqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1159
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDormqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1160
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDormqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const double*>A, lda, <const double*>tau,
 */
        __pyx_v_status = cusolverDnDormqr(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((double const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_tau), ((double *)__pyx_v_C), __pyx_v_ldc, ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1159
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDormqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1165
 *             <double*>C, ldc,
 *             <double*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cunmqr(intptr_t handle, int side, int trans,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1155
 *     check_status(status)
 * 
 * cpdef dormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_113dormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_112dormqr[] = "dormqr(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_113dormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dormqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 1); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 2); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 3); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 4); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 5); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 6); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 7); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 8); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 9); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 10); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 11); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 12); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, 13); __PYX_ERR(0, 1155, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dormqr") < 0)) __PYX_ERR(0, 1155, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1155, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1155, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1155, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1156, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1156, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1156, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1156, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1156, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1156, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1157, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1157, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1157, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1157, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1157, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dormqr", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1155, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_112dormqr(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_112dormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dormqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dormqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1167
 *     check_status(status)
 * 
 * cpdef cunmqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_115cunmqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cunmqr(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cunmqr", 0);

  /* "cupy/cuda/cusolver.pyx":1170
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCunmqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1171
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCunmqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1172
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCunmqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const cuComplex*>A, lda, <const cuComplex*>tau,
 */
        __pyx_v_status = cusolverDnCunmqr(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((cuComplex const *)__pyx_v_tau), ((cuComplex *)__pyx_v_C), __pyx_v_ldc, ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1171
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCunmqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1177
 *             <cuComplex*>C, ldc,
 *             <cuComplex*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zunmqr(intptr_t handle, int side, int trans,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1167
 *     check_status(status)
 * 
 * cpdef cunmqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cunmqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_115cunmqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_114cunmqr[] = "cunmqr(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_115cunmqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cunmqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 1); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 2); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 3); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 4); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 5); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 6); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 7); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 8); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 9); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 10); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 11); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 12); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, 13); __PYX_ERR(0, 1167, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cunmqr") < 0)) __PYX_ERR(0, 1167, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1167, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1167, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1167, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1168, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1169, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cunmqr", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1167, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cunmqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_114cunmqr(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_114cunmqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cunmqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cunmqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cunmqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1179
 *     check_status(status)
 * 
 * cpdef zunmqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_117zunmqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zunmqr(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zunmqr", 0);

  /* "cupy/cuda/cusolver.pyx":1182
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZunmqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1183
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZunmqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1184
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZunmqr(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 *             <const cuDoubleComplex*>A, lda, <const cuDoubleComplex*>tau,
 */
        __pyx_v_status = cusolverDnZunmqr(((cusolverDnHandle_t)__pyx_v_handle), ((cublasSideMode_t)__pyx_v_side), ((cublasOperation_t)__pyx_v_trans), __pyx_v_m, __pyx_v_n, __pyx_v_k, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((cuDoubleComplex const *)__pyx_v_tau), ((cuDoubleComplex *)__pyx_v_C), __pyx_v_ldc, ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1183
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZunmqr(
 *             <Handle>handle, <SideMode>side, <Operation>trans, m, n, k,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1189
 *             <cuDoubleComplex*>C, ldc,
 *             <cuDoubleComplex*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * # (obsoleted)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1179
 *     check_status(status)
 * 
 * cpdef zunmqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zunmqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_117zunmqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_116zunmqr[] = "zunmqr(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_117zunmqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zunmqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 1); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 2); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 3); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 4); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 5); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 6); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 7); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 8); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 9); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 10); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 11); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 12); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, 13); __PYX_ERR(0, 1179, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zunmqr") < 0)) __PYX_ERR(0, 1179, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1179, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1179, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1179, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1180, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1181, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zunmqr", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1179, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zunmqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_116zunmqr(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_116zunmqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zunmqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zunmqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zunmqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1192
 * 
 * # (obsoleted)
 * cpdef cormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_119cormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cormqr(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cormqr", 0);

  /* "cupy/cuda/cusolver.pyx":1195
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     return cunmqr(handle, side, trans, m, n, k, A, lda, tau,             # <<<<<<<<<<<<<<
 *                   C, ldc, work, lwork, devInfo)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "cupy/cuda/cusolver.pyx":1196
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     return cunmqr(handle, side, trans, m, n, k, A, lda, tau,
 *                   C, ldc, work, lwork, devInfo)             # <<<<<<<<<<<<<<
 * 
 * # (obsoleted)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cunmqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1192
 * 
 * # (obsoleted)
 * cpdef cormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_119cormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_118cormqr[] = "cormqr(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_119cormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cormqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 1); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 2); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 3); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 4); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 5); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 6); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 7); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 8); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 9); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 10); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 11); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 12); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, 13); __PYX_ERR(0, 1192, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cormqr") < 0)) __PYX_ERR(0, 1192, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1192, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1192, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1192, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1193, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1194, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cormqr", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1192, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_118cormqr(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_118cormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cormqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cormqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1199
 * 
 * # (obsoleted)
 * cpdef zormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_121zormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zormqr(intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zormqr", 0);

  /* "cupy/cuda/cusolver.pyx":1202
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     return zunmqr(handle, side, trans, m, n, k, A, lda, tau,             # <<<<<<<<<<<<<<
 *                   C, ldc, work, lwork, devInfo)
 * 
 */
  __Pyx_XDECREF(__pyx_r);

  /* "cupy/cuda/cusolver.pyx":1203
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 *     return zunmqr(handle, side, trans, m, n, k, A, lda, tau,
 *                   C, ldc, work, lwork, devInfo)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zunmqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1199
 * 
 * # (obsoleted)
 * cpdef zormqr(intptr_t handle, int side, int trans,             # <<<<<<<<<<<<<<
 *              int m, int n, int k, size_t A, int lda, size_t tau,
 *              size_t C, int ldc, size_t work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_121zormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_120zormqr[] = "zormqr(intptr_t handle, int side, int trans, int m, int n, int k, size_t A, int lda, size_t tau, size_t C, int ldc, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_121zormqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_side;
  int __pyx_v_trans;
  int __pyx_v_m;
  int __pyx_v_n;
  int __pyx_v_k;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_tau;
  size_t __pyx_v_C;
  int __pyx_v_ldc;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zormqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_side,&__pyx_n_s_trans,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_k,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_tau,&__pyx_n_s_C,&__pyx_n_s_ldc,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[14] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_side)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 1); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_trans)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 2); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 3); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 4); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_k)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 5); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 6); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 7); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tau)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 8); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_C)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 9); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldc)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 10); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 11); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 12); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, 13); __PYX_ERR(0, 1199, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zormqr") < 0)) __PYX_ERR(0, 1199, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 14) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1199, __pyx_L3_error)
    __pyx_v_side = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_side == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1199, __pyx_L3_error)
    __pyx_v_trans = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_trans == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1199, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_k = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_k == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_tau = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tau == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1200, __pyx_L3_error)
    __pyx_v_C = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_C == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1201, __pyx_L3_error)
    __pyx_v_ldc = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldc == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1201, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1201, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1201, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[13]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1201, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zormqr", 1, 14, 14, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1199, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_120zormqr(__pyx_self, __pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_120zormqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_side, int __pyx_v_trans, int __pyx_v_m, int __pyx_v_n, int __pyx_v_k, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_tau, size_t __pyx_v_C, int __pyx_v_ldc, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zormqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zormqr(__pyx_v_handle, __pyx_v_side, __pyx_v_trans, __pyx_v_m, __pyx_v_n, __pyx_v_k, __pyx_v_A, __pyx_v_lda, __pyx_v_tau, __pyx_v_C, __pyx_v_ldc, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zormqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1207
 * 
 * # L*D*L**T,U*D*U**T factorization
 * cpdef int ssytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_123ssytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_ssytrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ssytrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1210
 *                             int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSsytrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1211
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsytrf_bufferSize(
 *             <Handle>handle, n, <float*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1212
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSsytrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, n, <float*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnSsytrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1211
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsytrf_bufferSize(
 *             <Handle>handle, n, <float*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1214
 *         status = cusolverDnSsytrf_bufferSize(
 *             <Handle>handle, n, <float*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1215
 *             <Handle>handle, n, <float*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dsytrf_bufferSize(intptr_t handle, int n, size_t A,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1207
 * 
 * # L*D*L**T,U*D*U**T factorization
 * cpdef int ssytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_123ssytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_122ssytrf_bufferSize[] = "ssytrf_bufferSize(intptr_t handle, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_123ssytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ssytrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf_bufferSize", 1, 4, 4, 1); __PYX_ERR(0, 1207, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf_bufferSize", 1, 4, 4, 2); __PYX_ERR(0, 1207, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf_bufferSize", 1, 4, 4, 3); __PYX_ERR(0, 1207, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ssytrf_bufferSize") < 0)) __PYX_ERR(0, 1207, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1207, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1207, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1207, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1208, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ssytrf_bufferSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1207, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_122ssytrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_122ssytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("ssytrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ssytrf_bufferSize(__pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1207, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1217
 *     return lwork
 * 
 * cpdef int dsytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_125dsytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dsytrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dsytrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1220
 *                             int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDsytrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1221
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsytrf_bufferSize(
 *             <Handle>handle, n, <double*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1222
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDsytrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, n, <double*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnDsytrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1221
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsytrf_bufferSize(
 *             <Handle>handle, n, <double*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1224
 *         status = cusolverDnDsytrf_bufferSize(
 *             <Handle>handle, n, <double*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1225
 *             <Handle>handle, n, <double*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int csytrf_bufferSize(intptr_t handle, int n, size_t A,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1217
 *     return lwork
 * 
 * cpdef int dsytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_125dsytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_124dsytrf_bufferSize[] = "dsytrf_bufferSize(intptr_t handle, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_125dsytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dsytrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf_bufferSize", 1, 4, 4, 1); __PYX_ERR(0, 1217, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf_bufferSize", 1, 4, 4, 2); __PYX_ERR(0, 1217, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf_bufferSize", 1, 4, 4, 3); __PYX_ERR(0, 1217, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dsytrf_bufferSize") < 0)) __PYX_ERR(0, 1217, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1218, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dsytrf_bufferSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1217, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_124dsytrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_124dsytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dsytrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dsytrf_bufferSize(__pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1217, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1227
 *     return lwork
 * 
 * cpdef int csytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_127csytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_csytrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("csytrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1230
 *                             int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCsytrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1231
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCsytrf_bufferSize(
 *             <Handle>handle, n, <cuComplex*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1232
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCsytrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnCsytrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1231
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCsytrf_bufferSize(
 *             <Handle>handle, n, <cuComplex*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1234
 *         status = cusolverDnCsytrf_bufferSize(
 *             <Handle>handle, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1235
 *             <Handle>handle, n, <cuComplex*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zsytrf_bufferSize(intptr_t handle, int n, size_t A,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1227
 *     return lwork
 * 
 * cpdef int csytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.csytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_127csytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_126csytrf_bufferSize[] = "csytrf_bufferSize(intptr_t handle, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_127csytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("csytrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf_bufferSize", 1, 4, 4, 1); __PYX_ERR(0, 1227, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf_bufferSize", 1, 4, 4, 2); __PYX_ERR(0, 1227, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf_bufferSize", 1, 4, 4, 3); __PYX_ERR(0, 1227, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "csytrf_bufferSize") < 0)) __PYX_ERR(0, 1227, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1227, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1227, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1227, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1228, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("csytrf_bufferSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1227, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.csytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_126csytrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_126csytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("csytrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_csytrf_bufferSize(__pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1227, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.csytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1237
 *     return lwork
 * 
 * cpdef int zsytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_129zsytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zsytrf_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zsytrf_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1240
 *                             int lda) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZsytrf_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1241
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZsytrf_bufferSize(
 *             <Handle>handle, n, <cuDoubleComplex*>A, lda, &lwork)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1242
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZsytrf_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)
 */
        __pyx_v_status = cusolverDnZsytrf_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1241
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZsytrf_bufferSize(
 *             <Handle>handle, n, <cuDoubleComplex*>A, lda, &lwork)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1244
 *         status = cusolverDnZsytrf_bufferSize(
 *             <Handle>handle, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1244, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1245
 *             <Handle>handle, n, <cuDoubleComplex*>A, lda, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef ssytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1237
 *     return lwork
 * 
 * cpdef int zsytrf_bufferSize(intptr_t handle, int n, size_t A,             # <<<<<<<<<<<<<<
 *                             int lda) except? -1:
 *     cdef int lwork
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zsytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_129zsytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_128zsytrf_bufferSize[] = "zsytrf_bufferSize(intptr_t handle, int n, size_t A, int lda) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_129zsytrf_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zsytrf_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf_bufferSize", 1, 4, 4, 1); __PYX_ERR(0, 1237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf_bufferSize", 1, 4, 4, 2); __PYX_ERR(0, 1237, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf_bufferSize", 1, 4, 4, 3); __PYX_ERR(0, 1237, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zsytrf_bufferSize") < 0)) __PYX_ERR(0, 1237, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1237, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1237, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[2]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1237, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1238, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zsytrf_bufferSize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zsytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_128zsytrf_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_128zsytrf_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zsytrf_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zsytrf_bufferSize(__pyx_v_handle, __pyx_v_n, __pyx_v_A, __pyx_v_lda, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1237, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zsytrf_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1247
 *     return lwork
 * 
 * cpdef ssytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_131ssytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ssytrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ssytrf", 0);

  /* "cupy/cuda/cusolver.pyx":1249
 * cpdef ssytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSsytrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1249, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1250
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1251
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSsytrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <float*>A, lda,
 *             <int*>ipiv, <float*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnSsytrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((int *)__pyx_v_ipiv), ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1250
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1254
 *             <Handle>handle, <FillMode>uplo, n, <float*>A, lda,
 *             <int*>ipiv, <float*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1247
 *     return lwork
 * 
 * cpdef ssytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_131ssytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_130ssytrf[] = "ssytrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t ipiv, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_131ssytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_ipiv;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ssytrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_ipiv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 1); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 2); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 3); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 4); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ipiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 5); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 6); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 7); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, 8); __PYX_ERR(0, 1247, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ssytrf") < 0)) __PYX_ERR(0, 1247, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1247, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1247, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1247, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1247, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1247, __pyx_L3_error)
    __pyx_v_ipiv = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_ipiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1248, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ssytrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1247, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_130ssytrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_130ssytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ssytrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ssytrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1256
 *     check_status(status)
 * 
 * cpdef dsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_133dsytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dsytrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dsytrf", 0);

  /* "cupy/cuda/cusolver.pyx":1258
 * cpdef dsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDsytrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1259
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1260
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDsytrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <double*>A, lda,
 *             <int*>ipiv, <double*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnDsytrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((int *)__pyx_v_ipiv), ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1259
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1263
 *             <Handle>handle, <FillMode>uplo, n, <double*>A, lda,
 *             <int*>ipiv, <double*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef csytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1256
 *     check_status(status)
 * 
 * cpdef dsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_133dsytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_132dsytrf[] = "dsytrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t ipiv, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_133dsytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_ipiv;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dsytrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_ipiv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 1); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 2); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 3); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 4); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ipiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 5); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 6); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 7); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, 8); __PYX_ERR(0, 1256, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dsytrf") < 0)) __PYX_ERR(0, 1256, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1256, __pyx_L3_error)
    __pyx_v_ipiv = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_ipiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1257, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1257, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1257, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1257, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dsytrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1256, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_132dsytrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_132dsytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dsytrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dsytrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1265
 *     check_status(status)
 * 
 * cpdef csytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_135csytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_csytrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("csytrf", 0);

  /* "cupy/cuda/cusolver.pyx":1267
 * cpdef csytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCsytrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1267, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1268
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1269
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCsytrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A, lda,
 *             <int*>ipiv, <cuComplex*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnCsytrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((int *)__pyx_v_ipiv), ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1268
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1272
 *             <Handle>handle, <FillMode>uplo, n, <cuComplex*>A, lda,
 *             <int*>ipiv, <cuComplex*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1272, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1265
 *     check_status(status)
 * 
 * cpdef csytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.csytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_135csytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_134csytrf[] = "csytrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t ipiv, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_135csytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_ipiv;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("csytrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_ipiv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 1); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 2); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 3); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 4); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ipiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 5); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 6); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 7); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, 8); __PYX_ERR(0, 1265, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "csytrf") < 0)) __PYX_ERR(0, 1265, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1265, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1265, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1265, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1265, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1265, __pyx_L3_error)
    __pyx_v_ipiv = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_ipiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1266, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("csytrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1265, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.csytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_134csytrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_134csytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("csytrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_csytrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1265, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.csytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1274
 *     check_status(status)
 * 
 * cpdef zsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_137zsytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zsytrf(intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zsytrf", 0);

  /* "cupy/cuda/cusolver.pyx":1276
 * cpdef zsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZsytrf(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1277
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1278
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZsytrf(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A, lda,
 *             <int*>ipiv, <cuDoubleComplex*>work, lwork, <int*>devInfo)
 */
        __pyx_v_status = cusolverDnZsytrf(((cusolverDnHandle_t)__pyx_v_handle), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((int *)__pyx_v_ipiv), ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1277
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZsytrf(
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1281
 *             <Handle>handle, <FillMode>uplo, n, <cuDoubleComplex*>A, lda,
 *             <int*>ipiv, <cuDoubleComplex*>work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1274
 *     check_status(status)
 * 
 * cpdef zsytrf(intptr_t handle, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t ipiv, size_t work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zsytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_137zsytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_136zsytrf[] = "zsytrf(intptr_t handle, int uplo, int n, size_t A, int lda, size_t ipiv, size_t work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_137zsytrf(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_ipiv;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zsytrf (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_ipiv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 1); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 2); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 3); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 4); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ipiv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 5); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 6); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 7); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, 8); __PYX_ERR(0, 1274, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zsytrf") < 0)) __PYX_ERR(0, 1274, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 9) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1274, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1274, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1274, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1274, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1274, __pyx_L3_error)
    __pyx_v_ipiv = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_ipiv == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1275, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1275, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[7]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1275, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1275, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zsytrf", 1, 9, 9, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1274, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zsytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_136zsytrf(__pyx_self, __pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_136zsytrf(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_ipiv, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zsytrf", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zsytrf(__pyx_v_handle, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_ipiv, __pyx_v_work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zsytrf", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1289
 * 
 * # Bidiagonal factorization
 * cpdef int sgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_139sgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgebrd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgebrd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1291
 * cpdef int sgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1292
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1293
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgebrd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnSgebrd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1292
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1294
 *     with nogil:
 *         status = cusolverDnSgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1295
 *         status = cusolverDnSgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1289
 * 
 * # Bidiagonal factorization
 * cpdef int sgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_139sgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_138sgebrd_bufferSize[] = "sgebrd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_139sgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgebrd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1289, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1289, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgebrd_bufferSize") < 0)) __PYX_ERR(0, 1289, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgebrd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1289, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_138sgebrd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_138sgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sgebrd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgebrd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1289, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1297
 *     return lwork
 * 
 * cpdef int dgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_141dgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgebrd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgebrd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1299
 * cpdef int dgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1300
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1301
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgebrd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnDgebrd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1300
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1302
 *     with nogil:
 *         status = cusolverDnDgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1303
 *         status = cusolverDnDgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1297
 *     return lwork
 * 
 * cpdef int dgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_141dgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_140dgebrd_bufferSize[] = "dgebrd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_141dgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgebrd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1297, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1297, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgebrd_bufferSize") < 0)) __PYX_ERR(0, 1297, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgebrd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1297, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_140dgebrd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_140dgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dgebrd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgebrd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1297, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1305
 *     return lwork
 * 
 * cpdef int cgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_143cgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgebrd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgebrd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1307
 * cpdef int cgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1308
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1309
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgebrd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnCgebrd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1308
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1310
 *     with nogil:
 *         status = cusolverDnCgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1311
 *         status = cusolverDnCgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1305
 *     return lwork
 * 
 * cpdef int cgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_143cgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_142cgebrd_bufferSize[] = "cgebrd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_143cgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgebrd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1305, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgebrd_bufferSize") < 0)) __PYX_ERR(0, 1305, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1305, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1305, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1305, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgebrd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1305, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_142cgebrd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_142cgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cgebrd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgebrd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1305, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1313
 *     return lwork
 * 
 * cpdef int zgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_145zgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgebrd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgebrd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1315
 * cpdef int zgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1316
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1317
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgebrd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnZgebrd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1316
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1318
 *     with nogil:
 *         status = cusolverDnZgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1319
 *         status = cusolverDnZgebrd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgebrd(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1313
 *     return lwork
 * 
 * cpdef int zgebrd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_145zgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_144zgebrd_bufferSize[] = "zgebrd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_145zgebrd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgebrd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1313, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1313, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgebrd_bufferSize") < 0)) __PYX_ERR(0, 1313, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1313, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1313, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1313, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgebrd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1313, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_144zgebrd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_144zgebrd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zgebrd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgebrd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1313, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgebrd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1321
 *     return lwork
 * 
 * cpdef sgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_147sgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgebrd(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgebrd", 0);

  /* "cupy/cuda/cusolver.pyx":1324
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgebrd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1324, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1325
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgebrd(
 *             <Handle>handle, m, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1326
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgebrd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n,
 *             <float*>A, lda,
 */
        __pyx_v_status = cusolverDnSgebrd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_D), ((float *)__pyx_v_E), ((float *)__pyx_v_tauQ), ((float *)__pyx_v_tauP), ((float *)__pyx_v_Work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1325
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgebrd(
 *             <Handle>handle, m, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1332
 *             <float*>tauQ, <float*>tauP,
 *             <float*>Work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgebrd(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1321
 *     return lwork
 * 
 * cpdef sgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_147sgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_146sgebrd[] = "sgebrd(intptr_t handle, int m, int n, size_t A, int lda, size_t D, size_t E, size_t tauQ, size_t tauP, size_t Work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_147sgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_D;
  size_t __pyx_v_E;
  size_t __pyx_v_tauQ;
  size_t __pyx_v_tauP;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgebrd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_D,&__pyx_n_s_E,&__pyx_n_s_tauQ,&__pyx_n_s_tauP,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 1); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 2); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 3); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 4); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_D)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 5); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_E)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 6); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauQ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 7); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauP)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 8); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 9); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 10); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, 11); __PYX_ERR(0, 1321, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgebrd") < 0)) __PYX_ERR(0, 1321, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1321, __pyx_L3_error)
    __pyx_v_D = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_D == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L3_error)
    __pyx_v_E = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_E == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L3_error)
    __pyx_v_tauQ = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_tauQ == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L3_error)
    __pyx_v_tauP = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tauP == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1323, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1323, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1323, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgebrd", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1321, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_146sgebrd(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_146sgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgebrd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgebrd(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1334
 *     check_status(status)
 * 
 * cpdef dgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_149dgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgebrd(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgebrd", 0);

  /* "cupy/cuda/cusolver.pyx":1337
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgebrd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1338
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgebrd(
 *             <Handle>handle, m, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1339
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgebrd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n,
 *             <double*>A, lda,
 */
        __pyx_v_status = cusolverDnDgebrd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_D), ((double *)__pyx_v_E), ((double *)__pyx_v_tauQ), ((double *)__pyx_v_tauP), ((double *)__pyx_v_Work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1338
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgebrd(
 *             <Handle>handle, m, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1345
 *             <double*>tauQ, <double*>tauP,
 *             <double*>Work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgebrd(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1334
 *     check_status(status)
 * 
 * cpdef dgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_149dgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_148dgebrd[] = "dgebrd(intptr_t handle, int m, int n, size_t A, int lda, size_t D, size_t E, size_t tauQ, size_t tauP, size_t Work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_149dgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_D;
  size_t __pyx_v_E;
  size_t __pyx_v_tauQ;
  size_t __pyx_v_tauP;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgebrd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_D,&__pyx_n_s_E,&__pyx_n_s_tauQ,&__pyx_n_s_tauP,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 1); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 2); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 3); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 4); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_D)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 5); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_E)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 6); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauQ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 7); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauP)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 8); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 9); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 10); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, 11); __PYX_ERR(0, 1334, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgebrd") < 0)) __PYX_ERR(0, 1334, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1334, __pyx_L3_error)
    __pyx_v_D = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_D == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1335, __pyx_L3_error)
    __pyx_v_E = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_E == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1335, __pyx_L3_error)
    __pyx_v_tauQ = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_tauQ == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1335, __pyx_L3_error)
    __pyx_v_tauP = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tauP == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1335, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1336, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1336, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1336, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgebrd", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1334, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_148dgebrd(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_148dgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgebrd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgebrd(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1347
 *     check_status(status)
 * 
 * cpdef cgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_151cgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgebrd(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgebrd", 0);

  /* "cupy/cuda/cusolver.pyx":1350
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgebrd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1351
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgebrd(
 *             <Handle>handle, m, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1352
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgebrd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n,
 *             <cuComplex*>A, lda,
 */
        __pyx_v_status = cusolverDnCgebrd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_D), ((float *)__pyx_v_E), ((cuComplex *)__pyx_v_tauQ), ((cuComplex *)__pyx_v_tauP), ((cuComplex *)__pyx_v_Work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1351
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgebrd(
 *             <Handle>handle, m, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1358
 *             <cuComplex*>tauQ, <cuComplex*>tauP,
 *             <cuComplex*>Work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgebrd(intptr_t handle, int m, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1347
 *     check_status(status)
 * 
 * cpdef cgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_151cgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_150cgebrd[] = "cgebrd(intptr_t handle, int m, int n, size_t A, int lda, size_t D, size_t E, size_t tauQ, size_t tauP, size_t Work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_151cgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_D;
  size_t __pyx_v_E;
  size_t __pyx_v_tauQ;
  size_t __pyx_v_tauP;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgebrd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_D,&__pyx_n_s_E,&__pyx_n_s_tauQ,&__pyx_n_s_tauP,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 1); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 2); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 3); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 4); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_D)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 5); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_E)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 6); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauQ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 7); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauP)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 8); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 9); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 10); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, 11); __PYX_ERR(0, 1347, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgebrd") < 0)) __PYX_ERR(0, 1347, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L3_error)
    __pyx_v_D = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_D == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    __pyx_v_E = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_E == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    __pyx_v_tauQ = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_tauQ == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    __pyx_v_tauP = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tauP == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1349, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgebrd", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1347, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_150cgebrd(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_150cgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgebrd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgebrd(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1360
 *     check_status(status)
 * 
 * cpdef zgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_153zgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgebrd(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgebrd", 0);

  /* "cupy/cuda/cusolver.pyx":1363
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgebrd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1364
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgebrd(
 *             <Handle>handle, m, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1365
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgebrd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, m, n,
 *             <cuDoubleComplex*>A, lda,
 */
        __pyx_v_status = cusolverDnZgebrd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_D), ((double *)__pyx_v_E), ((cuDoubleComplex *)__pyx_v_tauQ), ((cuDoubleComplex *)__pyx_v_tauP), ((cuDoubleComplex *)__pyx_v_Work), __pyx_v_lwork, ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1364
 *              size_t Work, int lwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgebrd(
 *             <Handle>handle, m, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1371
 *             <cuDoubleComplex*>tauQ, <cuDoubleComplex*>tauP,
 *             <cuDoubleComplex*>Work, lwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1371, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1360
 *     check_status(status)
 * 
 * cpdef zgebrd(intptr_t handle, int m, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t D, size_t E, size_t tauQ, size_t tauP,
 *              size_t Work, int lwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_153zgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_152zgebrd[] = "zgebrd(intptr_t handle, int m, int n, size_t A, int lda, size_t D, size_t E, size_t tauQ, size_t tauP, size_t Work, int lwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_153zgebrd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_D;
  size_t __pyx_v_E;
  size_t __pyx_v_tauQ;
  size_t __pyx_v_tauP;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgebrd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_D,&__pyx_n_s_E,&__pyx_n_s_tauQ,&__pyx_n_s_tauP,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_devInfo,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 1); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 2); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 3); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 4); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_D)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 5); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_E)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 6); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauQ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 7); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tauP)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 8); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 9); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 10); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, 11); __PYX_ERR(0, 1360, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgebrd") < 0)) __PYX_ERR(0, 1360, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1360, __pyx_L3_error)
    __pyx_v_D = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_D == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_E = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_E == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_tauQ = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_tauQ == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_tauP = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_tauP == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1361, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1362, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgebrd", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1360, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_152zgebrd(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_152zgebrd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_D, size_t __pyx_v_E, size_t __pyx_v_tauQ, size_t __pyx_v_tauP, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgebrd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgebrd(__pyx_v_handle, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_D, __pyx_v_E, __pyx_v_tauQ, __pyx_v_tauP, __pyx_v_Work, __pyx_v_lwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1360, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgebrd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1375
 * 
 * # Singular value decomposition, A = U * Sigma * V^H
 * cpdef int sgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_155sgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1377
 * cpdef int sgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1377, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1378
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1379
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnSgesvd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1378
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1380
 *     with nogil:
 *         status = cusolverDnSgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1381
 *         status = cusolverDnSgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1375
 * 
 * # Singular value decomposition, A = U * Sigma * V^H
 * cpdef int sgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_155sgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_154sgesvd_bufferSize[] = "sgesvd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_155sgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1375, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1375, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvd_bufferSize") < 0)) __PYX_ERR(0, 1375, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1375, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1375, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1375, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1375, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_154sgesvd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_154sgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sgesvd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgesvd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1375, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1383
 *     return lwork
 * 
 * cpdef int dgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_157dgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1385
 * cpdef int dgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1386
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1387
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnDgesvd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1386
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1388
 *     with nogil:
 *         status = cusolverDnDgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1389
 *         status = cusolverDnDgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1383
 *     return lwork
 * 
 * cpdef int dgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_157dgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_156dgesvd_bufferSize[] = "dgesvd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_157dgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1383, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1383, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvd_bufferSize") < 0)) __PYX_ERR(0, 1383, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1383, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1383, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1383, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1383, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_156dgesvd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_156dgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dgesvd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgesvd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1383, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1383, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1391
 *     return lwork
 * 
 * cpdef int cgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_159cgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1393
 * cpdef int cgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1394
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1395
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnCgesvd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1394
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1396
 *     with nogil:
 *         status = cusolverDnCgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1397
 *         status = cusolverDnCgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1391
 *     return lwork
 * 
 * cpdef int cgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_159cgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_158cgesvd_bufferSize[] = "cgesvd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_159cgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1391, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1391, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvd_bufferSize") < 0)) __PYX_ERR(0, 1391, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1391, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_158cgesvd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_158cgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cgesvd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgesvd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1391, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1399
 *     return lwork
 * 
 * cpdef int zgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_161zgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1401
 * cpdef int zgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1402
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1403
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvd_bufferSize(<Handle>handle, m, n, &lwork)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return lwork
 */
        __pyx_v_status = cusolverDnZgesvd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_n, (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1402
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1404
 *     with nogil:
 *         status = cusolverDnZgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1405
 *         status = cusolverDnZgesvd_bufferSize(<Handle>handle, m, n, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1399
 *     return lwork
 * 
 * cpdef int zgesvd_bufferSize(intptr_t handle, int m, int n) except? -1:             # <<<<<<<<<<<<<<
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_161zgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_160zgesvd_bufferSize[] = "zgesvd_bufferSize(intptr_t handle, int m, int n) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_161zgesvd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_n;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_n,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd_bufferSize", 1, 3, 3, 1); __PYX_ERR(0, 1399, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd_bufferSize", 1, 3, 3, 2); __PYX_ERR(0, 1399, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvd_bufferSize") < 0)) __PYX_ERR(0, 1399, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1399, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1399, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1399, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvd_bufferSize", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1399, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_160zgesvd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_n);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_160zgesvd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_n) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zgesvd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgesvd_bufferSize(__pyx_v_handle, __pyx_v_m, __pyx_v_n, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1399, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1407
 *     return lwork
 * 
 * cpdef sgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_163sgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvd(intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvd", 0);

  /* "cupy/cuda/cusolver.pyx":1410
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1411
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1412
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, jobu, jobvt, m, n, <float*>A, lda,
 *             <float*>S, <float*>U, ldu, <float*>VT, ldvt,
 */
        __pyx_v_status = cusolverDnSgesvd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((float *)__pyx_v_U), __pyx_v_ldu, ((float *)__pyx_v_VT), __pyx_v_ldvt, ((float *)__pyx_v_Work), __pyx_v_lwork, ((float *)__pyx_v_rwork), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1411
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1416
 *             <float*>S, <float*>U, ldu, <float*>VT, ldvt,
 *             <float*>Work, lwork, <float*>rwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1416, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1407
 *     return lwork
 * 
 * cpdef sgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_163sgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_162sgesvd[] = "sgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A, int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt, size_t Work, int lwork, size_t rwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_163sgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  char __pyx_v_jobu;
  char __pyx_v_jobvt;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_S;
  size_t __pyx_v_U;
  int __pyx_v_ldu;
  size_t __pyx_v_VT;
  int __pyx_v_ldvt;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_rwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobu,&__pyx_n_s_jobvt,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_VT,&__pyx_n_s_ldvt,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_rwork,&__pyx_n_s_devInfo,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 1); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 2); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 3); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 4); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 5); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 6); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 7); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 8); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 9); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_VT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 10); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 11); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 12); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 13); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 14); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, 15); __PYX_ERR(0, 1407, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvd") < 0)) __PYX_ERR(0, 1407, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L3_error)
    __pyx_v_jobu = __Pyx_PyInt_As_char(values[1]); if (unlikely((__pyx_v_jobu == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L3_error)
    __pyx_v_jobvt = __Pyx_PyInt_As_char(values[2]); if (unlikely((__pyx_v_jobvt == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1407, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_S == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L3_error)
    __pyx_v_U = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_U == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L3_error)
    __pyx_v_VT = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_VT == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L3_error)
    __pyx_v_ldvt = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldvt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1408, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1409, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1409, __pyx_L3_error)
    __pyx_v_rwork = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_rwork == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1409, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1409, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvd", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1407, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_162sgesvd(__pyx_self, __pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_162sgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgesvd(__pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1418
 *     check_status(status)
 * 
 * cpdef dgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_165dgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvd(intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvd", 0);

  /* "cupy/cuda/cusolver.pyx":1421
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1422
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1423
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, jobu, jobvt, m, n, <double*>A, lda,
 *             <double*>S, <double*>U, ldu, <double*>VT, ldvt,
 */
        __pyx_v_status = cusolverDnDgesvd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((double *)__pyx_v_U), __pyx_v_ldu, ((double *)__pyx_v_VT), __pyx_v_ldvt, ((double *)__pyx_v_Work), __pyx_v_lwork, ((double *)__pyx_v_rwork), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1422
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1427
 *             <double*>S, <double*>U, ldu, <double*>VT, ldvt,
 *             <double*>Work, lwork, <double*>rwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1418
 *     check_status(status)
 * 
 * cpdef dgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_165dgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_164dgesvd[] = "dgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A, int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt, size_t Work, int lwork, size_t rwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_165dgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  char __pyx_v_jobu;
  char __pyx_v_jobvt;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_S;
  size_t __pyx_v_U;
  int __pyx_v_ldu;
  size_t __pyx_v_VT;
  int __pyx_v_ldvt;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_rwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobu,&__pyx_n_s_jobvt,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_VT,&__pyx_n_s_ldvt,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_rwork,&__pyx_n_s_devInfo,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 1); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 2); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 3); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 4); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 5); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 6); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 7); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 8); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 9); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_VT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 10); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 11); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 12); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 13); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 14); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, 15); __PYX_ERR(0, 1418, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvd") < 0)) __PYX_ERR(0, 1418, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1418, __pyx_L3_error)
    __pyx_v_jobu = __Pyx_PyInt_As_char(values[1]); if (unlikely((__pyx_v_jobu == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1418, __pyx_L3_error)
    __pyx_v_jobvt = __Pyx_PyInt_As_char(values[2]); if (unlikely((__pyx_v_jobvt == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1418, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1418, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1418, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1418, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1419, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_S == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1419, __pyx_L3_error)
    __pyx_v_U = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_U == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1419, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1419, __pyx_L3_error)
    __pyx_v_VT = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_VT == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1419, __pyx_L3_error)
    __pyx_v_ldvt = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldvt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1419, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1420, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1420, __pyx_L3_error)
    __pyx_v_rwork = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_rwork == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1420, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1420, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvd", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1418, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_164dgesvd(__pyx_self, __pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_164dgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgesvd(__pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1429
 *     check_status(status)
 * 
 * cpdef cgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_167cgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvd(intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvd", 0);

  /* "cupy/cuda/cusolver.pyx":1432
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1432, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1433
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1434
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, jobu, jobvt, m, n, <cuComplex*>A, lda,
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>VT, ldvt,
 */
        __pyx_v_status = cusolverDnCgesvd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((cuComplex *)__pyx_v_U), __pyx_v_ldu, ((cuComplex *)__pyx_v_VT), __pyx_v_ldvt, ((cuComplex *)__pyx_v_Work), __pyx_v_lwork, ((float *)__pyx_v_rwork), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1433
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1438
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>VT, ldvt,
 *             <cuComplex*>Work, lwork, <float*>rwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1429
 *     check_status(status)
 * 
 * cpdef cgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_167cgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_166cgesvd[] = "cgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A, int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt, size_t Work, int lwork, size_t rwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_167cgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  char __pyx_v_jobu;
  char __pyx_v_jobvt;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_S;
  size_t __pyx_v_U;
  int __pyx_v_ldu;
  size_t __pyx_v_VT;
  int __pyx_v_ldvt;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_rwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobu,&__pyx_n_s_jobvt,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_VT,&__pyx_n_s_ldvt,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_rwork,&__pyx_n_s_devInfo,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 1); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 2); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 3); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 4); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 5); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 6); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 7); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 8); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 9); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_VT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 10); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 11); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 12); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 13); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 14); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, 15); __PYX_ERR(0, 1429, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvd") < 0)) __PYX_ERR(0, 1429, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L3_error)
    __pyx_v_jobu = __Pyx_PyInt_As_char(values[1]); if (unlikely((__pyx_v_jobu == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L3_error)
    __pyx_v_jobvt = __Pyx_PyInt_As_char(values[2]); if (unlikely((__pyx_v_jobvt == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1429, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_S == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)
    __pyx_v_U = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_U == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)
    __pyx_v_VT = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_VT == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)
    __pyx_v_ldvt = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldvt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1430, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1431, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1431, __pyx_L3_error)
    __pyx_v_rwork = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_rwork == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1431, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1431, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvd", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1429, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_166cgesvd(__pyx_self, __pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_166cgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgesvd(__pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1440
 *     check_status(status)
 * 
 * cpdef zgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_169zgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvd(intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvd", 0);

  /* "cupy/cuda/cusolver.pyx":1443
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1444
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1445
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, jobu, jobvt, m, n, <cuDoubleComplex*>A, lda,
 *             <double*>S, <cuDoubleComplex*>U, ldu, <cuDoubleComplex*>VT, ldvt,
 */
        __pyx_v_status = cusolverDnZgesvd(((cusolverDnHandle_t)__pyx_v_handle), __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((cuDoubleComplex *)__pyx_v_U), __pyx_v_ldu, ((cuDoubleComplex *)__pyx_v_VT), __pyx_v_ldvt, ((cuDoubleComplex *)__pyx_v_Work), __pyx_v_lwork, ((double *)__pyx_v_rwork), ((int *)__pyx_v_devInfo));
      }

      /* "cupy/cuda/cusolver.pyx":1444
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvd(
 *             <Handle>handle, jobu, jobvt, m, n, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1449
 *             <double*>S, <cuDoubleComplex*>U, ldu, <cuDoubleComplex*>VT, ldvt,
 *             <cuDoubleComplex*>Work, lwork, <double*>rwork, <int*>devInfo)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * # gesvdj ... Singular value decomposition using Jacobi mathod
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1440
 *     check_status(status)
 * 
 * cpdef zgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A,             # <<<<<<<<<<<<<<
 *              int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt,
 *              size_t Work, int lwork, size_t rwork, size_t devInfo):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_169zgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_168zgesvd[] = "zgesvd(intptr_t handle, char jobu, char jobvt, int m, int n, size_t A, int lda, size_t S, size_t U, int ldu, size_t VT, int ldvt, size_t Work, int lwork, size_t rwork, size_t devInfo)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_169zgesvd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  char __pyx_v_jobu;
  char __pyx_v_jobvt;
  int __pyx_v_m;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_S;
  size_t __pyx_v_U;
  int __pyx_v_ldu;
  size_t __pyx_v_VT;
  int __pyx_v_ldvt;
  size_t __pyx_v_Work;
  int __pyx_v_lwork;
  size_t __pyx_v_rwork;
  size_t __pyx_v_devInfo;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobu,&__pyx_n_s_jobvt,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_VT,&__pyx_n_s_ldvt,&__pyx_n_s_Work,&__pyx_n_s_lwork,&__pyx_n_s_rwork,&__pyx_n_s_devInfo,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 1); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 2); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 3); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 4); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 5); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 6); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 7); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 8); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 9); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_VT)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 10); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldvt)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 11); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_Work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 12); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 13); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 14); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_devInfo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, 15); __PYX_ERR(0, 1440, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvd") < 0)) __PYX_ERR(0, 1440, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_jobu = __Pyx_PyInt_As_char(values[1]); if (unlikely((__pyx_v_jobu == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_jobvt = __Pyx_PyInt_As_char(values[2]); if (unlikely((__pyx_v_jobvt == (char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1440, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
    __pyx_v_S = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_S == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
    __pyx_v_U = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_U == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
    __pyx_v_VT = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_VT == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
    __pyx_v_ldvt = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldvt == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1441, __pyx_L3_error)
    __pyx_v_Work = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_Work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L3_error)
    __pyx_v_rwork = __Pyx_PyInt_As_size_t(values[14]); if (unlikely((__pyx_v_rwork == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L3_error)
    __pyx_v_devInfo = __Pyx_PyInt_As_size_t(values[15]); if (unlikely((__pyx_v_devInfo == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1442, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvd", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1440, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_168zgesvd(__pyx_self, __pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_168zgesvd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, char __pyx_v_jobu, char __pyx_v_jobvt, int __pyx_v_m, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_S, size_t __pyx_v_U, int __pyx_v_ldu, size_t __pyx_v_VT, int __pyx_v_ldvt, size_t __pyx_v_Work, int __pyx_v_lwork, size_t __pyx_v_rwork, size_t __pyx_v_devInfo) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgesvd(__pyx_v_handle, __pyx_v_jobu, __pyx_v_jobvt, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_VT, __pyx_v_ldvt, __pyx_v_Work, __pyx_v_lwork, __pyx_v_rwork, __pyx_v_devInfo, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1452
 * 
 * # gesvdj ... Singular value decomposition using Jacobi mathod
 * cpdef intptr_t createGesvdjInfo() except? 0:             # <<<<<<<<<<<<<<
 *     cdef GesvdjInfo info
 *     status = cusolverDnCreateGesvdjInfo(&info)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_171createGesvdjInfo(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static intptr_t __pyx_f_4cupy_4cuda_8cusolver_createGesvdjInfo(CYTHON_UNUSED int __pyx_skip_dispatch) {
  gesvdjInfo_t __pyx_v_info;
  int __pyx_v_status;
  intptr_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("createGesvdjInfo", 0);

  /* "cupy/cuda/cusolver.pyx":1454
 * cpdef intptr_t createGesvdjInfo() except? 0:
 *     cdef GesvdjInfo info
 *     status = cusolverDnCreateGesvdjInfo(&info)             # <<<<<<<<<<<<<<
 *     check_status(status)
 *     return <intptr_t>info
 */
  __pyx_v_status = cusolverDnCreateGesvdjInfo((&__pyx_v_info));

  /* "cupy/cuda/cusolver.pyx":1455
 *     cdef GesvdjInfo info
 *     status = cusolverDnCreateGesvdjInfo(&info)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return <intptr_t>info
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1455, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1456
 *     status = cusolverDnCreateGesvdjInfo(&info)
 *     check_status(status)
 *     return <intptr_t>info             # <<<<<<<<<<<<<<
 * 
 * cpdef destroyGesvdjInfo(intptr_t info):
 */
  __pyx_r = ((intptr_t)__pyx_v_info);
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1452
 * 
 * # gesvdj ... Singular value decomposition using Jacobi mathod
 * cpdef intptr_t createGesvdjInfo() except? 0:             # <<<<<<<<<<<<<<
 *     cdef GesvdjInfo info
 *     status = cusolverDnCreateGesvdjInfo(&info)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.createGesvdjInfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_171createGesvdjInfo(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_170createGesvdjInfo[] = "createGesvdjInfo() -> intptr_t";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_171createGesvdjInfo(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("createGesvdjInfo (wrapper)", 0);
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_170createGesvdjInfo(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_170createGesvdjInfo(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  intptr_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("createGesvdjInfo", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_createGesvdjInfo(0); if (unlikely(__pyx_t_1 == ((intptr_t)0) && PyErr_Occurred())) __PYX_ERR(0, 1452, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.createGesvdjInfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1458
 *     return <intptr_t>info
 * 
 * cpdef destroyGesvdjInfo(intptr_t info):             # <<<<<<<<<<<<<<
 *     status = cusolverDnDestroyGesvdjInfo(<GesvdjInfo>info)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_173destroyGesvdjInfo(PyObject *__pyx_self, PyObject *__pyx_arg_info); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_destroyGesvdjInfo(intptr_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyGesvdjInfo", 0);

  /* "cupy/cuda/cusolver.pyx":1459
 * 
 * cpdef destroyGesvdjInfo(intptr_t info):
 *     status = cusolverDnDestroyGesvdjInfo(<GesvdjInfo>info)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusolverDnDestroyGesvdjInfo(((gesvdjInfo_t)__pyx_v_info));

  /* "cupy/cuda/cusolver.pyx":1460
 * cpdef destroyGesvdjInfo(intptr_t info):
 *     status = cusolverDnDestroyGesvdjInfo(<GesvdjInfo>info)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef xgesvdjSetTolerance(intptr_t info, double tolerance):
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1458
 *     return <intptr_t>info
 * 
 * cpdef destroyGesvdjInfo(intptr_t info):             # <<<<<<<<<<<<<<
 *     status = cusolverDnDestroyGesvdjInfo(<GesvdjInfo>info)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.destroyGesvdjInfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_173destroyGesvdjInfo(PyObject *__pyx_self, PyObject *__pyx_arg_info); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_172destroyGesvdjInfo[] = "destroyGesvdjInfo(intptr_t info)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_173destroyGesvdjInfo(PyObject *__pyx_self, PyObject *__pyx_arg_info) {
  intptr_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("destroyGesvdjInfo (wrapper)", 0);
  assert(__pyx_arg_info); {
    __pyx_v_info = PyInt_AsSsize_t(__pyx_arg_info); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1458, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.destroyGesvdjInfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_172destroyGesvdjInfo(__pyx_self, ((intptr_t)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_172destroyGesvdjInfo(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("destroyGesvdjInfo", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_destroyGesvdjInfo(__pyx_v_info, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1458, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.destroyGesvdjInfo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1462
 *     check_status(status)
 * 
 * cpdef xgesvdjSetTolerance(intptr_t info, double tolerance):             # <<<<<<<<<<<<<<
 *     status = cusolverDnXgesvdjSetTolerance(<GesvdjInfo>info, tolerance)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_175xgesvdjSetTolerance(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetTolerance(intptr_t __pyx_v_info, double __pyx_v_tolerance, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjSetTolerance", 0);

  /* "cupy/cuda/cusolver.pyx":1463
 * 
 * cpdef xgesvdjSetTolerance(intptr_t info, double tolerance):
 *     status = cusolverDnXgesvdjSetTolerance(<GesvdjInfo>info, tolerance)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusolverDnXgesvdjSetTolerance(((gesvdjInfo_t)__pyx_v_info), __pyx_v_tolerance);

  /* "cupy/cuda/cusolver.pyx":1464
 * cpdef xgesvdjSetTolerance(intptr_t info, double tolerance):
 *     status = cusolverDnXgesvdjSetTolerance(<GesvdjInfo>info, tolerance)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef xgesvdjSetMaxSweeps(intptr_t info, int max_sweeps):
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1462
 *     check_status(status)
 * 
 * cpdef xgesvdjSetTolerance(intptr_t info, double tolerance):             # <<<<<<<<<<<<<<
 *     status = cusolverDnXgesvdjSetTolerance(<GesvdjInfo>info, tolerance)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetTolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_175xgesvdjSetTolerance(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_174xgesvdjSetTolerance[] = "xgesvdjSetTolerance(intptr_t info, double tolerance)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_175xgesvdjSetTolerance(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_info;
  double __pyx_v_tolerance;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xgesvdjSetTolerance (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_info,&__pyx_n_s_tolerance,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tolerance)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xgesvdjSetTolerance", 1, 2, 2, 1); __PYX_ERR(0, 1462, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xgesvdjSetTolerance") < 0)) __PYX_ERR(0, 1462, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_info = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1462, __pyx_L3_error)
    __pyx_v_tolerance = __pyx_PyFloat_AsDouble(values[1]); if (unlikely((__pyx_v_tolerance == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1462, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xgesvdjSetTolerance", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1462, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetTolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_174xgesvdjSetTolerance(__pyx_self, __pyx_v_info, __pyx_v_tolerance);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_174xgesvdjSetTolerance(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info, double __pyx_v_tolerance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjSetTolerance", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetTolerance(__pyx_v_info, __pyx_v_tolerance, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetTolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1466
 *     check_status(status)
 * 
 * cpdef xgesvdjSetMaxSweeps(intptr_t info, int max_sweeps):             # <<<<<<<<<<<<<<
 *     status = cusolverDnXgesvdjSetMaxSweeps(<GesvdjInfo>info, max_sweeps)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_177xgesvdjSetMaxSweeps(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetMaxSweeps(intptr_t __pyx_v_info, int __pyx_v_max_sweeps, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjSetMaxSweeps", 0);

  /* "cupy/cuda/cusolver.pyx":1467
 * 
 * cpdef xgesvdjSetMaxSweeps(intptr_t info, int max_sweeps):
 *     status = cusolverDnXgesvdjSetMaxSweeps(<GesvdjInfo>info, max_sweeps)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusolverDnXgesvdjSetMaxSweeps(((gesvdjInfo_t)__pyx_v_info), __pyx_v_max_sweeps);

  /* "cupy/cuda/cusolver.pyx":1468
 * cpdef xgesvdjSetMaxSweeps(intptr_t info, int max_sweeps):
 *     status = cusolverDnXgesvdjSetMaxSweeps(<GesvdjInfo>info, max_sweeps)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef xgesvdjSetSortEig(intptr_t info, int sort_svd):
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1466
 *     check_status(status)
 * 
 * cpdef xgesvdjSetMaxSweeps(intptr_t info, int max_sweeps):             # <<<<<<<<<<<<<<
 *     status = cusolverDnXgesvdjSetMaxSweeps(<GesvdjInfo>info, max_sweeps)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetMaxSweeps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_177xgesvdjSetMaxSweeps(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_176xgesvdjSetMaxSweeps[] = "xgesvdjSetMaxSweeps(intptr_t info, int max_sweeps)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_177xgesvdjSetMaxSweeps(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_info;
  int __pyx_v_max_sweeps;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xgesvdjSetMaxSweeps (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_info,&__pyx_n_s_max_sweeps,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_max_sweeps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xgesvdjSetMaxSweeps", 1, 2, 2, 1); __PYX_ERR(0, 1466, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xgesvdjSetMaxSweeps") < 0)) __PYX_ERR(0, 1466, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_info = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
    __pyx_v_max_sweeps = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_max_sweeps == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1466, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xgesvdjSetMaxSweeps", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1466, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetMaxSweeps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_176xgesvdjSetMaxSweeps(__pyx_self, __pyx_v_info, __pyx_v_max_sweeps);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_176xgesvdjSetMaxSweeps(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info, int __pyx_v_max_sweeps) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjSetMaxSweeps", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetMaxSweeps(__pyx_v_info, __pyx_v_max_sweeps, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetMaxSweeps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1470
 *     check_status(status)
 * 
 * cpdef xgesvdjSetSortEig(intptr_t info, int sort_svd):             # <<<<<<<<<<<<<<
 *     status = cusolverDnXgesvdjSetSortEig(<GesvdjInfo>info, sort_svd)
 *     check_status(status)
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_179xgesvdjSetSortEig(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetSortEig(intptr_t __pyx_v_info, int __pyx_v_sort_svd, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjSetSortEig", 0);

  /* "cupy/cuda/cusolver.pyx":1471
 * 
 * cpdef xgesvdjSetSortEig(intptr_t info, int sort_svd):
 *     status = cusolverDnXgesvdjSetSortEig(<GesvdjInfo>info, sort_svd)             # <<<<<<<<<<<<<<
 *     check_status(status)
 * 
 */
  __pyx_v_status = cusolverDnXgesvdjSetSortEig(((gesvdjInfo_t)__pyx_v_info), __pyx_v_sort_svd);

  /* "cupy/cuda/cusolver.pyx":1472
 * cpdef xgesvdjSetSortEig(intptr_t info, int sort_svd):
 *     status = cusolverDnXgesvdjSetSortEig(<GesvdjInfo>info, sort_svd)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef double xgesvdjGetResidual(intptr_t handle, intptr_t info):
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1470
 *     check_status(status)
 * 
 * cpdef xgesvdjSetSortEig(intptr_t info, int sort_svd):             # <<<<<<<<<<<<<<
 *     status = cusolverDnXgesvdjSetSortEig(<GesvdjInfo>info, sort_svd)
 *     check_status(status)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetSortEig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_179xgesvdjSetSortEig(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_178xgesvdjSetSortEig[] = "xgesvdjSetSortEig(intptr_t info, int sort_svd)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_179xgesvdjSetSortEig(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_info;
  int __pyx_v_sort_svd;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xgesvdjSetSortEig (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_info,&__pyx_n_s_sort_svd,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_sort_svd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xgesvdjSetSortEig", 1, 2, 2, 1); __PYX_ERR(0, 1470, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xgesvdjSetSortEig") < 0)) __PYX_ERR(0, 1470, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_info = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
    __pyx_v_sort_svd = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_sort_svd == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1470, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xgesvdjSetSortEig", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1470, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetSortEig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_178xgesvdjSetSortEig(__pyx_self, __pyx_v_info, __pyx_v_sort_svd);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_178xgesvdjSetSortEig(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_info, int __pyx_v_sort_svd) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjSetSortEig", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetSortEig(__pyx_v_info, __pyx_v_sort_svd, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjSetSortEig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1474
 *     check_status(status)
 * 
 * cpdef double xgesvdjGetResidual(intptr_t handle, intptr_t info):             # <<<<<<<<<<<<<<
 *     cdef double residual
 *     status = cusolverDnXgesvdjGetResidual(<Handle>handle, <GesvdjInfo>info,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_181xgesvdjGetResidual(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static double __pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetResidual(intptr_t __pyx_v_handle, intptr_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  double __pyx_v_residual;
  int __pyx_v_status;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjGetResidual", 0);

  /* "cupy/cuda/cusolver.pyx":1476
 * cpdef double xgesvdjGetResidual(intptr_t handle, intptr_t info):
 *     cdef double residual
 *     status = cusolverDnXgesvdjGetResidual(<Handle>handle, <GesvdjInfo>info,             # <<<<<<<<<<<<<<
 *                                           &residual)
 *     check_status(status)
 */
  __pyx_v_status = cusolverDnXgesvdjGetResidual(((cusolverDnHandle_t)__pyx_v_handle), ((gesvdjInfo_t)__pyx_v_info), (&__pyx_v_residual));

  /* "cupy/cuda/cusolver.pyx":1478
 *     status = cusolverDnXgesvdjGetResidual(<Handle>handle, <GesvdjInfo>info,
 *                                           &residual)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return residual
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1479
 *                                           &residual)
 *     check_status(status)
 *     return residual             # <<<<<<<<<<<<<<
 * 
 * cpdef int xgesvdjGetSweeps(intptr_t handle, intptr_t info):
 */
  __pyx_r = __pyx_v_residual;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1474
 *     check_status(status)
 * 
 * cpdef double xgesvdjGetResidual(intptr_t handle, intptr_t info):             # <<<<<<<<<<<<<<
 *     cdef double residual
 *     status = cusolverDnXgesvdjGetResidual(<Handle>handle, <GesvdjInfo>info,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.xgesvdjGetResidual", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_181xgesvdjGetResidual(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_180xgesvdjGetResidual[] = "xgesvdjGetResidual(intptr_t handle, intptr_t info) -> double";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_181xgesvdjGetResidual(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  intptr_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xgesvdjGetResidual (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_info,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xgesvdjGetResidual", 1, 2, 2, 1); __PYX_ERR(0, 1474, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xgesvdjGetResidual") < 0)) __PYX_ERR(0, 1474, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1474, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[1]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1474, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xgesvdjGetResidual", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1474, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjGetResidual", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_180xgesvdjGetResidual(__pyx_self, __pyx_v_handle, __pyx_v_info);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_180xgesvdjGetResidual(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, intptr_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjGetResidual", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetResidual(__pyx_v_handle, __pyx_v_info, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjGetResidual", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1481
 *     return residual
 * 
 * cpdef int xgesvdjGetSweeps(intptr_t handle, intptr_t info):             # <<<<<<<<<<<<<<
 *     cdef int executed_sweeps
 *     status = cusolverDnXgesvdjGetSweeps(<Handle>handle, <GesvdjInfo>info,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_183xgesvdjGetSweeps(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetSweeps(intptr_t __pyx_v_handle, intptr_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_executed_sweeps;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjGetSweeps", 0);

  /* "cupy/cuda/cusolver.pyx":1483
 * cpdef int xgesvdjGetSweeps(intptr_t handle, intptr_t info):
 *     cdef int executed_sweeps
 *     status = cusolverDnXgesvdjGetSweeps(<Handle>handle, <GesvdjInfo>info,             # <<<<<<<<<<<<<<
 *                                         &executed_sweeps)
 *     check_status(status)
 */
  __pyx_v_status = cusolverDnXgesvdjGetSweeps(((cusolverDnHandle_t)__pyx_v_handle), ((gesvdjInfo_t)__pyx_v_info), (&__pyx_v_executed_sweeps));

  /* "cupy/cuda/cusolver.pyx":1485
 *     status = cusolverDnXgesvdjGetSweeps(<Handle>handle, <GesvdjInfo>info,
 *                                         &executed_sweeps)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return executed_sweeps
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1485, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1486
 *                                         &executed_sweeps)
 *     check_status(status)
 *     return executed_sweeps             # <<<<<<<<<<<<<<
 * 
 * cpdef int sgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,
 */
  __pyx_r = __pyx_v_executed_sweeps;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1481
 *     return residual
 * 
 * cpdef int xgesvdjGetSweeps(intptr_t handle, intptr_t info):             # <<<<<<<<<<<<<<
 *     cdef int executed_sweeps
 *     status = cusolverDnXgesvdjGetSweeps(<Handle>handle, <GesvdjInfo>info,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.xgesvdjGetSweeps", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_183xgesvdjGetSweeps(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_182xgesvdjGetSweeps[] = "xgesvdjGetSweeps(intptr_t handle, intptr_t info) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_183xgesvdjGetSweeps(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  intptr_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("xgesvdjGetSweeps (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_info,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("xgesvdjGetSweeps", 1, 2, 2, 1); __PYX_ERR(0, 1481, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "xgesvdjGetSweeps") < 0)) __PYX_ERR(0, 1481, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1481, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[1]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1481, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("xgesvdjGetSweeps", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1481, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjGetSweeps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_182xgesvdjGetSweeps(__pyx_self, __pyx_v_handle, __pyx_v_info);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_182xgesvdjGetSweeps(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, intptr_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("xgesvdjGetSweeps", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetSweeps(__pyx_v_handle, __pyx_v_info, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.xgesvdjGetSweeps", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1488
 *     return executed_sweeps
 * 
 * cpdef int sgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_185sgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvdj_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdj_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1492
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvdj_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1493
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1494
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvdj_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const float*>A, lda,
 *             <const float*>S, <const float*>U, ldu, <const float*>V, ldv,
 */
        __pyx_v_status = cusolverDnSgesvdj_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((float const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_S), ((float const *)__pyx_v_U), __pyx_v_ldu, ((float const *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1493
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1498
 *             <const float*>S, <const float*>U, ldu, <const float*>V, ldv,
 *             &lwork, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1499
 *             &lwork, <GesvdjInfo>params)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1488
 *     return executed_sweeps
 * 
 * cpdef int sgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.sgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_185sgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_184sgesvdj_bufferSize[] = "sgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_185sgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvdj_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1488, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvdj_bufferSize") < 0)) __PYX_ERR(0, 1488, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1488, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1489, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1489, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1489, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1489, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1490, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1490, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1490, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1490, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvdj_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1488, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_184sgesvdj_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_184sgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdj_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_sgesvdj_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1501
 *     return lwork
 * 
 * cpdef int dgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_187dgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvdj_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdj_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1505
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvdj_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1506
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1507
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvdj_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const double*>A, lda,
 *             <const double*>S, <const double*>U, ldu, <const double*>V, ldv,
 */
        __pyx_v_status = cusolverDnDgesvdj_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((double const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_S), ((double const *)__pyx_v_U), __pyx_v_ldu, ((double const *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1506
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1511
 *             <const double*>S, <const double*>U, ldu, <const double*>V, ldv,
 *             &lwork, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1512
 *             &lwork, <GesvdjInfo>params)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1501
 *     return lwork
 * 
 * cpdef int dgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.dgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_187dgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_186dgesvdj_bufferSize[] = "dgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_187dgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvdj_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1501, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvdj_bufferSize") < 0)) __PYX_ERR(0, 1501, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1501, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1501, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1501, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1501, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1501, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1502, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1503, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvdj_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1501, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_186dgesvdj_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_186dgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdj_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_dgesvdj_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1514
 *     return lwork
 * 
 * cpdef int cgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_189cgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvdj_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdj_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1518
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvdj_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1519
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const cuComplex*>A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1520
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvdj_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const cuComplex*>A,
 *             lda, <const float*>S, <const cuComplex*>U, ldu,
 */
        __pyx_v_status = cusolverDnCgesvdj_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_S), ((cuComplex const *)__pyx_v_U), __pyx_v_ldu, ((cuComplex const *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1519
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <const cuComplex*>A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1524
 *             lda, <const float*>S, <const cuComplex*>U, ldu,
 *             <const cuComplex*>V, ldv, &lwork, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1525
 *             <const cuComplex*>V, ldv, &lwork, <GesvdjInfo>params)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1514
 *     return lwork
 * 
 * cpdef int cgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.cgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_189cgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_188cgesvdj_bufferSize[] = "cgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_189cgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvdj_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1514, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvdj_bufferSize") < 0)) __PYX_ERR(0, 1514, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1514, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1514, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1514, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1514, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1514, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1515, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1515, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1515, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1515, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1516, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1516, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1516, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1516, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvdj_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1514, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_188cgesvdj_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_188cgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdj_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_cgesvdj_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1514, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1527
 *     return lwork
 * 
 * cpdef int zgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_191zgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvdj_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdj_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1531
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvdj_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1532
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1533
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvdj_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, econ, m, n,
 *             <const cuDoubleComplex*>A, lda, <const double*>S,
 */
        __pyx_v_status = cusolverDnZgesvdj_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_S), ((cuDoubleComplex const *)__pyx_v_U), __pyx_v_ldu, ((cuDoubleComplex const *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1532
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdj_bufferSize(
 *             <Handle>handle, <EigMode>jobz, econ, m, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1538
 *             <const cuDoubleComplex*>U, ldu, <const cuDoubleComplex*>V,
 *             ldv, &lwork, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1539
 *             ldv, &lwork, <GesvdjInfo>params)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1527
 *     return lwork
 * 
 * cpdef int zgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n,             # <<<<<<<<<<<<<<
 *                              intptr_t A, int lda, intptr_t S, intptr_t U,
 *                              int ldu, intptr_t V, int ldv, intptr_t params):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.zgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_191zgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_190zgesvdj_bufferSize[] = "zgesvdj_bufferSize(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_191zgesvdj_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvdj_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1527, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvdj_bufferSize") < 0)) __PYX_ERR(0, 1527, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1527, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1528, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1528, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1528, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1528, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1529, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1529, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1529, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1529, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvdj_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1527, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_190zgesvdj_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_190zgesvdj_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdj_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_zgesvdj_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdj_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1541
 *     return lwork
 * 
 * cpdef sgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_193sgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvdj(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdj", 0);

  /* "cupy/cuda/cusolver.pyx":1544
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1545
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,
 *                                    <float*>A, lda, <float*>S, <float*>U, ldu,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1546
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,             # <<<<<<<<<<<<<<
 *                                    <float*>A, lda, <float*>S, <float*>U, ldu,
 *                                    <float*>V, ldv, <float*>work, lwork,
 */
        __pyx_v_status = cusolverDnSgesvdj(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((float *)__pyx_v_U), __pyx_v_ldu, ((float *)__pyx_v_V), __pyx_v_ldv, ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1545
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,
 *                                    <float*>A, lda, <float*>S, <float*>U, ldu,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1550
 *                                    <float*>V, ldv, <float*>work, lwork,
 *                                    <int*>info, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1541
 *     return lwork
 * 
 * cpdef sgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_193sgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_192sgesvdj[] = "sgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_193sgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvdj (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 1); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 2); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 3); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 4); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 5); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 6); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 7); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 8); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 9); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 10); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 11); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 12); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 13); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 14); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, 15); __PYX_ERR(0, 1541, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvdj") < 0)) __PYX_ERR(0, 1541, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1541, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1541, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1541, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1541, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1541, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1541, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1542, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1543, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1543, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1543, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[15]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1543, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvdj", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1541, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_192sgesvdj(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_192sgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdj", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgesvdj(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1552
 *     check_status(status)
 * 
 * cpdef dgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_195dgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvdj(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdj", 0);

  /* "cupy/cuda/cusolver.pyx":1555
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1555, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1556
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,
 *                                    <double*>A, lda, <double*>S, <double*>U,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1557
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,             # <<<<<<<<<<<<<<
 *                                    <double*>A, lda, <double*>S, <double*>U,
 *                                    ldu, <double*>V, ldv, <double*>work, lwork,
 */
        __pyx_v_status = cusolverDnDgesvdj(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((double *)__pyx_v_U), __pyx_v_ldu, ((double *)__pyx_v_V), __pyx_v_ldv, ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1556
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdj(<Handle>handle, <EigMode>jobz, econ, m, n,
 *                                    <double*>A, lda, <double*>S, <double*>U,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1561
 *                                    ldu, <double*>V, ldv, <double*>work, lwork,
 *                                    <int*>info, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1552
 *     check_status(status)
 * 
 * cpdef dgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_195dgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_194dgesvdj[] = "dgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_195dgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvdj (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 1); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 2); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 3); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 4); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 5); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 6); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 7); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 8); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 9); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 10); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 11); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 12); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 13); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 14); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, 15); __PYX_ERR(0, 1552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvdj") < 0)) __PYX_ERR(0, 1552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1553, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[15]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1554, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvdj", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_194dgesvdj(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_194dgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdj", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgesvdj(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1563
 *     check_status(status)
 * 
 * cpdef cgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_197cgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvdj(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdj", 0);

  /* "cupy/cuda/cusolver.pyx":1566
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvdj(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1567
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdj(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1568
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvdj(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <cuComplex*>A, lda,
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>V, ldv,
 */
        __pyx_v_status = cusolverDnCgesvdj(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((cuComplex *)__pyx_v_U), __pyx_v_ldu, ((cuComplex *)__pyx_v_V), __pyx_v_ldv, ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1567
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdj(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1572
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>V, ldv,
 *             <cuComplex*>work, lwork, <int*>info, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1563
 *     check_status(status)
 * 
 * cpdef cgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_197cgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_196cgesvdj[] = "cgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_197cgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvdj (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 1); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 2); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 3); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 4); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 5); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 6); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 7); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 8); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 9); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 10); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 11); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 12); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 13); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 14); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, 15); __PYX_ERR(0, 1563, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvdj") < 0)) __PYX_ERR(0, 1563, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1563, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1564, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1564, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1564, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1564, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1564, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1564, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1565, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1565, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1565, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[15]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1565, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvdj", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1563, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_196cgesvdj(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_196cgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdj", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgesvdj(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1574
 *     check_status(status)
 * 
 * cpdef zgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_199zgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvdj(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdj", 0);

  /* "cupy/cuda/cusolver.pyx":1577
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvdj(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1578
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdj(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <cuDoubleComplex*>A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1579
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvdj(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <cuDoubleComplex*>A,
 *             lda, <double*>S, <cuDoubleComplex*>U, ldu, <cuDoubleComplex*>V,
 */
        __pyx_v_status = cusolverDnZgesvdj(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_econ, __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((cuDoubleComplex *)__pyx_v_U), __pyx_v_ldu, ((cuDoubleComplex *)__pyx_v_V), __pyx_v_ldv, ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params));
      }

      /* "cupy/cuda/cusolver.pyx":1578
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdj(
 *             <Handle>handle, <EigMode>jobz, econ, m, n, <cuDoubleComplex*>A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1583
 *             lda, <double*>S, <cuDoubleComplex*>U, ldu, <cuDoubleComplex*>V,
 *             ldv, <cuDoubleComplex*>work, lwork, <int*>info, <GesvdjInfo>params)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef int sgesvdjBatched_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1574
 *     check_status(status)
 * 
 * cpdef zgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A,             # <<<<<<<<<<<<<<
 *               int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 *               intptr_t work, int lwork, intptr_t info, intptr_t params):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_199zgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_198zgesvdj[] = "zgesvdj(intptr_t handle, int jobz, int econ, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_199zgesvdj(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_econ;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvdj (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_econ,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 1); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_econ)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 2); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 3); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 4); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 5); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 6); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 7); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 8); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 9); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 10); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 11); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 12); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 13); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 14); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, 15); __PYX_ERR(0, 1574, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvdj") < 0)) __PYX_ERR(0, 1574, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1574, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1574, __pyx_L3_error)
    __pyx_v_econ = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_econ == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1574, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1574, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1574, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1574, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1575, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1575, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1575, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1575, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1575, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1575, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[12]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1576, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[13]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1576, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1576, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[15]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1576, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvdj", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1574, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_198zgesvdj(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_198zgesvdj(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_econ, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdj", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgesvdj(__pyx_v_handle, __pyx_v_jobz, __pyx_v_econ, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdj", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1585
 *     check_status(status)
 * 
 * cpdef int sgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_201sgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvdjBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdjBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1590
 *         intptr_t params, int batchSize) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvdjBatched_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1591
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1592
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <float*>A, lda,
 *             <float*>S, <float*>U, ldu, <float*>V, ldv, &lwork,
 */
        __pyx_v_status = cusolverDnSgesvdjBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((float *)__pyx_v_U), __pyx_v_ldu, ((float *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1591
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1596
 *             <float*>S, <float*>U, ldu, <float*>V, ldv, &lwork,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1596, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1597
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgesvdjBatched_bufferSize(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1585
 *     check_status(status)
 * 
 * cpdef int sgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_201sgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_200sgesvdjBatched_bufferSize[] = "sgesvdjBatched_bufferSize(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_201sgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvdjBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1585, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvdjBatched_bufferSize") < 0)) __PYX_ERR(0, 1585, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1586, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1586, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1586, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1586, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1586, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1587, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1588, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1588, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvdjBatched_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1585, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_200sgesvdjBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_200sgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("sgesvdjBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgesvdjBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1585, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1599
 *     return lwork
 * 
 * cpdef int dgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_203dgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvdjBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdjBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1604
 *         intptr_t params, int batchSize) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvdjBatched_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1605
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1606
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <double*>A, lda,
 *             <double*>S, <double*>U, ldu, <double*>V, ldv, &lwork,
 */
        __pyx_v_status = cusolverDnDgesvdjBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((double *)__pyx_v_U), __pyx_v_ldu, ((double *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1605
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1610
 *             <double*>S, <double*>U, ldu, <double*>V, ldv, &lwork,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1611
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgesvdjBatched_bufferSize(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1599
 *     return lwork
 * 
 * cpdef int dgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_203dgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_202dgesvdjBatched_bufferSize[] = "dgesvdjBatched_bufferSize(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_203dgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvdjBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1599, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvdjBatched_bufferSize") < 0)) __PYX_ERR(0, 1599, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1600, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1601, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1602, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1602, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvdjBatched_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1599, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_202dgesvdjBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_202dgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dgesvdjBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgesvdjBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1599, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1613
 *     return lwork
 * 
 * cpdef int cgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_205cgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvdjBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdjBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1618
 *         intptr_t params, int batchSize) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvdjBatched_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1618, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1619
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1620
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <cuComplex*>A, lda,
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>V, ldv, &lwork,
 */
        __pyx_v_status = cusolverDnCgesvdjBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((cuComplex *)__pyx_v_U), __pyx_v_ldu, ((cuComplex *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1619
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1624
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>V, ldv, &lwork,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1625
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgesvdjBatched_bufferSize(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1613
 *     return lwork
 * 
 * cpdef int cgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_205cgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_204cgesvdjBatched_bufferSize[] = "cgesvdjBatched_bufferSize(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_205cgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvdjBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1613, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvdjBatched_bufferSize") < 0)) __PYX_ERR(0, 1613, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1614, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1614, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1614, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1614, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1614, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1615, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1615, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1615, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1615, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1615, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1615, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1616, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1616, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvdjBatched_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1613, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_204cgesvdjBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_204cgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cgesvdjBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgesvdjBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1613, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1627
 *     return lwork
 * 
 * cpdef int zgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_207zgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvdjBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdjBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1632
 *         intptr_t params, int batchSize) except? -1:
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvdjBatched_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1632, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1633
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1634
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <cuDoubleComplex*>A, lda,
 *             <double*>S, <cuDoubleComplex*>U, ldu, <cuDoubleComplex*>V, ldv,
 */
        __pyx_v_status = cusolverDnZgesvdjBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((cuDoubleComplex *)__pyx_v_U), __pyx_v_ldu, ((cuDoubleComplex *)__pyx_v_V), __pyx_v_ldv, (&__pyx_v_lwork), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1633
 *     cdef int lwork
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdjBatched_bufferSize(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1639
 *             &lwork,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1639, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1640
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgesvdjBatched(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1627
 *     return lwork
 * 
 * cpdef int zgesvdjBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_207zgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_206zgesvdjBatched_bufferSize[] = "zgesvdjBatched_bufferSize(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t params, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_207zgesvdjBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvdjBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 1); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 2); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 3); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 4); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 5); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 6); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 7); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 8); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 9); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 10); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 11); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, 12); __PYX_ERR(0, 1627, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvdjBatched_bufferSize") < 0)) __PYX_ERR(0, 1627, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1628, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1629, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1630, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1630, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvdjBatched_bufferSize", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1627, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_206zgesvdjBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_206zgesvdjBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zgesvdjBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgesvdjBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1627, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1627, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdjBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1642
 *     return lwork
 * 
 * cpdef sgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_209sgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvdjBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdjBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1647
 *         intptr_t work, int lwork, intptr_t info,
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1648
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <float*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1649
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvdjBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <float*>A, lda,
 *             <float*>S, <float*>U, ldu, <float*>V, ldv,
 */
        __pyx_v_status = cusolverDnSgesvdjBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((float *)__pyx_v_U), __pyx_v_ldu, ((float *)__pyx_v_V), __pyx_v_ldv, ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1648
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <float*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1654
 *             <float*>work, lwork, <int*>info,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1654, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1642
 *     return lwork
 * 
 * cpdef sgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_209sgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_208sgesvdjBatched[] = "sgesvdjBatched(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_209sgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvdjBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 1); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 2); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 3); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 4); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 5); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 6); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 7); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 8); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 9); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 10); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 11); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 12); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 13); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 14); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, 15); __PYX_ERR(0, 1642, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvdjBatched") < 0)) __PYX_ERR(0, 1642, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1644, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1646, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[15]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1646, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvdjBatched", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1642, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_208sgesvdjBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_208sgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdjBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgesvdjBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1656
 *     check_status(status)
 * 
 * cpdef dgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_211dgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvdjBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdjBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1661
 *         intptr_t work, int lwork, intptr_t info,
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1661, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1662
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <double*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1663
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvdjBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <double*>A, lda,
 *             <double*>S, <double*>U, ldu, <double*>V, ldv,
 */
        __pyx_v_status = cusolverDnDgesvdjBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((double *)__pyx_v_U), __pyx_v_ldu, ((double *)__pyx_v_V), __pyx_v_ldv, ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1662
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <double*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1668
 *             <double*>work, lwork, <int*>info,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1656
 *     check_status(status)
 * 
 * cpdef dgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_211dgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_210dgesvdjBatched[] = "dgesvdjBatched(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_211dgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvdjBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 1); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 2); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 3); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 4); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 5); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 6); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 7); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 8); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 9); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 10); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 11); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 12); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 13); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 14); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, 15); __PYX_ERR(0, 1656, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvdjBatched") < 0)) __PYX_ERR(0, 1656, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1657, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1657, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1657, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1657, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1657, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1658, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1658, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1658, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1658, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1658, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1658, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1659, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1659, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1659, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1660, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[15]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1660, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvdjBatched", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1656, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_210dgesvdjBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_210dgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdjBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgesvdjBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1670
 *     check_status(status)
 * 
 * cpdef cgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_213cgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvdjBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdjBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1675
 *         intptr_t work, int lwork, intptr_t info,
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1676
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1677
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvdjBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <cuComplex*>A, lda,
 *             <float*>S, <cuComplex*>U, ldu, <cuComplex*>V, ldv,
 */
        __pyx_v_status = cusolverDnCgesvdjBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_S), ((cuComplex *)__pyx_v_U), __pyx_v_ldu, ((cuComplex *)__pyx_v_V), __pyx_v_ldv, ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1676
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1682
 *             <cuComplex*>work, lwork, <int*>info,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1670
 *     check_status(status)
 * 
 * cpdef cgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_213cgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_212cgesvdjBatched[] = "cgesvdjBatched(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_213cgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvdjBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 1); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 2); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 3); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 4); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 5); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 6); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 7); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 8); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 9); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 10); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 11); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 12); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 13); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 14); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, 15); __PYX_ERR(0, 1670, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvdjBatched") < 0)) __PYX_ERR(0, 1670, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1672, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1672, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1672, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1672, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1672, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1672, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1673, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1673, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1673, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1674, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[15]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1674, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvdjBatched", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1670, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_212cgesvdjBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_212cgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdjBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgesvdjBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1684
 *     check_status(status)
 * 
 * cpdef zgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_215zgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvdjBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdjBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1689
 *         intptr_t work, int lwork, intptr_t info,
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvdjBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1690
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuDoubleComplex*>A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1691
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvdjBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, m, n, <cuDoubleComplex*>A, lda,
 *             <double*>S, <cuDoubleComplex*>U, ldu, <cuDoubleComplex*>V, ldv,
 */
        __pyx_v_status = cusolverDnZgesvdjBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_m, __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_S), ((cuDoubleComplex *)__pyx_v_U), __pyx_v_ldu, ((cuDoubleComplex *)__pyx_v_V), __pyx_v_ldv, ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info), ((gesvdjInfo_t)__pyx_v_params), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1690
 *         intptr_t params, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdjBatched(
 *             <Handle>handle, <EigMode>jobz, m, n, <cuDoubleComplex*>A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1696
 *             <cuDoubleComplex*>work, lwork, <int*>info,
 *             <GesvdjInfo>params, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * # gesvda ... Approximate singular value decomposition
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1684
 *     check_status(status)
 * 
 * cpdef zgesvdjBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int m, int n, intptr_t A,
 *         int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_215zgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_214zgesvdjBatched[] = "zgesvdjBatched(intptr_t handle, int jobz, int m, int n, intptr_t A, int lda, intptr_t S, intptr_t U, int ldu, intptr_t V, int ldv, intptr_t work, int lwork, intptr_t info, intptr_t params, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_215zgesvdjBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_A;
  int __pyx_v_lda;
  intptr_t __pyx_v_S;
  intptr_t __pyx_v_U;
  int __pyx_v_ldu;
  intptr_t __pyx_v_V;
  int __pyx_v_ldv;
  intptr_t __pyx_v_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_info;
  intptr_t __pyx_v_params;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvdjBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_S,&__pyx_n_s_U,&__pyx_n_s_ldu,&__pyx_n_s_V,&__pyx_n_s_ldv,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,&__pyx_n_s_params,&__pyx_n_s_batchSize,0};
    PyObject* values[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 1); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 2); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 3); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 4); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 5); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 6); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 7); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 8); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 9); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 10); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 11); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 12); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 13); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 14); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, 15); __PYX_ERR(0, 1684, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvdjBatched") < 0)) __PYX_ERR(0, 1684, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 16) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1685, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1685, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1685, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1685, __pyx_L3_error)
    __pyx_v_A = PyInt_AsSsize_t(values[4]); if (unlikely((__pyx_v_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1685, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1686, __pyx_L3_error)
    __pyx_v_S = PyInt_AsSsize_t(values[6]); if (unlikely((__pyx_v_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1686, __pyx_L3_error)
    __pyx_v_U = PyInt_AsSsize_t(values[7]); if (unlikely((__pyx_v_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1686, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1686, __pyx_L3_error)
    __pyx_v_V = PyInt_AsSsize_t(values[9]); if (unlikely((__pyx_v_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1686, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[10]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1686, __pyx_L3_error)
    __pyx_v_work = PyInt_AsSsize_t(values[11]); if (unlikely((__pyx_v_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1687, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[12]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1687, __pyx_L3_error)
    __pyx_v_info = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1687, __pyx_L3_error)
    __pyx_v_params = PyInt_AsSsize_t(values[14]); if (unlikely((__pyx_v_params == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1688, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[15]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1688, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvdjBatched", 1, 16, 16, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1684, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_214zgesvdjBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_214zgesvdjBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_A, int __pyx_v_lda, intptr_t __pyx_v_S, intptr_t __pyx_v_U, int __pyx_v_ldu, intptr_t __pyx_v_V, int __pyx_v_ldv, intptr_t __pyx_v_work, int __pyx_v_lwork, intptr_t __pyx_v_info, intptr_t __pyx_v_params, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdjBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgesvdjBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_m, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_S, __pyx_v_U, __pyx_v_ldu, __pyx_v_V, __pyx_v_ldv, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, __pyx_v_params, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1684, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdjBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1699
 * 
 * # gesvda ... Approximate singular value decomposition
 * cpdef int sgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_217sgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdaStridedBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1705
 *         long long int strideV, int batchSize):
 *     cdef int lwork
 *     status = cusolverDnSgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <EigMode>jobz, rank, m, n, <const float*>d_A, lda,
 *         strideA, <const float*>d_S, strideS, <const float*>d_U, ldu, strideU,
 */
  __pyx_v_status = cusolverDnSgesvdaStridedBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((float const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((float const *)__pyx_v_d_S), __pyx_v_strideS, ((float const *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((float const *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, (&__pyx_v_lwork), __pyx_v_batchSize);

  /* "cupy/cuda/cusolver.pyx":1709
 *         strideA, <const float*>d_S, strideS, <const float*>d_U, ldu, strideU,
 *         <const float*>d_V, ldv, strideV, &lwork, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1710
 *         <const float*>d_V, ldv, strideV, &lwork, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dgesvdaStridedBatched_bufferSize(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1699
 * 
 * # gesvda ... Approximate singular value decomposition
 * cpdef int sgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.sgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_217sgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_216sgesvdaStridedBatched_bufferSize[] = "sgesvdaStridedBatched_bufferSize(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_217sgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvdaStridedBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_batchSize,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 1); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 2); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 3); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 4); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 5); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 6); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 7); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 8); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 9); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 10); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 11); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 12); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 13); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 14); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 15); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, 16); __PYX_ERR(0, 1699, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvdaStridedBatched_bufferSize") < 0)) __PYX_ERR(0, 1699, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1700, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1700, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1700, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1700, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1700, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1700, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1701, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1701, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1701, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1701, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1702, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1702, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1702, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1702, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1702, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1703, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[16]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1703, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched_bufferSize", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1699, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_216sgesvdaStridedBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_216sgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdaStridedBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1712
 *     return lwork
 * 
 * cpdef int dgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_219dgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdaStridedBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1718
 *         long long int strideV, int batchSize):
 *     cdef int lwork
 *     status = cusolverDnDgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <EigMode>jobz, rank, m, n, <const double*>d_A, lda,
 *         strideA, <const double*>d_S, strideS, <const double*>d_U, ldu, strideU,
 */
  __pyx_v_status = cusolverDnDgesvdaStridedBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((double const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((double const *)__pyx_v_d_S), __pyx_v_strideS, ((double const *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((double const *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, (&__pyx_v_lwork), __pyx_v_batchSize);

  /* "cupy/cuda/cusolver.pyx":1722
 *         strideA, <const double*>d_S, strideS, <const double*>d_U, ldu, strideU,
 *         <const double*>d_V, ldv, strideV, &lwork, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1723
 *         <const double*>d_V, ldv, strideV, &lwork, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cgesvdaStridedBatched_bufferSize(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1712
 *     return lwork
 * 
 * cpdef int dgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.dgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_219dgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_218dgesvdaStridedBatched_bufferSize[] = "dgesvdaStridedBatched_bufferSize(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_219dgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvdaStridedBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_batchSize,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 1); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 2); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 3); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 4); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 5); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 6); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 7); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 8); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 9); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 10); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 11); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 12); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 13); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 14); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 15); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, 16); __PYX_ERR(0, 1712, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvdaStridedBatched_bufferSize") < 0)) __PYX_ERR(0, 1712, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1713, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1714, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1714, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1714, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1714, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1715, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1715, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1715, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1715, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1715, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1716, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[16]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1716, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched_bufferSize", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1712, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_218dgesvdaStridedBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_218dgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdaStridedBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1712, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1725
 *     return lwork
 * 
 * cpdef int cgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_221cgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdaStridedBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1731
 *         long long int strideV, int batchSize):
 *     cdef int lwork
 *     status = cusolverDnCgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <EigMode>jobz, rank, m, n, <const cuComplex*>d_A, lda,
 *         strideA, <const float*>d_S, strideS, <const cuComplex*>d_U, ldu,
 */
  __pyx_v_status = cusolverDnCgesvdaStridedBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((cuComplex const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((float const *)__pyx_v_d_S), __pyx_v_strideS, ((cuComplex const *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((cuComplex const *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, (&__pyx_v_lwork), __pyx_v_batchSize);

  /* "cupy/cuda/cusolver.pyx":1735
 *         strideA, <const float*>d_S, strideS, <const cuComplex*>d_U, ldu,
 *         strideU, <const cuComplex*>d_V, ldv, strideV, &lwork, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1735, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1736
 *         strideU, <const cuComplex*>d_V, ldv, strideV, &lwork, batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zgesvdaStridedBatched_bufferSize(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1725
 *     return lwork
 * 
 * cpdef int cgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.cgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_221cgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_220cgesvdaStridedBatched_bufferSize[] = "cgesvdaStridedBatched_bufferSize(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_221cgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvdaStridedBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_batchSize,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 1); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 2); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 3); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 4); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 5); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 6); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 7); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 8); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 9); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 10); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 11); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 12); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 13); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 14); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 15); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, 16); __PYX_ERR(0, 1725, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvdaStridedBatched_bufferSize") < 0)) __PYX_ERR(0, 1725, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1726, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1726, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1726, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1726, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1726, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1726, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1727, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1727, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1727, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1727, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1728, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1729, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[16]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1729, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched_bufferSize", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1725, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_220cgesvdaStridedBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_220cgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdaStridedBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1725, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1738
 *     return lwork
 * 
 * cpdef int zgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_223zgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdaStridedBatched_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1744
 *         long long int strideV, int batchSize):
 *     cdef int lwork
 *     status = cusolverDnZgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         <Handle>handle, <EigMode>jobz, rank, m, n, <const cuDoubleComplex*>d_A,
 *         lda, strideA, <const double*>d_S, strideS, <const cuDoubleComplex*>d_U,
 */
  __pyx_v_status = cusolverDnZgesvdaStridedBatched_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((cuDoubleComplex const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((double const *)__pyx_v_d_S), __pyx_v_strideS, ((cuDoubleComplex const *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((cuDoubleComplex const *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, (&__pyx_v_lwork), __pyx_v_batchSize);

  /* "cupy/cuda/cusolver.pyx":1749
 *         ldu, strideU, <const cuDoubleComplex*>d_V, ldv, strideV, &lwork,
 *         batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1750
 *         batchSize)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef sgesvdaStridedBatched(
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1738
 *     return lwork
 * 
 * cpdef int zgesvdaStridedBatched_bufferSize(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("cupy.cuda.cusolver.zgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_223zgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_222zgesvdaStridedBatched_bufferSize[] = "zgesvdaStridedBatched_bufferSize(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, int batchSize) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_223zgesvdaStridedBatched_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvdaStridedBatched_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_batchSize,0};
    PyObject* values[17] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 1); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 2); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 3); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 4); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 5); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 6); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 7); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 8); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 9); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 10); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 11); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 12); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 13); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 14); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 15); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, 16); __PYX_ERR(0, 1738, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvdaStridedBatched_bufferSize") < 0)) __PYX_ERR(0, 1738, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 17) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1739, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1739, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1739, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1739, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1739, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1739, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1740, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1740, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1740, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1740, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1741, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1741, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1741, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1741, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1741, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1742, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[16]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1742, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched_bufferSize", 1, 17, 17, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1738, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_222zgesvdaStridedBatched_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_222zgesvdaStridedBatched_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdaStridedBatched_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_batchSize, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdaStridedBatched_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1752
 *     return lwork
 * 
 * cpdef sgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_225sgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdaStridedBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1758
 *         long long int strideV, intptr_t d_work, int lwork, intptr_t d_info,
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1758, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1759
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const float*>d_A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1760
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const float*>d_A, lda,
 *             strideA, <float*>d_S, strideS, <float*>d_U, ldu, strideU,
 */
        __pyx_v_status = cusolverDnSgesvdaStridedBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((float const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((float *)__pyx_v_d_S), __pyx_v_strideS, ((float *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((float *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, ((float *)__pyx_v_d_work), __pyx_v_lwork, ((int *)__pyx_v_d_info), ((double *)__pyx_v_h_R_nrmF), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1759
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const float*>d_A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1765
 *             <float*>d_V, ldv, strideV, <float*>d_work, lwork, <int*>d_info,
 *             <double*>h_R_nrmF, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1765, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1752
 *     return lwork
 * 
 * cpdef sgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_225sgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_224sgesvdaStridedBatched[] = "sgesvdaStridedBatched(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, intptr_t d_work, int lwork, intptr_t d_info, intptr_t h_R_nrmF, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_225sgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  intptr_t __pyx_v_d_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_d_info;
  intptr_t __pyx_v_h_R_nrmF;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sgesvdaStridedBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_d_work,&__pyx_n_s_lwork,&__pyx_n_s_d_info,&__pyx_n_s_h_R_nrmF,&__pyx_n_s_batchSize,0};
    PyObject* values[21] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 1); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 2); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 3); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 4); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 5); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 6); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 7); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 8); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 9); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 10); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 11); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 12); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 13); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 14); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 15); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 16); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 17); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 18); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h_R_nrmF)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 19); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, 20); __PYX_ERR(0, 1752, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "sgesvdaStridedBatched") < 0)) __PYX_ERR(0, 1752, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 21) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1753, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1753, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1753, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1753, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1753, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1753, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1754, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1754, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1754, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1754, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1755, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1755, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1755, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1755, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1755, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1756, __pyx_L3_error)
    __pyx_v_d_work = PyInt_AsSsize_t(values[16]); if (unlikely((__pyx_v_d_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1756, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[17]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1756, __pyx_L3_error)
    __pyx_v_d_info = PyInt_AsSsize_t(values[18]); if (unlikely((__pyx_v_d_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1756, __pyx_L3_error)
    __pyx_v_h_R_nrmF = PyInt_AsSsize_t(values[19]); if (unlikely((__pyx_v_h_R_nrmF == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1757, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[20]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1757, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sgesvdaStridedBatched", 1, 21, 21, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1752, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_224sgesvdaStridedBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_224sgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("sgesvdaStridedBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1752, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.sgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1767
 *     check_status(status)
 * 
 * cpdef dgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_227dgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdaStridedBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1773
 *         long long int strideV, intptr_t d_work, int lwork, intptr_t d_info,
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1773, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1774
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const double*>d_A, lda,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1775
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const double*>d_A, lda,
 *             strideA, <double*>d_S, strideS, <double*>d_U, ldu, strideU,
 */
        __pyx_v_status = cusolverDnDgesvdaStridedBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((double const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((double *)__pyx_v_d_S), __pyx_v_strideS, ((double *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((double *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, ((double *)__pyx_v_d_work), __pyx_v_lwork, ((int *)__pyx_v_d_info), ((double *)__pyx_v_h_R_nrmF), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1774
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const double*>d_A, lda,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1780
 *             <double*>d_V, ldv, strideV, <double*>d_work, lwork, <int*>d_info,
 *             <double*>h_R_nrmF, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1767
 *     check_status(status)
 * 
 * cpdef dgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_227dgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_226dgesvdaStridedBatched[] = "dgesvdaStridedBatched(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, intptr_t d_work, int lwork, intptr_t d_info, intptr_t h_R_nrmF, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_227dgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  intptr_t __pyx_v_d_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_d_info;
  intptr_t __pyx_v_h_R_nrmF;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dgesvdaStridedBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_d_work,&__pyx_n_s_lwork,&__pyx_n_s_d_info,&__pyx_n_s_h_R_nrmF,&__pyx_n_s_batchSize,0};
    PyObject* values[21] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 1); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 2); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 3); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 4); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 5); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 6); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 7); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 8); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 9); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 10); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 11); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 12); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 13); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 14); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 15); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 16); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 17); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 18); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h_R_nrmF)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 19); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, 20); __PYX_ERR(0, 1767, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dgesvdaStridedBatched") < 0)) __PYX_ERR(0, 1767, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 21) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1768, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1768, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1768, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1768, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1768, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1768, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1769, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1769, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1769, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1769, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1770, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1770, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1770, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1770, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1770, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1771, __pyx_L3_error)
    __pyx_v_d_work = PyInt_AsSsize_t(values[16]); if (unlikely((__pyx_v_d_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1771, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[17]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1771, __pyx_L3_error)
    __pyx_v_d_info = PyInt_AsSsize_t(values[18]); if (unlikely((__pyx_v_d_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1771, __pyx_L3_error)
    __pyx_v_h_R_nrmF = PyInt_AsSsize_t(values[19]); if (unlikely((__pyx_v_h_R_nrmF == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1772, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[20]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1772, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dgesvdaStridedBatched", 1, 21, 21, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1767, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_226dgesvdaStridedBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_226dgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dgesvdaStridedBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1767, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1782
 *     check_status(status)
 * 
 * cpdef cgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_229cgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdaStridedBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1788
 *         long long int strideV, intptr_t d_work, int lwork, intptr_t d_info,
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1789
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const cuComplex*>d_A,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1790
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const cuComplex*>d_A,
 *             lda, strideA, <float*>d_S, strideS, <cuComplex*>d_U, ldu, strideU,
 */
        __pyx_v_status = cusolverDnCgesvdaStridedBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((cuComplex const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((float *)__pyx_v_d_S), __pyx_v_strideS, ((cuComplex *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((cuComplex *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, ((cuComplex *)__pyx_v_d_work), __pyx_v_lwork, ((int *)__pyx_v_d_info), ((double *)__pyx_v_h_R_nrmF), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1789
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n, <const cuComplex*>d_A,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1795
 *             <cuComplex*>d_V, ldv, strideV, <cuComplex*>d_work, lwork,
 *             <int*>d_info, <double*>h_R_nrmF, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1782
 *     check_status(status)
 * 
 * cpdef cgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_229cgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_228cgesvdaStridedBatched[] = "cgesvdaStridedBatched(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, intptr_t d_work, int lwork, intptr_t d_info, intptr_t h_R_nrmF, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_229cgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  intptr_t __pyx_v_d_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_d_info;
  intptr_t __pyx_v_h_R_nrmF;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cgesvdaStridedBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_d_work,&__pyx_n_s_lwork,&__pyx_n_s_d_info,&__pyx_n_s_h_R_nrmF,&__pyx_n_s_batchSize,0};
    PyObject* values[21] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 1); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 2); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 3); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 4); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 5); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 6); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 7); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 8); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 9); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 10); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 11); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 12); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 13); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 14); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 15); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 16); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 17); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 18); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h_R_nrmF)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 19); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, 20); __PYX_ERR(0, 1782, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cgesvdaStridedBatched") < 0)) __PYX_ERR(0, 1782, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 21) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1783, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1783, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1783, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1783, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1783, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1783, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1784, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1784, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1784, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1784, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1785, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1785, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1785, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1785, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1785, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1786, __pyx_L3_error)
    __pyx_v_d_work = PyInt_AsSsize_t(values[16]); if (unlikely((__pyx_v_d_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1786, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[17]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1786, __pyx_L3_error)
    __pyx_v_d_info = PyInt_AsSsize_t(values[18]); if (unlikely((__pyx_v_d_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1786, __pyx_L3_error)
    __pyx_v_h_R_nrmF = PyInt_AsSsize_t(values[19]); if (unlikely((__pyx_v_h_R_nrmF == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1787, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[20]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1787, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cgesvdaStridedBatched", 1, 21, 21, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1782, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_228cgesvdaStridedBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_228cgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cgesvdaStridedBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1782, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1797
 *     check_status(status)
 * 
 * cpdef zgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_231zgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdaStridedBatched", 0);

  /* "cupy/cuda/cusolver.pyx":1803
 *         long long int strideV, intptr_t d_work, int lwork, intptr_t d_info,
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZgesvdaStridedBatched(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1804
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1805
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, rank, m, n,
 *             <const cuDoubleComplex*>d_A, lda, strideA, <double*>d_S, strideS,
 */
        __pyx_v_status = cusolverDnZgesvdaStridedBatched(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), __pyx_v_rank, __pyx_v_m, __pyx_v_n, ((cuDoubleComplex const *)__pyx_v_d_A), __pyx_v_lda, __pyx_v_strideA, ((double *)__pyx_v_d_S), __pyx_v_strideS, ((cuDoubleComplex *)__pyx_v_d_U), __pyx_v_ldu, __pyx_v_strideU, ((cuDoubleComplex *)__pyx_v_d_V), __pyx_v_ldv, __pyx_v_strideV, ((cuDoubleComplex *)__pyx_v_d_work), __pyx_v_lwork, ((int *)__pyx_v_d_info), ((double *)__pyx_v_h_R_nrmF), __pyx_v_batchSize);
      }

      /* "cupy/cuda/cusolver.pyx":1804
 *         intptr_t h_R_nrmF, int batchSize):
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZgesvdaStridedBatched(
 *             <Handle>handle, <EigMode>jobz, rank, m, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1811
 *             strideV, <cuDoubleComplex*>d_work, lwork, <int*>d_info,
 *             <double*>h_R_nrmF, batchSize)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * # Standard symmetric eigenvalue solver
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1797
 *     check_status(status)
 * 
 * cpdef zgesvdaStridedBatched(             # <<<<<<<<<<<<<<
 *         intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A,
 *         int lda, long long int strideA, intptr_t d_S, long long int strideS,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_231zgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_230zgesvdaStridedBatched[] = "zgesvdaStridedBatched(intptr_t handle, int jobz, int rank, int m, int n, intptr_t d_A, int lda, long long strideA, intptr_t d_S, long long strideS, intptr_t d_U, int ldu, long long strideU, intptr_t d_V, int ldv, long long strideV, intptr_t d_work, int lwork, intptr_t d_info, intptr_t h_R_nrmF, int batchSize)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_231zgesvdaStridedBatched(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_rank;
  int __pyx_v_m;
  int __pyx_v_n;
  intptr_t __pyx_v_d_A;
  int __pyx_v_lda;
  PY_LONG_LONG __pyx_v_strideA;
  intptr_t __pyx_v_d_S;
  PY_LONG_LONG __pyx_v_strideS;
  intptr_t __pyx_v_d_U;
  int __pyx_v_ldu;
  PY_LONG_LONG __pyx_v_strideU;
  intptr_t __pyx_v_d_V;
  int __pyx_v_ldv;
  PY_LONG_LONG __pyx_v_strideV;
  intptr_t __pyx_v_d_work;
  int __pyx_v_lwork;
  intptr_t __pyx_v_d_info;
  intptr_t __pyx_v_h_R_nrmF;
  int __pyx_v_batchSize;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zgesvdaStridedBatched (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_rank,&__pyx_n_s_m,&__pyx_n_s_n,&__pyx_n_s_d_A,&__pyx_n_s_lda,&__pyx_n_s_strideA,&__pyx_n_s_d_S,&__pyx_n_s_strideS,&__pyx_n_s_d_U,&__pyx_n_s_ldu,&__pyx_n_s_strideU,&__pyx_n_s_d_V,&__pyx_n_s_ldv,&__pyx_n_s_strideV,&__pyx_n_s_d_work,&__pyx_n_s_lwork,&__pyx_n_s_d_info,&__pyx_n_s_h_R_nrmF,&__pyx_n_s_batchSize,0};
    PyObject* values[21] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 21: values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
        CYTHON_FALLTHROUGH;
        case 20: values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
        CYTHON_FALLTHROUGH;
        case 19: values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
        CYTHON_FALLTHROUGH;
        case 18: values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
        CYTHON_FALLTHROUGH;
        case 17: values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
        CYTHON_FALLTHROUGH;
        case 16: values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
        CYTHON_FALLTHROUGH;
        case 15: values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
        CYTHON_FALLTHROUGH;
        case 14: values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
        CYTHON_FALLTHROUGH;
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 1); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_rank)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 2); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 3); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 4); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 5); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 6); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 7); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_S)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 8); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideS)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 9); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_U)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 10); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 11); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideU)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 12); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 13:
        if (likely((values[13] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_V)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 13); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 14:
        if (likely((values[14] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ldv)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 14); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 15:
        if (likely((values[15] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_strideV)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 15); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 16:
        if (likely((values[16] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 16); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 17:
        if (likely((values[17] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 17); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 18:
        if (likely((values[18] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_d_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 18); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 19:
        if (likely((values[19] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_h_R_nrmF)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 19); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 20:
        if (likely((values[20] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_batchSize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, 20); __PYX_ERR(0, 1797, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zgesvdaStridedBatched") < 0)) __PYX_ERR(0, 1797, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 21) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
      values[13] = PyTuple_GET_ITEM(__pyx_args, 13);
      values[14] = PyTuple_GET_ITEM(__pyx_args, 14);
      values[15] = PyTuple_GET_ITEM(__pyx_args, 15);
      values[16] = PyTuple_GET_ITEM(__pyx_args, 16);
      values[17] = PyTuple_GET_ITEM(__pyx_args, 17);
      values[18] = PyTuple_GET_ITEM(__pyx_args, 18);
      values[19] = PyTuple_GET_ITEM(__pyx_args, 19);
      values[20] = PyTuple_GET_ITEM(__pyx_args, 20);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1798, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1798, __pyx_L3_error)
    __pyx_v_rank = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_rank == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1798, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1798, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1798, __pyx_L3_error)
    __pyx_v_d_A = PyInt_AsSsize_t(values[5]); if (unlikely((__pyx_v_d_A == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1798, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    __pyx_v_strideA = __Pyx_PyInt_As_PY_LONG_LONG(values[7]); if (unlikely((__pyx_v_strideA == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    __pyx_v_d_S = PyInt_AsSsize_t(values[8]); if (unlikely((__pyx_v_d_S == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    __pyx_v_strideS = __Pyx_PyInt_As_PY_LONG_LONG(values[9]); if (unlikely((__pyx_v_strideS == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1799, __pyx_L3_error)
    __pyx_v_d_U = PyInt_AsSsize_t(values[10]); if (unlikely((__pyx_v_d_U == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1800, __pyx_L3_error)
    __pyx_v_ldu = __Pyx_PyInt_As_int(values[11]); if (unlikely((__pyx_v_ldu == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1800, __pyx_L3_error)
    __pyx_v_strideU = __Pyx_PyInt_As_PY_LONG_LONG(values[12]); if (unlikely((__pyx_v_strideU == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1800, __pyx_L3_error)
    __pyx_v_d_V = PyInt_AsSsize_t(values[13]); if (unlikely((__pyx_v_d_V == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1800, __pyx_L3_error)
    __pyx_v_ldv = __Pyx_PyInt_As_int(values[14]); if (unlikely((__pyx_v_ldv == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1800, __pyx_L3_error)
    __pyx_v_strideV = __Pyx_PyInt_As_PY_LONG_LONG(values[15]); if (unlikely((__pyx_v_strideV == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 1801, __pyx_L3_error)
    __pyx_v_d_work = PyInt_AsSsize_t(values[16]); if (unlikely((__pyx_v_d_work == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1801, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[17]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1801, __pyx_L3_error)
    __pyx_v_d_info = PyInt_AsSsize_t(values[18]); if (unlikely((__pyx_v_d_info == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1801, __pyx_L3_error)
    __pyx_v_h_R_nrmF = PyInt_AsSsize_t(values[19]); if (unlikely((__pyx_v_h_R_nrmF == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1802, __pyx_L3_error)
    __pyx_v_batchSize = __Pyx_PyInt_As_int(values[20]); if (unlikely((__pyx_v_batchSize == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1802, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zgesvdaStridedBatched", 1, 21, 21, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1797, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_230zgesvdaStridedBatched(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_230zgesvdaStridedBatched(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_rank, int __pyx_v_m, int __pyx_v_n, intptr_t __pyx_v_d_A, int __pyx_v_lda, PY_LONG_LONG __pyx_v_strideA, intptr_t __pyx_v_d_S, PY_LONG_LONG __pyx_v_strideS, intptr_t __pyx_v_d_U, int __pyx_v_ldu, PY_LONG_LONG __pyx_v_strideU, intptr_t __pyx_v_d_V, int __pyx_v_ldv, PY_LONG_LONG __pyx_v_strideV, intptr_t __pyx_v_d_work, int __pyx_v_lwork, intptr_t __pyx_v_d_info, intptr_t __pyx_v_h_R_nrmF, int __pyx_v_batchSize) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zgesvdaStridedBatched", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched(__pyx_v_handle, __pyx_v_jobz, __pyx_v_rank, __pyx_v_m, __pyx_v_n, __pyx_v_d_A, __pyx_v_lda, __pyx_v_strideA, __pyx_v_d_S, __pyx_v_strideS, __pyx_v_d_U, __pyx_v_ldu, __pyx_v_strideU, __pyx_v_d_V, __pyx_v_ldv, __pyx_v_strideV, __pyx_v_d_work, __pyx_v_lwork, __pyx_v_d_info, __pyx_v_h_R_nrmF, __pyx_v_batchSize, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zgesvdaStridedBatched", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1814
 * 
 * # Standard symmetric eigenvalue solver
 * cpdef int ssyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_233ssyevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_ssyevd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ssyevd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1817
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSsyevd_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1818
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsyevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1819
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSsyevd_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <const float*>A,
 */
        __pyx_v_status = cusolverDnSsyevd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((float const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_W), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1818
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsyevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1823
 *             <const float*>A,
 *             lda, <const float*>W, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1824
 *             lda, <const float*>W, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int dsyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1814
 * 
 * # Standard symmetric eigenvalue solver
 * cpdef int ssyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssyevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_233ssyevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_232ssyevd_bufferSize[] = "ssyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_233ssyevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ssyevd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1814, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1814, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1814, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1814, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1814, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1814, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ssyevd_bufferSize") < 0)) __PYX_ERR(0, 1814, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1814, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1814, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1814, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1814, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1815, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1815, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1815, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ssyevd_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1814, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssyevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_232ssyevd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_232ssyevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("ssyevd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ssyevd_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1814, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1814, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssyevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1826
 *     return lwork
 * 
 * cpdef int dsyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_235dsyevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_dsyevd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dsyevd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1829
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDsyevd_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1829, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1830
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsyevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1831
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDsyevd_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <const double*>A,
 */
        __pyx_v_status = cusolverDnDsyevd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((double const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_W), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1830
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsyevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1835
 *             <const double*>A,
 *             lda, <const double*>W, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1836
 *             lda, <const double*>W, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int cheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1826
 *     return lwork
 * 
 * cpdef int dsyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsyevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_235dsyevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_234dsyevd_bufferSize[] = "dsyevd_bufferSize(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_235dsyevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dsyevd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1826, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1826, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dsyevd_bufferSize") < 0)) __PYX_ERR(0, 1826, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1826, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1826, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1826, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1826, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1827, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1827, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1827, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dsyevd_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsyevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_234dsyevd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_234dsyevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("dsyevd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dsyevd_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1826, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsyevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1838
 *     return lwork
 * 
 * cpdef int cheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_237cheevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_cheevd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cheevd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1841
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCheevd_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1842
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCheevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1843
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCheevd_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <const cuComplex*>A,
 */
        __pyx_v_status = cusolverDnCheevd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuComplex const *)__pyx_v_A), __pyx_v_lda, ((float const *)__pyx_v_W), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1842
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCheevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1847
 *             <const cuComplex*>A,
 *             lda, <const float*>W, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1848
 *             lda, <const float*>W, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef int zheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1838
 *     return lwork
 * 
 * cpdef int cheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cheevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_237cheevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_236cheevd_bufferSize[] = "cheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_237cheevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cheevd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1838, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1838, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cheevd_bufferSize") < 0)) __PYX_ERR(0, 1838, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1838, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1838, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1838, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1838, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1839, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1839, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1839, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cheevd_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1838, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cheevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_236cheevd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_236cheevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("cheevd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cheevd_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1838, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cheevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1850
 *     return lwork
 * 
 * cpdef int zheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_239zheevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_f_4cupy_4cuda_8cusolver_zheevd_bufferSize(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_lwork;
  int __pyx_v_status;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zheevd_bufferSize", 0);

  /* "cupy/cuda/cusolver.pyx":1853
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZheevd_bufferSize(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1854
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZheevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1855
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZheevd_bufferSize(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <const cuDoubleComplex*>A,
 */
        __pyx_v_status = cusolverDnZheevd_bufferSize(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuDoubleComplex const *)__pyx_v_A), __pyx_v_lda, ((double const *)__pyx_v_W), (&__pyx_v_lwork));
      }

      /* "cupy/cuda/cusolver.pyx":1854
 *     cdef int lwork, status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZheevd_bufferSize(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1859
 *             <const cuDoubleComplex*>A,
 *             lda, <const double*>W, &lwork)
 *     check_status(status)             # <<<<<<<<<<<<<<
 *     return lwork
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1860
 *             lda, <const double*>W, &lwork)
 *     check_status(status)
 *     return lwork             # <<<<<<<<<<<<<<
 * 
 * cpdef ssyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,
 */
  __pyx_r = __pyx_v_lwork;
  goto __pyx_L0;

  /* "cupy/cuda/cusolver.pyx":1850
 *     return lwork
 * 
 * cpdef int zheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n,             # <<<<<<<<<<<<<<
 *                             size_t A, int lda, size_t W) except? -1:
 *     cdef int lwork, status
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zheevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_239zheevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_238zheevd_bufferSize[] = "zheevd_bufferSize(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W) -> int";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_239zheevd_bufferSize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zheevd_bufferSize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, 1); __PYX_ERR(0, 1850, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, 2); __PYX_ERR(0, 1850, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, 3); __PYX_ERR(0, 1850, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, 4); __PYX_ERR(0, 1850, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, 5); __PYX_ERR(0, 1850, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, 6); __PYX_ERR(0, 1850, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zheevd_bufferSize") < 0)) __PYX_ERR(0, 1850, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 7) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1850, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1850, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1850, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1850, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1851, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1851, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1851, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zheevd_bufferSize", 1, 7, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1850, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zheevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_238zheevd_bufferSize(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_238zheevd_bufferSize(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("zheevd_bufferSize", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zheevd_bufferSize(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, 0); if (unlikely(__pyx_t_1 == ((int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1850, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1850, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zheevd_bufferSize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1862
 *     return lwork
 * 
 * cpdef ssyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_241ssyevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ssyevd(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ssyevd", 0);

  /* "cupy/cuda/cusolver.pyx":1865
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnSsyevd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1865, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1866
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsyevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1867
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnSsyevd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <float*>A, lda, <float*>W,
 */
        __pyx_v_status = cusolverDnSsyevd(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((float *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_W), ((float *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info));
      }

      /* "cupy/cuda/cusolver.pyx":1866
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnSsyevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1871
 *             <float*>A, lda, <float*>W,
 *             <float*>work, lwork, <int*>info)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dsyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1862
 *     return lwork
 * 
 * cpdef ssyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssyevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_241ssyevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_240ssyevd[] = "ssyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W, size_t work, int lwork, size_t info)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_241ssyevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ssyevd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 1); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 2); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 3); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 4); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 5); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 6); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 7); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 8); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, 9); __PYX_ERR(0, 1862, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ssyevd") < 0)) __PYX_ERR(0, 1862, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1862, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1862, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1862, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1862, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1862, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1862, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1863, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1863, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1863, __pyx_L3_error)
    __pyx_v_info = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_info == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1863, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ssyevd", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1862, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssyevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_240ssyevd(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_240ssyevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ssyevd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ssyevd(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ssyevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1873
 *     check_status(status)
 * 
 * cpdef dsyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_243dsyevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dsyevd(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dsyevd", 0);

  /* "cupy/cuda/cusolver.pyx":1876
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnDsyevd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1877
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsyevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1878
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnDsyevd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <double*>A, lda, <double*>W,
 */
        __pyx_v_status = cusolverDnDsyevd(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((double *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_W), ((double *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info));
      }

      /* "cupy/cuda/cusolver.pyx":1877
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnDsyevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1882
 *             <double*>A, lda, <double*>W,
 *             <double*>work, lwork, <int*>info)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef cheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1873
 *     check_status(status)
 * 
 * cpdef dsyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsyevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_243dsyevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_242dsyevd[] = "dsyevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W, size_t work, int lwork, size_t info)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_243dsyevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dsyevd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 1); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 2); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 3); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 4); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 5); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 6); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 7); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 8); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, 9); __PYX_ERR(0, 1873, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dsyevd") < 0)) __PYX_ERR(0, 1873, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1873, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
    __pyx_v_info = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_info == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1874, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dsyevd", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1873, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsyevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_242dsyevd(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_242dsyevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dsyevd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dsyevd(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dsyevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1884
 *     check_status(status)
 * 
 * cpdef cheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_245cheevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_cheevd(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cheevd", 0);

  /* "cupy/cuda/cusolver.pyx":1887
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnCheevd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1887, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1888
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCheevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1889
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnCheevd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <cuComplex*>A, lda, <float*>W,
 */
        __pyx_v_status = cusolverDnCheevd(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuComplex *)__pyx_v_A), __pyx_v_lda, ((float *)__pyx_v_W), ((cuComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info));
      }

      /* "cupy/cuda/cusolver.pyx":1888
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnCheevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1893
 *             <cuComplex*>A, lda, <float*>W,
 *             <cuComplex*>work, lwork, <int*>info)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1893, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1884
 *     check_status(status)
 * 
 * cpdef cheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cheevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_245cheevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_244cheevd[] = "cheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W, size_t work, int lwork, size_t info)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_245cheevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cheevd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 1); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 2); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 3); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 4); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 5); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 6); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 7); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 8); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, 9); __PYX_ERR(0, 1884, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "cheevd") < 0)) __PYX_ERR(0, 1884, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1884, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1885, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1885, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1885, __pyx_L3_error)
    __pyx_v_info = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_info == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1885, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cheevd", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1884, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.cheevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_244cheevd(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_244cheevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("cheevd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_cheevd(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.cheevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1895
 *     check_status(status)
 * 
 * cpdef zheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_247zheevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zheevd(intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zheevd", 0);

  /* "cupy/cuda/cusolver.pyx":1898
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverDnZheevd(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_setStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1899
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZheevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1900
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverDnZheevd(             # <<<<<<<<<<<<<<
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 *             <cuDoubleComplex*>A, lda, <double*>W,
 */
        __pyx_v_status = cusolverDnZheevd(((cusolverDnHandle_t)__pyx_v_handle), ((cusolverEigMode_t)__pyx_v_jobz), ((cublasFillMode_t)__pyx_v_uplo), __pyx_v_n, ((cuDoubleComplex *)__pyx_v_A), __pyx_v_lda, ((double *)__pyx_v_W), ((cuDoubleComplex *)__pyx_v_work), __pyx_v_lwork, ((int *)__pyx_v_info));
      }

      /* "cupy/cuda/cusolver.pyx":1899
 *     cdef int status
 *     setStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverDnZheevd(
 *             <Handle>handle, <EigMode>jobz, <FillMode>uplo, n,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1904
 *             <cuDoubleComplex*>A, lda, <double*>W,
 *             <cuDoubleComplex*>work, lwork, <int*>info)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1895
 *     check_status(status)
 * 
 * cpdef zheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda,             # <<<<<<<<<<<<<<
 *              size_t W, size_t work, int lwork, size_t info):
 *     cdef int status
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zheevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_247zheevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_246zheevd[] = "zheevd(intptr_t handle, int jobz, int uplo, int n, size_t A, int lda, size_t W, size_t work, int lwork, size_t info)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_247zheevd(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_jobz;
  int __pyx_v_uplo;
  int __pyx_v_n;
  size_t __pyx_v_A;
  int __pyx_v_lda;
  size_t __pyx_v_W;
  size_t __pyx_v_work;
  int __pyx_v_lwork;
  size_t __pyx_v_info;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zheevd (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_jobz,&__pyx_n_s_uplo,&__pyx_n_s_n,&__pyx_n_s_A,&__pyx_n_s_lda,&__pyx_n_s_W,&__pyx_n_s_work,&__pyx_n_s_lwork,&__pyx_n_s_info,0};
    PyObject* values[10] = {0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_jobz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 1); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_uplo)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 2); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_n)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 3); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_A)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 4); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lda)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 5); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_W)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 6); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_work)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 7); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_lwork)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 8); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_info)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, 9); __PYX_ERR(0, 1895, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zheevd") < 0)) __PYX_ERR(0, 1895, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 10) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1895, __pyx_L3_error)
    __pyx_v_jobz = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_jobz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1895, __pyx_L3_error)
    __pyx_v_uplo = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_uplo == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1895, __pyx_L3_error)
    __pyx_v_n = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_n == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1895, __pyx_L3_error)
    __pyx_v_A = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_A == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1895, __pyx_L3_error)
    __pyx_v_lda = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_lda == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1895, __pyx_L3_error)
    __pyx_v_W = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_W == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1896, __pyx_L3_error)
    __pyx_v_work = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_work == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1896, __pyx_L3_error)
    __pyx_v_lwork = __Pyx_PyInt_As_int(values[8]); if (unlikely((__pyx_v_lwork == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1896, __pyx_L3_error)
    __pyx_v_info = __Pyx_PyInt_As_size_t(values[9]); if (unlikely((__pyx_v_info == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1896, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zheevd", 1, 10, 10, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1895, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zheevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_246zheevd(__pyx_self, __pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_246zheevd(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_jobz, int __pyx_v_uplo, int __pyx_v_n, size_t __pyx_v_A, int __pyx_v_lda, size_t __pyx_v_W, size_t __pyx_v_work, int __pyx_v_lwork, size_t __pyx_v_info) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zheevd", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zheevd(__pyx_v_handle, __pyx_v_jobz, __pyx_v_uplo, __pyx_v_n, __pyx_v_A, __pyx_v_lda, __pyx_v_W, __pyx_v_work, __pyx_v_lwork, __pyx_v_info, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1895, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zheevd", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1910
 * # Sparse LAPACK Functions
 * ###############################################################################
 * cpdef scsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                   size_t b, float tol, int reorder, size_t x,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_249scsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_scsrlsvchol(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrlsvchol", 0);

  /* "cupy/cuda/cusolver.pyx":1915
 *                   size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpScsrlsvchol(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1916
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpScsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1917
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpScsrlsvchol(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 *             <const float*> csrValA, <const int*> csrRowPtrA,
 */
        __pyx_v_status = cusolverSpScsrlsvchol(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((float const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((float const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((float *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1916
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpScsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1922
 *             <const int*> csrColIndA, <const float*> b,
 *             tol, reorder, <float*> x, <int*> singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1910
 * # Sparse LAPACK Functions
 * ###############################################################################
 * cpdef scsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                   size_t b, float tol, int reorder, size_t x,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_249scsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_248scsrlsvchol[] = "scsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t b, float tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_249scsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_b;
  float __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrlsvchol (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 1); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 2); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 3); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 4); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 5); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 6); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 7); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 8); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 9); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 10); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, 11); __PYX_ERR(0, 1910, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrlsvchol") < 0)) __PYX_ERR(0, 1910, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1910, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1910, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1910, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1910, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1911, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1911, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1911, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[8]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1912, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1913, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrlsvchol", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1910, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_248scsrlsvchol(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_248scsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrlsvchol", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_scsrlsvchol(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1924
 *     check_status(status)
 * 
 * cpdef dcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                   size_t b, double tol, int reorder, size_t x,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_251dcsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dcsrlsvchol(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrlsvchol", 0);

  /* "cupy/cuda/cusolver.pyx":1929
 *                   size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpDcsrlsvchol(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1929, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1930
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDcsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1931
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpDcsrlsvchol(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 *             <const double*> csrValA, <const int*> csrRowPtrA,
 */
        __pyx_v_status = cusolverSpDcsrlsvchol(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((double const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((double const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((double *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1930
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDcsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1936
 *             <const int*> csrColIndA, <const double*> b,
 *             tol, reorder, <double*> x, <int*> singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef ccsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1924
 *     check_status(status)
 * 
 * cpdef dcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                   size_t b, double tol, int reorder, size_t x,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_251dcsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_250dcsrlsvchol[] = "dcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t b, double tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_251dcsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_b;
  double __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrlsvchol (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 1); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 2); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 3); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 4); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 5); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 6); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 7); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 8); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 9); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 10); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, 11); __PYX_ERR(0, 1924, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrlsvchol") < 0)) __PYX_ERR(0, 1924, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1924, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1924, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1924, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1924, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1925, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1925, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1925, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1926, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_tol == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1926, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1926, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1926, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1927, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrlsvchol", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1924, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_250dcsrlsvchol(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_250dcsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrlsvchol", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dcsrlsvchol(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1938
 *     check_status(status)
 * 
 * cpdef ccsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b,
 *                   float tol, int reorder, size_t x, size_t singularity):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_253ccsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ccsrlsvchol(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ccsrlsvchol", 0);

  /* "cupy/cuda/cusolver.pyx":1942
 *                   float tol, int reorder, size_t x, size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpCcsrlsvchol(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1942, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1943
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCcsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1944
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpCcsrlsvchol(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const cuComplex*>csrVal, <const int*>csrRowPtr,
 */
        __pyx_v_status = cusolverSpCcsrlsvchol(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((cuComplex const *)__pyx_v_csrVal), ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), ((cuComplex const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((cuComplex *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1943
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCcsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1949
 *             <const int*>csrColInd, <const cuComplex*>b, tol, reorder,
 *             <cuComplex*>x, <int*>singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1949, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1938
 *     check_status(status)
 * 
 * cpdef ccsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b,
 *                   float tol, int reorder, size_t x, size_t singularity):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_253ccsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_252ccsrlsvchol[] = "ccsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b, float tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_253ccsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrVal;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_b;
  float __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ccsrlsvchol (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrVal,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 1); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 2); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 3); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 4); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 5); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 6); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 7); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 8); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 9); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 10); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, 11); __PYX_ERR(0, 1938, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ccsrlsvchol") < 0)) __PYX_ERR(0, 1938, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1938, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1938, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1938, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1938, __pyx_L3_error)
    __pyx_v_csrVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1939, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1939, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1939, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1939, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[8]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1940, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1940, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1940, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1940, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ccsrlsvchol", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1938, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_252ccsrlsvchol(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_252ccsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ccsrlsvchol", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ccsrlsvchol(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1938, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1951
 *     check_status(status)
 * 
 * cpdef zcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b,
 *                   double tol, int reorder, size_t x, size_t singularity):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_255zcsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zcsrlsvchol(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zcsrlsvchol", 0);

  /* "cupy/cuda/cusolver.pyx":1955
 *                   double tol, int reorder, size_t x, size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpZcsrlsvchol(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1956
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpZcsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1957
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpZcsrlsvchol(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const cuDoubleComplex*>csrVal, <const int*>csrRowPtr,
 */
        __pyx_v_status = cusolverSpZcsrlsvchol(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((cuDoubleComplex const *)__pyx_v_csrVal), ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), ((cuDoubleComplex const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((cuDoubleComplex *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1956
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpZcsrlsvchol(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1962
 *             <const int*>csrColInd, <const cuDoubleComplex*>b, tol, reorder,
 *             <cuDoubleComplex*>x, <int*>singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef scsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1962, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1951
 *     check_status(status)
 * 
 * cpdef zcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                   size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b,
 *                   double tol, int reorder, size_t x, size_t singularity):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_255zcsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_254zcsrlsvchol[] = "zcsrlsvchol(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b, double tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_255zcsrlsvchol(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrVal;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_b;
  double __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zcsrlsvchol (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrVal,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 1); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 2); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 3); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 4); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 5); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 6); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 7); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 8); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 9); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 10); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, 11); __PYX_ERR(0, 1951, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zcsrlsvchol") < 0)) __PYX_ERR(0, 1951, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1951, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1951, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1951, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1951, __pyx_L3_error)
    __pyx_v_csrVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1952, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1952, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1952, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1952, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_tol == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1953, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1953, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1953, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1953, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zcsrlsvchol", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1951, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_254zcsrlsvchol(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_254zcsrlsvchol(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zcsrlsvchol", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zcsrlsvchol(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1951, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsrlsvchol", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1964
 *     check_status(status)
 * 
 * cpdef scsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtrA, size_t csrColIndA, size_t b, float tol,
 *                 int reorder, size_t x, size_t singularity):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_257scsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_scsrlsvqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrlsvqr", 0);

  /* "cupy/cuda/cusolver.pyx":1968
 *                 int reorder, size_t x, size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpScsrlsvqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1969
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpScsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1970
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpScsrlsvqr(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 *             <const float*> csrValA, <const int*> csrRowPtrA,
 */
        __pyx_v_status = cusolverSpScsrlsvqr(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((float const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((float const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((float *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1969
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpScsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1975
 *             <const int*> csrColIndA, <const float*> b,
 *             tol, reorder, <float*> x, <int*> singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1964
 *     check_status(status)
 * 
 * cpdef scsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtrA, size_t csrColIndA, size_t b, float tol,
 *                 int reorder, size_t x, size_t singularity):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_257scsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_256scsrlsvqr[] = "scsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t b, float tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_257scsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_b;
  float __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsrlsvqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 1); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 2); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 3); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 4); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 5); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 6); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 7); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 8); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 9); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 10); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, 11); __PYX_ERR(0, 1964, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsrlsvqr") < 0)) __PYX_ERR(0, 1964, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1964, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1965, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1965, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1965, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[8]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1965, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1966, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1966, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1966, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsrlsvqr", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1964, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_256scsrlsvqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_256scsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsrlsvqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_scsrlsvqr(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1977
 *     check_status(status)
 * 
 * cpdef dcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtrA, size_t csrColIndA, size_t b, double tol,
 *                 int reorder, size_t x, size_t singularity):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_259dcsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dcsrlsvqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrlsvqr", 0);

  /* "cupy/cuda/cusolver.pyx":1981
 *                 int reorder, size_t x, size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpDcsrlsvqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1982
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDcsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1983
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpDcsrlsvqr(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 *             <const double*> csrValA, <const int*> csrRowPtrA,
 */
        __pyx_v_status = cusolverSpDcsrlsvqr(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((double const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), ((double const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((double *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1982
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDcsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr> descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":1988
 *             <const int*> csrColIndA, <const double*> b,
 *             tol, reorder, <double*> x, <int*> singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef ccsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1977
 *     check_status(status)
 * 
 * cpdef dcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtrA, size_t csrColIndA, size_t b, double tol,
 *                 int reorder, size_t x, size_t singularity):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_259dcsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_258dcsrlsvqr[] = "dcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t b, double tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_259dcsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_b;
  double __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsrlsvqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 1); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 2); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 3); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 4); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 5); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 6); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 7); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 8); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 9); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 10); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, 11); __PYX_ERR(0, 1977, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsrlsvqr") < 0)) __PYX_ERR(0, 1977, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1977, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1977, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1977, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1977, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1977, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1978, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1978, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1978, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_tol == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1978, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1979, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1979, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1979, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsrlsvqr", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1977, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_258dcsrlsvqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_258dcsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsrlsvqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dcsrlsvqr(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1977, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":1990
 *     check_status(status)
 * 
 * cpdef ccsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtr, size_t csrColInd, size_t b, float tol,
 *                 int reorder, size_t x, size_t singularity):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_261ccsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ccsrlsvqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ccsrlsvqr", 0);

  /* "cupy/cuda/cusolver.pyx":1994
 *                 int reorder, size_t x, size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpCcsrlsvqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1994, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1995
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCcsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":1996
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpCcsrlsvqr(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const cuComplex*>csrVal, <const int*>csrRowPtr,
 */
        __pyx_v_status = cusolverSpCcsrlsvqr(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((cuComplex const *)__pyx_v_csrVal), ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), ((cuComplex const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((cuComplex *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":1995
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCcsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":2001
 *             <const int*>csrColInd, <const cuComplex*>b, tol, reorder,
 *             <cuComplex*>x, <int*>singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1990
 *     check_status(status)
 * 
 * cpdef ccsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtr, size_t csrColInd, size_t b, float tol,
 *                 int reorder, size_t x, size_t singularity):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_261ccsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_260ccsrlsvqr[] = "ccsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b, float tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_261ccsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrVal;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_b;
  float __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ccsrlsvqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrVal,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 1); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 2); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 3); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 4); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 5); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 6); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 7); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 8); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 9); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 10); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, 11); __PYX_ERR(0, 1990, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ccsrlsvqr") < 0)) __PYX_ERR(0, 1990, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1990, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1990, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1990, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1990, __pyx_L3_error)
    __pyx_v_csrVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1990, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1991, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1991, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1991, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsFloat(values[8]); if (unlikely((__pyx_v_tol == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 1991, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1992, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1992, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1992, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ccsrlsvqr", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1990, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_260ccsrlsvqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_260ccsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, float __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ccsrlsvqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ccsrlsvqr(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":2003
 *     check_status(status)
 * 
 * cpdef zcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtr, size_t csrColInd, size_t b, double tol,
 *                 int reorder, size_t x, size_t singularity):
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_263zcsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zcsrlsvqr(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zcsrlsvqr", 0);

  /* "cupy/cuda/cusolver.pyx":2007
 *                 int reorder, size_t x, size_t singularity):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpZcsrlsvqr(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2007, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2008
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpZcsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":2009
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpZcsrlsvqr(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const cuDoubleComplex*>csrVal, <const int*>csrRowPtr,
 */
        __pyx_v_status = cusolverSpZcsrlsvqr(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((cuDoubleComplex const *)__pyx_v_csrVal), ((int const *)__pyx_v_csrRowPtr), ((int const *)__pyx_v_csrColInd), ((cuDoubleComplex const *)__pyx_v_b), __pyx_v_tol, __pyx_v_reorder, ((cuDoubleComplex *)__pyx_v_x), ((int *)__pyx_v_singularity));
      }

      /* "cupy/cuda/cusolver.pyx":2008
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpZcsrlsvqr(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":2014
 *             <const int*>csrColInd, <const cuDoubleComplex*>b, tol, reorder,
 *             <cuDoubleComplex*>x, <int*>singularity)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef scsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2014, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2003
 *     check_status(status)
 * 
 * cpdef zcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal,             # <<<<<<<<<<<<<<
 *                 size_t csrRowPtr, size_t csrColInd, size_t b, double tol,
 *                 int reorder, size_t x, size_t singularity):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_263zcsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_262zcsrlsvqr[] = "zcsrlsvqr(intptr_t handle, int m, int nnz, size_t descrA, size_t csrVal, size_t csrRowPtr, size_t csrColInd, size_t b, double tol, int reorder, size_t x, size_t singularity)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_263zcsrlsvqr(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrVal;
  size_t __pyx_v_csrRowPtr;
  size_t __pyx_v_csrColInd;
  size_t __pyx_v_b;
  double __pyx_v_tol;
  int __pyx_v_reorder;
  size_t __pyx_v_x;
  size_t __pyx_v_singularity;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zcsrlsvqr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrVal,&__pyx_n_s_csrRowPtr,&__pyx_n_s_csrColInd,&__pyx_n_s_b,&__pyx_n_s_tol,&__pyx_n_s_reorder,&__pyx_n_s_x,&__pyx_n_s_singularity,0};
    PyObject* values[12] = {0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 1); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 2); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 3); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrVal)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 4); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 5); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColInd)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 6); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_b)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 7); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_tol)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 8); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_reorder)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 9); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 10); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_singularity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, 11); __PYX_ERR(0, 2003, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zcsrlsvqr") < 0)) __PYX_ERR(0, 2003, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 12) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2003, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2003, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2003, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2003, __pyx_L3_error)
    __pyx_v_csrVal = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrVal == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2003, __pyx_L3_error)
    __pyx_v_csrRowPtr = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtr == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2004, __pyx_L3_error)
    __pyx_v_csrColInd = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColInd == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2004, __pyx_L3_error)
    __pyx_v_b = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_b == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2004, __pyx_L3_error)
    __pyx_v_tol = __pyx_PyFloat_AsDouble(values[8]); if (unlikely((__pyx_v_tol == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2004, __pyx_L3_error)
    __pyx_v_reorder = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_reorder == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2005, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[10]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2005, __pyx_L3_error)
    __pyx_v_singularity = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_singularity == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2005, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zcsrlsvqr", 1, 12, 12, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2003, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_262zcsrlsvqr(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_262zcsrlsvqr(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrVal, size_t __pyx_v_csrRowPtr, size_t __pyx_v_csrColInd, size_t __pyx_v_b, double __pyx_v_tol, int __pyx_v_reorder, size_t __pyx_v_x, size_t __pyx_v_singularity) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zcsrlsvqr", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zcsrlsvqr(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrVal, __pyx_v_csrRowPtr, __pyx_v_csrColInd, __pyx_v_b, __pyx_v_tol, __pyx_v_reorder, __pyx_v_x, __pyx_v_singularity, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsrlsvqr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":2016
 *     check_status(status)
 * 
 * cpdef scsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  float mu0, size_t x0, int maxite, float eps, size_t mu,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_265scsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_scsreigvsi(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, float __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, float __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsreigvsi", 0);

  /* "cupy/cuda/cusolver.pyx":2021
 *                  size_t x):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpScsreigvsi(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2022
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpScsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":2023
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpScsreigvsi(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const float*>csrValA, <const int*>csrRowPtrA,
 */
        __pyx_v_status = cusolverSpScsreigvsi(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((float const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), __pyx_v_mu0, ((float const *)__pyx_v_x0), __pyx_v_maxite, __pyx_v_eps, ((float *)__pyx_v_mu), ((float *)__pyx_v_x));
      }

      /* "cupy/cuda/cusolver.pyx":2022
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpScsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":2028
 *             <const int*>csrColIndA, mu0, <const float*>x0, maxite, eps,
 *             <float*>mu, <float*>x)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef dcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2028, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2016
 *     check_status(status)
 * 
 * cpdef scsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  float mu0, size_t x0, int maxite, float eps, size_t mu,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_265scsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_264scsreigvsi[] = "scsreigvsi(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, float mu0, size_t x0, int maxite, float eps, size_t mu, size_t x)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_265scsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  float __pyx_v_mu0;
  size_t __pyx_v_x0;
  int __pyx_v_maxite;
  float __pyx_v_eps;
  size_t __pyx_v_mu;
  size_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("scsreigvsi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_mu0,&__pyx_n_s_x0,&__pyx_n_s_maxite,&__pyx_n_s_eps,&__pyx_n_s_mu,&__pyx_n_s_x,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 1); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 2); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 3); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 4); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 5); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 6); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 7); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 8); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxite)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 9); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 10); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 11); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, 12); __PYX_ERR(0, 2016, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "scsreigvsi") < 0)) __PYX_ERR(0, 2016, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2016, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2016, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2016, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2016, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2017, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2017, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2017, __pyx_L3_error)
    __pyx_v_mu0 = __pyx_PyFloat_AsFloat(values[7]); if (unlikely((__pyx_v_mu0 == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L3_error)
    __pyx_v_x0 = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_x0 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L3_error)
    __pyx_v_maxite = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_maxite == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L3_error)
    __pyx_v_eps = __pyx_PyFloat_AsFloat(values[10]); if (unlikely((__pyx_v_eps == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L3_error)
    __pyx_v_mu = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_mu == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2019, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("scsreigvsi", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2016, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_264scsreigvsi(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_264scsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, float __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, float __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("scsreigvsi", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_scsreigvsi(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2016, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.scsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":2030
 *     check_status(status)
 * 
 * cpdef dcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  double mu0, size_t x0, int maxite, double eps, size_t mu,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_267dcsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_dcsreigvsi(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, double __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, double __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsreigvsi", 0);

  /* "cupy/cuda/cusolver.pyx":2035
 *                  size_t x):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpDcsreigvsi(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2036
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDcsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":2037
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpDcsreigvsi(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const double*>csrValA, <const int*>csrRowPtrA,
 */
        __pyx_v_status = cusolverSpDcsreigvsi(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((double const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), __pyx_v_mu0, ((double const *)__pyx_v_x0), __pyx_v_maxite, __pyx_v_eps, ((double *)__pyx_v_mu), ((double *)__pyx_v_x));
      }

      /* "cupy/cuda/cusolver.pyx":2036
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpDcsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":2042
 *             <const int*>csrColIndA, mu0, <const double*>x0, maxite, eps,
 *             <double*>mu, <double*>x)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef ccsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2030
 *     check_status(status)
 * 
 * cpdef dcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  double mu0, size_t x0, int maxite, double eps, size_t mu,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_267dcsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_266dcsreigvsi[] = "dcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, double mu0, size_t x0, int maxite, double eps, size_t mu, size_t x)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_267dcsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  double __pyx_v_mu0;
  size_t __pyx_v_x0;
  int __pyx_v_maxite;
  double __pyx_v_eps;
  size_t __pyx_v_mu;
  size_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dcsreigvsi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_mu0,&__pyx_n_s_x0,&__pyx_n_s_maxite,&__pyx_n_s_eps,&__pyx_n_s_mu,&__pyx_n_s_x,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 1); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 2); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 3); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 4); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 5); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 6); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 7); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 8); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxite)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 9); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 10); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 11); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, 12); __PYX_ERR(0, 2030, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "dcsreigvsi") < 0)) __PYX_ERR(0, 2030, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2030, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2030, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2030, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2030, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2031, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2031, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2031, __pyx_L3_error)
    __pyx_v_mu0 = __pyx_PyFloat_AsDouble(values[7]); if (unlikely((__pyx_v_mu0 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2032, __pyx_L3_error)
    __pyx_v_x0 = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_x0 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2032, __pyx_L3_error)
    __pyx_v_maxite = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_maxite == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2032, __pyx_L3_error)
    __pyx_v_eps = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2032, __pyx_L3_error)
    __pyx_v_mu = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_mu == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2032, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2033, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("dcsreigvsi", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2030, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_266dcsreigvsi(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_266dcsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, double __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, double __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("dcsreigvsi", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_dcsreigvsi(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2030, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.dcsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":2044
 *     check_status(status)
 * 
 * cpdef ccsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  size_t mu0, size_t x0, int maxite, float eps, size_t mu,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_269ccsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_ccsreigvsi(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, float __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ccsreigvsi", 0);

  /* "cupy/cuda/cusolver.pyx":2049
 *                  size_t x):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpCcsreigvsi(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2049, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2050
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCcsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":2051
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpCcsreigvsi(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const cuComplex*>csrValA, <const int*>csrRowPtrA,
 */
        __pyx_v_status = cusolverSpCcsreigvsi(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((cuComplex const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), (((cuComplex *)__pyx_v_mu0)[0]), ((cuComplex const *)__pyx_v_x0), __pyx_v_maxite, __pyx_v_eps, ((cuComplex *)__pyx_v_mu), ((cuComplex *)__pyx_v_x));
      }

      /* "cupy/cuda/cusolver.pyx":2050
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpCcsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":2056
 *             <const int*>csrColIndA, (<cuComplex*>mu0)[0], <const cuComplex*>x0,
 *             maxite, eps, <cuComplex*>mu, <cuComplex*>x)
 *     check_status(status)             # <<<<<<<<<<<<<<
 * 
 * cpdef zcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2056, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2044
 *     check_status(status)
 * 
 * cpdef ccsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  size_t mu0, size_t x0, int maxite, float eps, size_t mu,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_269ccsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_268ccsreigvsi[] = "ccsreigvsi(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t mu0, size_t x0, int maxite, float eps, size_t mu, size_t x)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_269ccsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_mu0;
  size_t __pyx_v_x0;
  int __pyx_v_maxite;
  float __pyx_v_eps;
  size_t __pyx_v_mu;
  size_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ccsreigvsi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_mu0,&__pyx_n_s_x0,&__pyx_n_s_maxite,&__pyx_n_s_eps,&__pyx_n_s_mu,&__pyx_n_s_x,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 1); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 2); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 3); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 4); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 5); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 6); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 7); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 8); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxite)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 9); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 10); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 11); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, 12); __PYX_ERR(0, 2044, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ccsreigvsi") < 0)) __PYX_ERR(0, 2044, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2044, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2044, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2044, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2044, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2045, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2045, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2045, __pyx_L3_error)
    __pyx_v_mu0 = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_mu0 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2046, __pyx_L3_error)
    __pyx_v_x0 = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_x0 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2046, __pyx_L3_error)
    __pyx_v_maxite = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_maxite == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2046, __pyx_L3_error)
    __pyx_v_eps = __pyx_PyFloat_AsFloat(values[10]); if (unlikely((__pyx_v_eps == (float)-1) && PyErr_Occurred())) __PYX_ERR(0, 2046, __pyx_L3_error)
    __pyx_v_mu = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_mu == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2046, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2047, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ccsreigvsi", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2044, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_268ccsreigvsi(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_268ccsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, float __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("ccsreigvsi", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_ccsreigvsi(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2044, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.ccsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cupy/cuda/cusolver.pyx":2058
 *     check_status(status)
 * 
 * cpdef zcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  size_t mu0, size_t x0, int maxite, double eps, size_t mu,
 */

static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_271zcsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_f_4cupy_4cuda_8cusolver_zcsreigvsi(intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, double __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x, CYTHON_UNUSED int __pyx_skip_dispatch) {
  int __pyx_v_status;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zcsreigvsi", 0);

  /* "cupy/cuda/cusolver.pyx":2063
 *                  size_t x):
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())             # <<<<<<<<<<<<<<
 *     with nogil:
 *         status = cusolverSpZcsreigvsi(
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_spSetStream(__pyx_v_handle, __pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr(), 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2063, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2064
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpZcsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "cupy/cuda/cusolver.pyx":2065
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:
 *         status = cusolverSpZcsreigvsi(             # <<<<<<<<<<<<<<
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 *             <const cuDoubleComplex*>csrValA, <const int*>csrRowPtrA,
 */
        __pyx_v_status = cusolverSpZcsreigvsi(((cusolverSpHandle_t)__pyx_v_handle), __pyx_v_m, __pyx_v_nnz, ((cusparseMatDescr_t const )__pyx_v_descrA), ((cuDoubleComplex const *)__pyx_v_csrValA), ((int const *)__pyx_v_csrRowPtrA), ((int const *)__pyx_v_csrColIndA), (((cuDoubleComplex *)__pyx_v_mu0)[0]), ((cuDoubleComplex const *)__pyx_v_x0), __pyx_v_maxite, __pyx_v_eps, ((cuDoubleComplex *)__pyx_v_mu), ((cuDoubleComplex *)__pyx_v_x));
      }

      /* "cupy/cuda/cusolver.pyx":2064
 *     cdef int status
 *     spSetStream(handle, stream_module.get_current_stream_ptr())
 *     with nogil:             # <<<<<<<<<<<<<<
 *         status = cusolverSpZcsreigvsi(
 *             <SpHandle>handle, m, nnz, <const MatDescr>descrA,
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "cupy/cuda/cusolver.pyx":2071
 *             <const cuDoubleComplex*>x0, maxite,
 *             eps, <cuDoubleComplex*>mu, <cuDoubleComplex*>x)
 *     check_status(status)             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_check_status(__pyx_v_status, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2071, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":2058
 *     check_status(status)
 * 
 * cpdef zcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA,             # <<<<<<<<<<<<<<
 *                  size_t csrValA, size_t csrRowPtrA, size_t csrColIndA,
 *                  size_t mu0, size_t x0, int maxite, double eps, size_t mu,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_271zcsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_4cupy_4cuda_8cusolver_270zcsreigvsi[] = "zcsreigvsi(intptr_t handle, int m, int nnz, size_t descrA, size_t csrValA, size_t csrRowPtrA, size_t csrColIndA, size_t mu0, size_t x0, int maxite, double eps, size_t mu, size_t x)";
static PyObject *__pyx_pw_4cupy_4cuda_8cusolver_271zcsreigvsi(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  intptr_t __pyx_v_handle;
  int __pyx_v_m;
  int __pyx_v_nnz;
  size_t __pyx_v_descrA;
  size_t __pyx_v_csrValA;
  size_t __pyx_v_csrRowPtrA;
  size_t __pyx_v_csrColIndA;
  size_t __pyx_v_mu0;
  size_t __pyx_v_x0;
  int __pyx_v_maxite;
  double __pyx_v_eps;
  size_t __pyx_v_mu;
  size_t __pyx_v_x;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("zcsreigvsi (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_handle,&__pyx_n_s_m,&__pyx_n_s_nnz,&__pyx_n_s_descrA,&__pyx_n_s_csrValA,&__pyx_n_s_csrRowPtrA,&__pyx_n_s_csrColIndA,&__pyx_n_s_mu0,&__pyx_n_s_x0,&__pyx_n_s_maxite,&__pyx_n_s_eps,&__pyx_n_s_mu,&__pyx_n_s_x,0};
    PyObject* values[13] = {0,0,0,0,0,0,0,0,0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case 13: values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
        CYTHON_FALLTHROUGH;
        case 12: values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
        CYTHON_FALLTHROUGH;
        case 11: values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
        CYTHON_FALLTHROUGH;
        case 10: values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
        CYTHON_FALLTHROUGH;
        case  9: values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
        CYTHON_FALLTHROUGH;
        case  8: values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
        CYTHON_FALLTHROUGH;
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_handle)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_m)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 1); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_nnz)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 2); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_descrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 3); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrValA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 4); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (likely((values[5] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrRowPtrA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 5); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (likely((values[6] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_csrColIndA)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 6); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  7:
        if (likely((values[7] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 7); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  8:
        if (likely((values[8] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x0)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 8); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  9:
        if (likely((values[9] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_maxite)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 9); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 10:
        if (likely((values[10] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_eps)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 10); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 11:
        if (likely((values[11] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mu)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 11); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case 12:
        if (likely((values[12] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_x)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, 12); __PYX_ERR(0, 2058, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "zcsreigvsi") < 0)) __PYX_ERR(0, 2058, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 13) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
      values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
      values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
      values[7] = PyTuple_GET_ITEM(__pyx_args, 7);
      values[8] = PyTuple_GET_ITEM(__pyx_args, 8);
      values[9] = PyTuple_GET_ITEM(__pyx_args, 9);
      values[10] = PyTuple_GET_ITEM(__pyx_args, 10);
      values[11] = PyTuple_GET_ITEM(__pyx_args, 11);
      values[12] = PyTuple_GET_ITEM(__pyx_args, 12);
    }
    __pyx_v_handle = PyInt_AsSsize_t(values[0]); if (unlikely((__pyx_v_handle == ((intptr_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 2058, __pyx_L3_error)
    __pyx_v_m = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_m == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2058, __pyx_L3_error)
    __pyx_v_nnz = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_nnz == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2058, __pyx_L3_error)
    __pyx_v_descrA = __Pyx_PyInt_As_size_t(values[3]); if (unlikely((__pyx_v_descrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2058, __pyx_L3_error)
    __pyx_v_csrValA = __Pyx_PyInt_As_size_t(values[4]); if (unlikely((__pyx_v_csrValA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2059, __pyx_L3_error)
    __pyx_v_csrRowPtrA = __Pyx_PyInt_As_size_t(values[5]); if (unlikely((__pyx_v_csrRowPtrA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2059, __pyx_L3_error)
    __pyx_v_csrColIndA = __Pyx_PyInt_As_size_t(values[6]); if (unlikely((__pyx_v_csrColIndA == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2059, __pyx_L3_error)
    __pyx_v_mu0 = __Pyx_PyInt_As_size_t(values[7]); if (unlikely((__pyx_v_mu0 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2060, __pyx_L3_error)
    __pyx_v_x0 = __Pyx_PyInt_As_size_t(values[8]); if (unlikely((__pyx_v_x0 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2060, __pyx_L3_error)
    __pyx_v_maxite = __Pyx_PyInt_As_int(values[9]); if (unlikely((__pyx_v_maxite == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2060, __pyx_L3_error)
    __pyx_v_eps = __pyx_PyFloat_AsDouble(values[10]); if (unlikely((__pyx_v_eps == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 2060, __pyx_L3_error)
    __pyx_v_mu = __Pyx_PyInt_As_size_t(values[11]); if (unlikely((__pyx_v_mu == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2060, __pyx_L3_error)
    __pyx_v_x = __Pyx_PyInt_As_size_t(values[12]); if (unlikely((__pyx_v_x == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 2061, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("zcsreigvsi", 1, 13, 13, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 2058, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_4cupy_4cuda_8cusolver_270zcsreigvsi(__pyx_self, __pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_4cupy_4cuda_8cusolver_270zcsreigvsi(CYTHON_UNUSED PyObject *__pyx_self, intptr_t __pyx_v_handle, int __pyx_v_m, int __pyx_v_nnz, size_t __pyx_v_descrA, size_t __pyx_v_csrValA, size_t __pyx_v_csrRowPtrA, size_t __pyx_v_csrColIndA, size_t __pyx_v_mu0, size_t __pyx_v_x0, int __pyx_v_maxite, double __pyx_v_eps, size_t __pyx_v_mu, size_t __pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("zcsreigvsi", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_4cupy_4cuda_8cusolver_zcsreigvsi(__pyx_v_handle, __pyx_v_m, __pyx_v_nnz, __pyx_v_descrA, __pyx_v_csrValA, __pyx_v_csrRowPtrA, __pyx_v_csrColIndA, __pyx_v_mu0, __pyx_v_x0, __pyx_v_maxite, __pyx_v_eps, __pyx_v_mu, __pyx_v_x, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2058, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("cupy.cuda.cusolver.zcsreigvsi", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {"check_status", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_1check_status, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_check_status},
  {"getProperty", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_3getProperty, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_2getProperty},
  {"_getVersion", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_5_getVersion, METH_NOARGS, __pyx_doc_4cupy_4cuda_8cusolver_4_getVersion},
  {"create", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_7create, METH_NOARGS, __pyx_doc_4cupy_4cuda_8cusolver_6create},
  {"spCreate", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_9spCreate, METH_NOARGS, __pyx_doc_4cupy_4cuda_8cusolver_8spCreate},
  {"destroy", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_11destroy, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_10destroy},
  {"spDestroy", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_13spDestroy, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_12spDestroy},
  {"setStream", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_15setStream, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_14setStream},
  {"getStream", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_17getStream, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_16getStream},
  {"spSetStream", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_19spSetStream, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_18spSetStream},
  {"spGetStream", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_21spGetStream, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_20spGetStream},
  {"spotrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_23spotrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_22spotrf_bufferSize},
  {"dpotrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_25dpotrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_24dpotrf_bufferSize},
  {"cpotrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_27cpotrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_26cpotrf_bufferSize},
  {"zpotrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_29zpotrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_28zpotrf_bufferSize},
  {"spotrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_31spotrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_30spotrf},
  {"dpotrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_33dpotrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_32dpotrf},
  {"cpotrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_35cpotrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_34cpotrf},
  {"zpotrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_37zpotrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_36zpotrf},
  {"spotrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_39spotrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_38spotrs},
  {"dpotrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_41dpotrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_40dpotrs},
  {"cpotrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_43cpotrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_42cpotrs},
  {"zpotrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_45zpotrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_44zpotrs},
  {"sgetrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_47sgetrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_46sgetrf_bufferSize},
  {"dgetrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_49dgetrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_48dgetrf_bufferSize},
  {"cgetrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_51cgetrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_50cgetrf_bufferSize},
  {"zgetrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_53zgetrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_52zgetrf_bufferSize},
  {"sgetrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_55sgetrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_54sgetrf},
  {"dgetrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_57dgetrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_56dgetrf},
  {"cgetrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_59cgetrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_58cgetrf},
  {"zgetrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_61zgetrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_60zgetrf},
  {"sgetrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_63sgetrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_62sgetrs},
  {"dgetrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_65dgetrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_64dgetrs},
  {"cgetrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_67cgetrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_66cgetrs},
  {"zgetrs", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_69zgetrs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_68zgetrs},
  {"sgeqrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_71sgeqrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_70sgeqrf_bufferSize},
  {"dgeqrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_73dgeqrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_72dgeqrf_bufferSize},
  {"cgeqrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_75cgeqrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_74cgeqrf_bufferSize},
  {"zgeqrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_77zgeqrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_76zgeqrf_bufferSize},
  {"sgeqrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_79sgeqrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_78sgeqrf},
  {"dgeqrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_81dgeqrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_80dgeqrf},
  {"cgeqrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_83cgeqrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_82cgeqrf},
  {"zgeqrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_85zgeqrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_84zgeqrf},
  {"sorgqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_87sorgqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_86sorgqr_bufferSize},
  {"dorgqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_89dorgqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_88dorgqr_bufferSize},
  {"cungqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_91cungqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_90cungqr_bufferSize},
  {"zungqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_93zungqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_92zungqr_bufferSize},
  {"sorgqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_95sorgqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_94sorgqr},
  {"dorgqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_97dorgqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_96dorgqr},
  {"cungqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_99cungqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_98cungqr},
  {"zungqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_101zungqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_100zungqr},
  {"sormqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_103sormqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_102sormqr_bufferSize},
  {"dormqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_105dormqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_104dormqr_bufferSize},
  {"cunmqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_107cunmqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_106cunmqr_bufferSize},
  {"zunmqr_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_109zunmqr_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_108zunmqr_bufferSize},
  {"sormqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_111sormqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_110sormqr},
  {"dormqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_113dormqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_112dormqr},
  {"cunmqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_115cunmqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_114cunmqr},
  {"zunmqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_117zunmqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_116zunmqr},
  {"cormqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_119cormqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_118cormqr},
  {"zormqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_121zormqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_120zormqr},
  {"ssytrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_123ssytrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_122ssytrf_bufferSize},
  {"dsytrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_125dsytrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_124dsytrf_bufferSize},
  {"csytrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_127csytrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_126csytrf_bufferSize},
  {"zsytrf_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_129zsytrf_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_128zsytrf_bufferSize},
  {"ssytrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_131ssytrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_130ssytrf},
  {"dsytrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_133dsytrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_132dsytrf},
  {"csytrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_135csytrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_134csytrf},
  {"zsytrf", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_137zsytrf, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_136zsytrf},
  {"sgebrd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_139sgebrd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_138sgebrd_bufferSize},
  {"dgebrd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_141dgebrd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_140dgebrd_bufferSize},
  {"cgebrd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_143cgebrd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_142cgebrd_bufferSize},
  {"zgebrd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_145zgebrd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_144zgebrd_bufferSize},
  {"sgebrd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_147sgebrd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_146sgebrd},
  {"dgebrd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_149dgebrd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_148dgebrd},
  {"cgebrd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_151cgebrd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_150cgebrd},
  {"zgebrd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_153zgebrd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_152zgebrd},
  {"sgesvd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_155sgesvd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_154sgesvd_bufferSize},
  {"dgesvd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_157dgesvd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_156dgesvd_bufferSize},
  {"cgesvd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_159cgesvd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_158cgesvd_bufferSize},
  {"zgesvd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_161zgesvd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_160zgesvd_bufferSize},
  {"sgesvd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_163sgesvd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_162sgesvd},
  {"dgesvd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_165dgesvd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_164dgesvd},
  {"cgesvd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_167cgesvd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_166cgesvd},
  {"zgesvd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_169zgesvd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_168zgesvd},
  {"createGesvdjInfo", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_171createGesvdjInfo, METH_NOARGS, __pyx_doc_4cupy_4cuda_8cusolver_170createGesvdjInfo},
  {"destroyGesvdjInfo", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_173destroyGesvdjInfo, METH_O, __pyx_doc_4cupy_4cuda_8cusolver_172destroyGesvdjInfo},
  {"xgesvdjSetTolerance", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_175xgesvdjSetTolerance, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_174xgesvdjSetTolerance},
  {"xgesvdjSetMaxSweeps", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_177xgesvdjSetMaxSweeps, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_176xgesvdjSetMaxSweeps},
  {"xgesvdjSetSortEig", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_179xgesvdjSetSortEig, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_178xgesvdjSetSortEig},
  {"xgesvdjGetResidual", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_181xgesvdjGetResidual, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_180xgesvdjGetResidual},
  {"xgesvdjGetSweeps", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_183xgesvdjGetSweeps, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_182xgesvdjGetSweeps},
  {"sgesvdj_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_185sgesvdj_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_184sgesvdj_bufferSize},
  {"dgesvdj_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_187dgesvdj_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_186dgesvdj_bufferSize},
  {"cgesvdj_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_189cgesvdj_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_188cgesvdj_bufferSize},
  {"zgesvdj_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_191zgesvdj_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_190zgesvdj_bufferSize},
  {"sgesvdj", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_193sgesvdj, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_192sgesvdj},
  {"dgesvdj", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_195dgesvdj, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_194dgesvdj},
  {"cgesvdj", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_197cgesvdj, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_196cgesvdj},
  {"zgesvdj", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_199zgesvdj, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_198zgesvdj},
  {"sgesvdjBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_201sgesvdjBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_200sgesvdjBatched_bufferSize},
  {"dgesvdjBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_203dgesvdjBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_202dgesvdjBatched_bufferSize},
  {"cgesvdjBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_205cgesvdjBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_204cgesvdjBatched_bufferSize},
  {"zgesvdjBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_207zgesvdjBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_206zgesvdjBatched_bufferSize},
  {"sgesvdjBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_209sgesvdjBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_208sgesvdjBatched},
  {"dgesvdjBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_211dgesvdjBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_210dgesvdjBatched},
  {"cgesvdjBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_213cgesvdjBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_212cgesvdjBatched},
  {"zgesvdjBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_215zgesvdjBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_214zgesvdjBatched},
  {"sgesvdaStridedBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_217sgesvdaStridedBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_216sgesvdaStridedBatched_bufferSize},
  {"dgesvdaStridedBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_219dgesvdaStridedBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_218dgesvdaStridedBatched_bufferSize},
  {"cgesvdaStridedBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_221cgesvdaStridedBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_220cgesvdaStridedBatched_bufferSize},
  {"zgesvdaStridedBatched_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_223zgesvdaStridedBatched_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_222zgesvdaStridedBatched_bufferSize},
  {"sgesvdaStridedBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_225sgesvdaStridedBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_224sgesvdaStridedBatched},
  {"dgesvdaStridedBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_227dgesvdaStridedBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_226dgesvdaStridedBatched},
  {"cgesvdaStridedBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_229cgesvdaStridedBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_228cgesvdaStridedBatched},
  {"zgesvdaStridedBatched", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_231zgesvdaStridedBatched, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_230zgesvdaStridedBatched},
  {"ssyevd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_233ssyevd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_232ssyevd_bufferSize},
  {"dsyevd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_235dsyevd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_234dsyevd_bufferSize},
  {"cheevd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_237cheevd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_236cheevd_bufferSize},
  {"zheevd_bufferSize", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_239zheevd_bufferSize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_238zheevd_bufferSize},
  {"ssyevd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_241ssyevd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_240ssyevd},
  {"dsyevd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_243dsyevd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_242dsyevd},
  {"cheevd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_245cheevd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_244cheevd},
  {"zheevd", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_247zheevd, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_246zheevd},
  {"scsrlsvchol", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_249scsrlsvchol, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_248scsrlsvchol},
  {"dcsrlsvchol", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_251dcsrlsvchol, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_250dcsrlsvchol},
  {"ccsrlsvchol", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_253ccsrlsvchol, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_252ccsrlsvchol},
  {"zcsrlsvchol", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_255zcsrlsvchol, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_254zcsrlsvchol},
  {"scsrlsvqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_257scsrlsvqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_256scsrlsvqr},
  {"dcsrlsvqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_259dcsrlsvqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_258dcsrlsvqr},
  {"ccsrlsvqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_261ccsrlsvqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_260ccsrlsvqr},
  {"zcsrlsvqr", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_263zcsrlsvqr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_262zcsrlsvqr},
  {"scsreigvsi", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_265scsreigvsi, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_264scsreigvsi},
  {"dcsreigvsi", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_267dcsreigvsi, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_266dcsreigvsi},
  {"ccsreigvsi", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_269ccsreigvsi, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_268ccsreigvsi},
  {"zcsreigvsi", (PyCFunction)__pyx_pw_4cupy_4cuda_8cusolver_271zcsreigvsi, METH_VARARGS|METH_KEYWORDS, __pyx_doc_4cupy_4cuda_8cusolver_270zcsreigvsi},
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_cusolver(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_cusolver},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "cusolver",
    __pyx_k_Thin_wrapper_of_CUSOLVER, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 0, 1, 1},
  {&__pyx_n_s_B, __pyx_k_B, sizeof(__pyx_k_B), 0, 0, 1, 1},
  {&__pyx_n_s_C, __pyx_k_C, sizeof(__pyx_k_C), 0, 0, 1, 1},
  {&__pyx_n_s_CUSOLVERError, __pyx_k_CUSOLVERError, sizeof(__pyx_k_CUSOLVERError), 0, 0, 1, 1},
  {&__pyx_n_s_CUSOLVERError___init, __pyx_k_CUSOLVERError___init, sizeof(__pyx_k_CUSOLVERError___init), 0, 0, 1, 1},
  {&__pyx_n_s_CUSOLVERError___reduce, __pyx_k_CUSOLVERError___reduce, sizeof(__pyx_k_CUSOLVERError___reduce), 0, 0, 1, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_ALLOC_FAILED, __pyx_k_CUSOLVER_STATUS_ALLOC_FAILED, sizeof(__pyx_k_CUSOLVER_STATUS_ALLOC_FAILED), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_ARCH_MISMATCH, __pyx_k_CUSOLVER_STATUS_ARCH_MISMATCH, sizeof(__pyx_k_CUSOLVER_STATUS_ARCH_MISMATCH), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_EXECUTION_FAILED, __pyx_k_CUSOLVER_STATUS_EXECUTION_FAILED, sizeof(__pyx_k_CUSOLVER_STATUS_EXECUTION_FAILED), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_INTERNAL_ERROR, __pyx_k_CUSOLVER_STATUS_INTERNAL_ERROR, sizeof(__pyx_k_CUSOLVER_STATUS_INTERNAL_ERROR), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_INVALID_LICENSE, __pyx_k_CUSOLVER_STATUS_INVALID_LICENSE, sizeof(__pyx_k_CUSOLVER_STATUS_INVALID_LICENSE), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_INVALID_VALUE, __pyx_k_CUSOLVER_STATUS_INVALID_VALUE, sizeof(__pyx_k_CUSOLVER_STATUS_INVALID_VALUE), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_MAPPING_ERROR, __pyx_k_CUSOLVER_STATUS_MAPPING_ERROR, sizeof(__pyx_k_CUSOLVER_STATUS_MAPPING_ERROR), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_MATRIX_TYPE_NOT, __pyx_k_CUSOLVER_STATUS_MATRIX_TYPE_NOT, sizeof(__pyx_k_CUSOLVER_STATUS_MATRIX_TYPE_NOT), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_NOT_INITIALIZED, __pyx_k_CUSOLVER_STATUS_NOT_INITIALIZED, sizeof(__pyx_k_CUSOLVER_STATUS_NOT_INITIALIZED), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_NOT_SUPPORTED, __pyx_k_CUSOLVER_STATUS_NOT_SUPPORTED, sizeof(__pyx_k_CUSOLVER_STATUS_NOT_SUPPORTED), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_SUCCESS, __pyx_k_CUSOLVER_STATUS_SUCCESS, sizeof(__pyx_k_CUSOLVER_STATUS_SUCCESS), 0, 1, 0, 1},
  {&__pyx_n_u_CUSOLVER_STATUS_ZERO_PIVOT, __pyx_k_CUSOLVER_STATUS_ZERO_PIVOT, sizeof(__pyx_k_CUSOLVER_STATUS_ZERO_PIVOT), 0, 1, 0, 1},
  {&__pyx_n_s_D, __pyx_k_D, sizeof(__pyx_k_D), 0, 0, 1, 1},
  {&__pyx_n_s_E, __pyx_k_E, sizeof(__pyx_k_E), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_S, __pyx_k_S, sizeof(__pyx_k_S), 0, 0, 1, 1},
  {&__pyx_n_s_U, __pyx_k_U, sizeof(__pyx_k_U), 0, 0, 1, 1},
  {&__pyx_n_s_V, __pyx_k_V, sizeof(__pyx_k_V), 0, 0, 1, 1},
  {&__pyx_n_s_VT, __pyx_k_VT, sizeof(__pyx_k_VT), 0, 0, 1, 1},
  {&__pyx_n_s_W, __pyx_k_W, sizeof(__pyx_k_W), 0, 0, 1, 1},
  {&__pyx_n_s_Work, __pyx_k_Work, sizeof(__pyx_k_Work), 0, 0, 1, 1},
  {&__pyx_n_s_b, __pyx_k_b, sizeof(__pyx_k_b), 0, 0, 1, 1},
  {&__pyx_n_s_batchSize, __pyx_k_batchSize, sizeof(__pyx_k_batchSize), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_csrColInd, __pyx_k_csrColInd, sizeof(__pyx_k_csrColInd), 0, 0, 1, 1},
  {&__pyx_n_s_csrColIndA, __pyx_k_csrColIndA, sizeof(__pyx_k_csrColIndA), 0, 0, 1, 1},
  {&__pyx_n_s_csrRowPtr, __pyx_k_csrRowPtr, sizeof(__pyx_k_csrRowPtr), 0, 0, 1, 1},
  {&__pyx_n_s_csrRowPtrA, __pyx_k_csrRowPtrA, sizeof(__pyx_k_csrRowPtrA), 0, 0, 1, 1},
  {&__pyx_n_s_csrVal, __pyx_k_csrVal, sizeof(__pyx_k_csrVal), 0, 0, 1, 1},
  {&__pyx_n_s_csrValA, __pyx_k_csrValA, sizeof(__pyx_k_csrValA), 0, 0, 1, 1},
  {&__pyx_n_s_cupy_cuda_cusolver, __pyx_k_cupy_cuda_cusolver, sizeof(__pyx_k_cupy_cuda_cusolver), 0, 0, 1, 1},
  {&__pyx_kp_s_cupy_cuda_cusolver_pyx, __pyx_k_cupy_cuda_cusolver_pyx, sizeof(__pyx_k_cupy_cuda_cusolver_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_d_A, __pyx_k_d_A, sizeof(__pyx_k_d_A), 0, 0, 1, 1},
  {&__pyx_n_s_d_S, __pyx_k_d_S, sizeof(__pyx_k_d_S), 0, 0, 1, 1},
  {&__pyx_n_s_d_U, __pyx_k_d_U, sizeof(__pyx_k_d_U), 0, 0, 1, 1},
  {&__pyx_n_s_d_V, __pyx_k_d_V, sizeof(__pyx_k_d_V), 0, 0, 1, 1},
  {&__pyx_n_s_d_info, __pyx_k_d_info, sizeof(__pyx_k_d_info), 0, 0, 1, 1},
  {&__pyx_n_s_d_work, __pyx_k_d_work, sizeof(__pyx_k_d_work), 0, 0, 1, 1},
  {&__pyx_n_s_descrA, __pyx_k_descrA, sizeof(__pyx_k_descrA), 0, 0, 1, 1},
  {&__pyx_n_s_devInfo, __pyx_k_devInfo, sizeof(__pyx_k_devInfo), 0, 0, 1, 1},
  {&__pyx_n_s_devIpiv, __pyx_k_devIpiv, sizeof(__pyx_k_devIpiv), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_econ, __pyx_k_econ, sizeof(__pyx_k_econ), 0, 0, 1, 1},
  {&__pyx_n_s_eps, __pyx_k_eps, sizeof(__pyx_k_eps), 0, 0, 1, 1},
  {&__pyx_n_s_h_R_nrmF, __pyx_k_h_R_nrmF, sizeof(__pyx_k_h_R_nrmF), 0, 0, 1, 1},
  {&__pyx_n_s_handle, __pyx_k_handle, sizeof(__pyx_k_handle), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
  {&__pyx_n_s_ipiv, __pyx_k_ipiv, sizeof(__pyx_k_ipiv), 0, 0, 1, 1},
  {&__pyx_n_s_jobu, __pyx_k_jobu, sizeof(__pyx_k_jobu), 0, 0, 1, 1},
  {&__pyx_n_s_jobvt, __pyx_k_jobvt, sizeof(__pyx_k_jobvt), 0, 0, 1, 1},
  {&__pyx_n_s_jobz, __pyx_k_jobz, sizeof(__pyx_k_jobz), 0, 0, 1, 1},
  {&__pyx_n_s_k, __pyx_k_k, sizeof(__pyx_k_k), 0, 0, 1, 1},
  {&__pyx_n_s_lda, __pyx_k_lda, sizeof(__pyx_k_lda), 0, 0, 1, 1},
  {&__pyx_n_s_ldb, __pyx_k_ldb, sizeof(__pyx_k_ldb), 0, 0, 1, 1},
  {&__pyx_n_s_ldc, __pyx_k_ldc, sizeof(__pyx_k_ldc), 0, 0, 1, 1},
  {&__pyx_n_s_ldu, __pyx_k_ldu, sizeof(__pyx_k_ldu), 0, 0, 1, 1},
  {&__pyx_n_s_ldv, __pyx_k_ldv, sizeof(__pyx_k_ldv), 0, 0, 1, 1},
  {&__pyx_n_s_ldvt, __pyx_k_ldvt, sizeof(__pyx_k_ldvt), 0, 0, 1, 1},
  {&__pyx_n_s_lwork, __pyx_k_lwork, sizeof(__pyx_k_lwork), 0, 0, 1, 1},
  {&__pyx_n_s_m, __pyx_k_m, sizeof(__pyx_k_m), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_max_sweeps, __pyx_k_max_sweeps, sizeof(__pyx_k_max_sweeps), 0, 0, 1, 1},
  {&__pyx_n_s_maxite, __pyx_k_maxite, sizeof(__pyx_k_maxite), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_mu, __pyx_k_mu, sizeof(__pyx_k_mu), 0, 0, 1, 1},
  {&__pyx_n_s_mu0, __pyx_k_mu0, sizeof(__pyx_k_mu0), 0, 0, 1, 1},
  {&__pyx_n_s_n, __pyx_k_n, sizeof(__pyx_k_n), 0, 0, 1, 1},
  {&__pyx_n_s_nnz, __pyx_k_nnz, sizeof(__pyx_k_nnz), 0, 0, 1, 1},
  {&__pyx_n_s_nrhs, __pyx_k_nrhs, sizeof(__pyx_k_nrhs), 0, 0, 1, 1},
  {&__pyx_n_s_params, __pyx_k_params, sizeof(__pyx_k_params), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_rank, __pyx_k_rank, sizeof(__pyx_k_rank), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reorder, __pyx_k_reorder, sizeof(__pyx_k_reorder), 0, 0, 1, 1},
  {&__pyx_n_s_rwork, __pyx_k_rwork, sizeof(__pyx_k_rwork), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_side, __pyx_k_side, sizeof(__pyx_k_side), 0, 0, 1, 1},
  {&__pyx_n_s_singularity, __pyx_k_singularity, sizeof(__pyx_k_singularity), 0, 0, 1, 1},
  {&__pyx_n_s_sort_svd, __pyx_k_sort_svd, sizeof(__pyx_k_sort_svd), 0, 0, 1, 1},
  {&__pyx_n_s_status, __pyx_k_status, sizeof(__pyx_k_status), 0, 0, 1, 1},
  {&__pyx_n_s_stream, __pyx_k_stream, sizeof(__pyx_k_stream), 0, 0, 1, 1},
  {&__pyx_n_s_strideA, __pyx_k_strideA, sizeof(__pyx_k_strideA), 0, 0, 1, 1},
  {&__pyx_n_s_strideS, __pyx_k_strideS, sizeof(__pyx_k_strideS), 0, 0, 1, 1},
  {&__pyx_n_s_strideU, __pyx_k_strideU, sizeof(__pyx_k_strideU), 0, 0, 1, 1},
  {&__pyx_n_s_strideV, __pyx_k_strideV, sizeof(__pyx_k_strideV), 0, 0, 1, 1},
  {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
  {&__pyx_n_s_tau, __pyx_k_tau, sizeof(__pyx_k_tau), 0, 0, 1, 1},
  {&__pyx_n_s_tauP, __pyx_k_tauP, sizeof(__pyx_k_tauP), 0, 0, 1, 1},
  {&__pyx_n_s_tauQ, __pyx_k_tauQ, sizeof(__pyx_k_tauQ), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_tol, __pyx_k_tol, sizeof(__pyx_k_tol), 0, 0, 1, 1},
  {&__pyx_n_s_tolerance, __pyx_k_tolerance, sizeof(__pyx_k_tolerance), 0, 0, 1, 1},
  {&__pyx_n_s_trans, __pyx_k_trans, sizeof(__pyx_k_trans), 0, 0, 1, 1},
  {&__pyx_n_s_uplo, __pyx_k_uplo, sizeof(__pyx_k_uplo), 0, 0, 1, 1},
  {&__pyx_n_s_work, __pyx_k_work, sizeof(__pyx_k_work), 0, 0, 1, 1},
  {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
  {&__pyx_n_s_x0, __pyx_k_x0, sizeof(__pyx_k_x0), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 573, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 577, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "cupy/cuda/cusolver.pyx":573
 * 
 * 
 * class CUSOLVERError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, status):
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_builtin_RuntimeError); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "cupy/cuda/cusolver.pyx":575
 * class CUSOLVERError(RuntimeError):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 */
  __pyx_tuple__2 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_status); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);
  __pyx_codeobj__3 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__2, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cupy_cuda_cusolver_pyx, __pyx_n_s_init, 575, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__3)) __PYX_ERR(0, 575, __pyx_L1_error)

  /* "cupy/cuda/cusolver.pyx":579
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_cupy_cuda_cusolver_pyx, __pyx_n_s_reduce, 579, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_11 = PyInt_FromLong(11); if (unlikely(!__pyx_int_11)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __pyx_v_4cupy_4cuda_8cusolver_STATUS = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  if (__Pyx_ExportFunction("getProperty", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_getProperty, "int (int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("_getVersion", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver__getVersion, "PyObject *(int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("create", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_create, "intptr_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spCreate", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_spCreate, "intptr_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroy", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_destroy, "PyObject *(intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spDestroy", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_spDestroy, "PyObject *(intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("setStream", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_setStream, "PyObject *(intptr_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("getStream", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_getStream, "size_t (intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spotrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_spotrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dpotrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dpotrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cpotrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cpotrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zpotrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zpotrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spotrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_spotrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dpotrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dpotrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cpotrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cpotrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zpotrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zpotrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("spotrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_spotrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dpotrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dpotrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cpotrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cpotrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zpotrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zpotrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgetrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgetrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgetrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgetrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgetrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgetrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgetrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgetrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgetrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgetrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgetrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgetrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgetrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgetrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgetrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgetrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgetrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgetrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgetrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgetrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgetrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgetrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgetrs", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgetrs, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgeqrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgeqrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgeqrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgeqrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgeqrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgeqrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgeqrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgeqrf_bufferSize, "int (intptr_t, int, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgeqrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgeqrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgeqrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgeqrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgeqrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgeqrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgeqrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgeqrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sorgqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sorgqr_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dorgqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dorgqr_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cungqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cungqr_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zungqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zungqr_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sorgqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sorgqr, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dorgqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dorgqr, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cungqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cungqr, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zungqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zungqr, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sormqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sormqr_bufferSize, "int (intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dormqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dormqr_bufferSize, "int (intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cunmqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cunmqr_bufferSize, "int (intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zunmqr_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zunmqr_bufferSize, "int (intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sormqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sormqr, "PyObject *(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dormqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dormqr, "PyObject *(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cunmqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cunmqr, "PyObject *(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zunmqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zunmqr, "PyObject *(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cormqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cormqr, "PyObject *(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zormqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zormqr, "PyObject *(intptr_t, int, int, int, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ssytrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ssytrf_bufferSize, "int (intptr_t, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dsytrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dsytrf_bufferSize, "int (intptr_t, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("csytrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_csytrf_bufferSize, "int (intptr_t, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zsytrf_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zsytrf_bufferSize, "int (intptr_t, int, size_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ssytrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ssytrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dsytrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dsytrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("csytrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_csytrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zsytrf", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zsytrf, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgebrd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgebrd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgebrd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgebrd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgebrd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgebrd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgebrd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgebrd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgebrd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgebrd, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgebrd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgebrd, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgebrd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgebrd, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgebrd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgebrd, "PyObject *(intptr_t, int, int, size_t, int, size_t, size_t, size_t, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgesvd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgesvd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgesvd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgesvd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgesvd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgesvd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgesvd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgesvd_bufferSize, "int (intptr_t, int, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgesvd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgesvd, "PyObject *(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgesvd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgesvd, "PyObject *(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgesvd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgesvd, "PyObject *(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgesvd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgesvd, "PyObject *(intptr_t, char, char, int, int, size_t, int, size_t, size_t, int, size_t, int, size_t, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("createGesvdjInfo", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_createGesvdjInfo, "intptr_t (int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("destroyGesvdjInfo", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_destroyGesvdjInfo, "PyObject *(intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("xgesvdjSetTolerance", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetTolerance, "PyObject *(intptr_t, double, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("xgesvdjSetMaxSweeps", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetMaxSweeps, "PyObject *(intptr_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("xgesvdjSetSortEig", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_xgesvdjSetSortEig, "PyObject *(intptr_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("xgesvdjGetResidual", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetResidual, "double (intptr_t, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("xgesvdjGetSweeps", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_xgesvdjGetSweeps, "int (intptr_t, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgesvdj_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgesvdj_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgesvdj_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgesvdj_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgesvdj_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgesvdj_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgesvdj_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgesvdj_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgesvdj", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgesvdj, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgesvdj", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgesvdj, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgesvdj", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgesvdj, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgesvdj", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgesvdj, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, intptr_t, intptr_t, int, intptr_t, int, intptr_t, int, intptr_t, intptr_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgesvdaStridedBatched_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgesvdaStridedBatched_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgesvdaStridedBatched_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgesvdaStridedBatched_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched_bufferSize, "int (intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("sgesvdaStridedBatched", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_sgesvdaStridedBatched, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dgesvdaStridedBatched", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dgesvdaStridedBatched, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cgesvdaStridedBatched", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cgesvdaStridedBatched, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zgesvdaStridedBatched", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zgesvdaStridedBatched, "PyObject *(intptr_t, int, int, int, int, intptr_t, int, PY_LONG_LONG, intptr_t, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, PY_LONG_LONG, intptr_t, int, intptr_t, intptr_t, int, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ssyevd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ssyevd_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dsyevd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dsyevd_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cheevd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cheevd_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zheevd_bufferSize", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zheevd_bufferSize, "int (intptr_t, int, int, int, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ssyevd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ssyevd, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dsyevd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dsyevd, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("cheevd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_cheevd, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zheevd", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zheevd, "PyObject *(intptr_t, int, int, int, size_t, int, size_t, size_t, int, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("scsrlsvchol", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_scsrlsvchol, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dcsrlsvchol", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dcsrlsvchol, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ccsrlsvchol", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ccsrlsvchol, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zcsrlsvchol", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zcsrlsvchol, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("scsrlsvqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_scsrlsvqr, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dcsrlsvqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dcsrlsvqr, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ccsrlsvqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ccsrlsvqr, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, float, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zcsrlsvqr", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zcsrlsvqr, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, double, int, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("scsreigvsi", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_scsreigvsi, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, float, size_t, int, float, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("dcsreigvsi", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_dcsreigvsi, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, double, size_t, int, double, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("ccsreigvsi", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_ccsreigvsi, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, float, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ExportFunction("zcsreigvsi", (void (*)(void))__pyx_f_4cupy_4cuda_8cusolver_zcsreigvsi, "PyObject *(intptr_t, int, int, size_t, size_t, size_t, size_t, size_t, size_t, int, double, size_t, size_t, int __pyx_skip_dispatch)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("cupy.cuda.stream"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_current_stream_ptr", (void (**)(void))&__pyx_f_4cupy_4cuda_6stream_get_current_stream_ptr, "intptr_t (void)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((optimize("Os")))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initcusolver(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initcusolver(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_cusolver(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_cusolver(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_cusolver(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_cusolver(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("cusolver", __pyx_methods, __pyx_k_Thin_wrapper_of_CUSOLVER, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_cupy__cuda__cusolver) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "cupy.cuda.cusolver")) {
      if (unlikely(PyDict_SetItemString(modules, "cupy.cuda.cusolver", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  if (unlikely(__Pyx_modinit_function_export_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "cupy/cuda/cusolver.pyx":558
 * 
 * cdef dict STATUS = {
 *     0: 'CUSOLVER_STATUS_SUCCESS',             # <<<<<<<<<<<<<<
 *     1: 'CUSOLVER_STATUS_NOT_INITIALIZED',
 *     2: 'CUSOLVER_STATUS_ALLOC_FAILED',
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_0, __pyx_n_u_CUSOLVER_STATUS_SUCCESS) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_1, __pyx_n_u_CUSOLVER_STATUS_NOT_INITIALIZED) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_2, __pyx_n_u_CUSOLVER_STATUS_ALLOC_FAILED) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_3, __pyx_n_u_CUSOLVER_STATUS_INVALID_VALUE) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_4, __pyx_n_u_CUSOLVER_STATUS_ARCH_MISMATCH) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_5, __pyx_n_u_CUSOLVER_STATUS_MAPPING_ERROR) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_6, __pyx_n_u_CUSOLVER_STATUS_EXECUTION_FAILED) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_7, __pyx_n_u_CUSOLVER_STATUS_INTERNAL_ERROR) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_8, __pyx_n_u_CUSOLVER_STATUS_MATRIX_TYPE_NOT) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_9, __pyx_n_u_CUSOLVER_STATUS_NOT_SUPPORTED) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_10, __pyx_n_u_CUSOLVER_STATUS_ZERO_PIVOT) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_int_11, __pyx_n_u_CUSOLVER_STATUS_INVALID_LICENSE) < 0) __PYX_ERR(0, 558, __pyx_L1_error)
  __Pyx_XGOTREF(__pyx_v_4cupy_4cuda_8cusolver_STATUS);
  __Pyx_DECREF_SET(__pyx_v_4cupy_4cuda_8cusolver_STATUS, ((PyObject*)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":573
 * 
 * 
 * class CUSOLVERError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, status):
 */
  __pyx_t_1 = __Pyx_CalculateMetaclass(NULL, __pyx_tuple_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Py3MetaclassPrepare(__pyx_t_1, __pyx_tuple_, __pyx_n_s_CUSOLVERError, __pyx_n_s_CUSOLVERError, (PyObject *) NULL, __pyx_n_s_cupy_cuda_cusolver, (PyObject *) NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "cupy/cuda/cusolver.pyx":575
 * class CUSOLVERError(RuntimeError):
 * 
 *     def __init__(self, status):             # <<<<<<<<<<<<<<
 *         self.status = status
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 */
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4cupy_4cuda_8cusolver_13CUSOLVERError_1__init__, 0, __pyx_n_s_CUSOLVERError___init, NULL, __pyx_n_s_cupy_cuda_cusolver, __pyx_d, ((PyObject *)__pyx_codeobj__3)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_init, __pyx_t_3) < 0) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cupy/cuda/cusolver.pyx":579
 *         super(CUSOLVERError, self).__init__(STATUS[status])
 * 
 *     def __reduce__(self):             # <<<<<<<<<<<<<<
 *         return (type(self), (self.status,))
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_NewEx(&__pyx_mdef_4cupy_4cuda_8cusolver_13CUSOLVERError_3__reduce__, 0, __pyx_n_s_CUSOLVERError___reduce, NULL, __pyx_n_s_cupy_cuda_cusolver, __pyx_d, ((PyObject *)__pyx_codeobj__5)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetNameInClass(__pyx_t_2, __pyx_n_s_reduce, __pyx_t_3) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "cupy/cuda/cusolver.pyx":573
 * 
 * 
 * class CUSOLVERError(RuntimeError):             # <<<<<<<<<<<<<<
 * 
 *     def __init__(self, status):
 */
  __pyx_t_3 = __Pyx_Py3ClassCreate(__pyx_t_1, __pyx_n_s_CUSOLVERError, __pyx_tuple_, __pyx_t_2, NULL, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CUSOLVERError, __pyx_t_3) < 0) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "cupy/cuda/cusolver.pyx":1
 * """Thin wrapper of CUSOLVER."""             # <<<<<<<<<<<<<<
 * 
 * cimport cython  # NOQA
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_MODE_NOVECTOR);
    if (unlikely(!wrapped)) __PYX_ERR(1, 40, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSOLVER_EIG_MODE_NOVECTOR", wrapped) < 0) __PYX_ERR(1, 40, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_MODE_VECTOR);
    if (unlikely(!wrapped)) __PYX_ERR(1, 41, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSOLVER_EIG_MODE_VECTOR", wrapped) < 0) __PYX_ERR(1, 41, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_TYPE_1);
    if (unlikely(!wrapped)) __PYX_ERR(1, 36, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSOLVER_EIG_TYPE_1", wrapped) < 0) __PYX_ERR(1, 36, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_TYPE_2);
    if (unlikely(!wrapped)) __PYX_ERR(1, 37, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSOLVER_EIG_TYPE_2", wrapped) < 0) __PYX_ERR(1, 37, __pyx_L1_error);
  }
  {
    PyObject* wrapped = __Pyx_PyInt_From_int(__pyx_e_4cupy_4cuda_8cusolver_CUSOLVER_EIG_TYPE_3);
    if (unlikely(!wrapped)) __PYX_ERR(1, 38, __pyx_L1_error)
    if (PyObject_SetAttrString(__pyx_m, "CUSOLVER_EIG_TYPE_3", wrapped) < 0) __PYX_ERR(1, 38, __pyx_L1_error);
  }

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init cupy.cuda.cusolver", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init cupy.cuda.cusolver");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    if (likely(result)) {
        Py_INCREF(result);
    } else if (unlikely(PyErr_Occurred())) {
        result = NULL;
    } else {
#else
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCall */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* DictGetItem */
    #if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyCFunctionFastCall */
    #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
    #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCallMethO */
    #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
    #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyErrFetchRestore */
    #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
    #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* WriteUnraisableException */
    static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* CalculateMetaclass */
    static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* FetchCommonType */
    static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
    #include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Py3ClassCreate */
        static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CLineInTraceback */
        #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
        static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
        #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
        #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) -1, const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CIntFromPy */
        static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (PY_LONG_LONG) -1;
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntToPy */
        static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
        static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FastTypeChecks */
        #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
        static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionExport */
        static int __Pyx_ExportFunction(const char *name, void (*f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(__pyx_m, (char *)"__pyx_capi__");
    if (!d) {
        PyErr_Clear();
        d = PyDict_New();
        if (!d)
            goto bad;
        Py_INCREF(d);
        if (PyModule_AddObject(__pyx_m, (char *)"__pyx_capi__", d) < 0)
            goto bad;
    }
    tmp.fp = f;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(tmp.p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(tmp.p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItemString(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* ModuleImport */
        #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* FunctionImport */
        #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
        static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
